// cfpaccessor.C

#include "cfpinc.h"

//Not implemented in gcc yet (2.8.1)
//using namespace std;
//namespace Configpp{
















//--------------------------------------------------------------------------------
//---- CfpVAccessor --------------------------------------------------------------
//--------------------------------------------------------------------------------

CfpVAccessor::CfpVAccessor(string inname,CfpObject *objp,CfpVal *invalp)
  : CfpAccessor(inname)
{
  // base constructor
  // Note that objp points to the object which owns/created the accessor, it may be null.
  //  if it is not null, the object class is retrieved from the object pointer,
  //  and the class is queried for information on the variable/arg that the accessor name matches.
  //  Sometimes what is wanted is internal accessors which don't match any variable/arg.
  //  in these cases, the first character of the name should be an underscore.  this tells
  //  the constructor not to look for a matching member var/arg.
  CfpClass *classp;

  // set object, class, and member pointers
  objectp=objp;
  if (objp)
    classp=objp->getclassp();
  else
    classp=NULL;

  if (classp!=NULL && inname!="" && inname.substr(0,1)!="_")
    memberp=classp->lookup_memberva(inname,true);
  else
    memberp=NULL;

  if (invalp)
    {
      // caller allocated our cfpval
      valislocal=false;
      valp=invalp;
    }
  else
    {
      // we allocate a new cfpval and take care of it
      valislocal=true;
      valp=new CfpVal();
    }

  listenercount=0;
  speakercount=0;
  secretchanges=false;
  // test
  secretchanges=true;
}


CfpVAccessor::~CfpVAccessor()
{
  // destruct the accessor
  if (valislocal)
    delete valp;

  // delete the listener stubs which we created
  list<CfpVAccstub *>::iterator li=listeners.begin();
  list<CfpVAccstub *>::iterator lend;
  lend=listeners.end();
  while (li!=lend)
    {
      delete (*li);
      li++;
    }


  //  // delete the speaker stubs which we created
  //  list<CfpVAccstub *>::iterator si=speakers.begin();
  //  list<CfpVAccstub *>::iterator send;
  //  
  //  send=speakers.end();
  //  while (si!=send)
  //    {
  //      delete (*si);
  //      si++;
  //    }
}






string CfpVAccessor::getlabel()
  {
    // get the label for the accessor, if one exists
    if (memberp)
      return memberp->getlabel();
    else
      return "";
  }

string CfpVAccessor::getunitstr()
  {
    // get the units for the accessor, if one exists
    if (memberp)
      return memberp->getunitstr();
    else
      return "";
  }

string CfpVAccessor::getdesc()
  {
    // get the description for the accessor, if one exists
    if (memberp)
      return memberp->getdesc();
    else
      return "";
  }


string CfpVAccessor::getclassname()
  {
    // get the name of the ownerclass, if one exists
    if (objectp)
      return (objectp->classname());
    else
      return "";
  }

string CfpVAccessor::getobjname()
  {
    // get the name of the ownerobject, if one exists
    if (objectp)
      return objectp->prettyname();
    else
      return "";
  }

string CfpVAccessor::getobjtrail()
  {
    // get the trail of the ownerobject, if one exists
    if (objectp)
      return objectp->gettrail();
    else
      return "";
  }





bool CfpVAccessor::isdefined()
  {
    if (valp==NULL)
      return false;
    if (!valp->isdefined())
      return false;
    return true;
  }




//--------------------------------------------------------------------------------
//---- CfpVAccessor::addlistener() -----------------------------------------------
//--------------------------------------------------------------------------------

CfpVAccstub *CfpVAccessor::addlistener(CfpVAccessor *vlistenerp, int fromindex=-1, int toindex=-1)
{
  // register an accessor which will get notified about changes

  // first create a stub that lets us hold onto the index into the accessor
  CfpVAccstub *listenerp=new CfpVAccstub(this,vlistenerp,fromindex,toindex);

  //  cout << "added a listener from '" << getname() << "' index " << JRtostr(fromindex) << " to '"<< vlistenerp->getname() << "' index " << JRtostr(toindex) << ".\n";

  listeners.push_back(listenerp);
  ++listenercount;

  // now add the complementary speaker entry for the target
  vlistenerp->addspeaker(listenerp);

  return listenerp;
}


//--------------------------------------------------------------------------------
//---- CfpVAccessor::addspeaker() -----------------------------------------------
//--------------------------------------------------------------------------------

void CfpVAccessor::addspeaker(CfpVAccstub *astub)
{
  // keep track of who we listen to
  speakers.push_back(astub);
  ++speakercount;
}



//--------------------------------------------------------------------------------
//---- CfpVAccessor::correctspeaker() -----------------------------------------------
//--------------------------------------------------------------------------------

void CfpVAccessor::correctspeaker(CfpVAccstub *oldstub,CfpVAccstub *newstub)
{
  // due to a splicing operation, who we listen to has changed.
  // so we need to find the place we stored our listener stub and replace it

  //  cout << "... Begining checking speaker of '"+getname()+"' which has "+JRtostr(speakercount)+" speakers.\n";

  list<CfpVAccstub *>::iterator si=speakers.begin();
  list<CfpVAccstub *>::iterator send;
  send=speakers.end();
  while (si!=send)
    {
      //      cout << "... Checking speaker of '"+getname()+".\n";
      if (*si==oldstub)
	{
	  //	  cout << "... Succesfully corrected a speaker\n";
	  speakers.erase(si);
	  *si=newstub;
	  return;
	}
      si++;
    }
  cout << ".. UNSuccesful at correcteding a speaker\n";
}





//--------------------------------------------------------------------------------
//---- CfpVAccessor::notify() ----------------------------------------------------
//--------------------------------------------------------------------------------

void CfpVAccessor::notify(CfpVal *outvalp)
{
  // someone has changed our value, so we notify any listeners
  list<CfpVAccstub *>::iterator li=listeners.begin();
  list<CfpVAccstub *>::iterator lend;

  lend=listeners.end();
  while (li!=lend)
    {
      //      cout << "NOTIFYING.\n";
      //      cout << "NOTIFYING " << (*li)->getname() " << " of " << outvalp->get_string()<<".\n";
      (*li)->set(outvalp);
      li++;
    }
}







//--------------------------------------------------------------------------------
//---- CfpVAccessor:set(CfpVal val, int index) -----------------------------------
//--------------------------------------------------------------------------------
void CfpVAccessor::set(CfpVal *valp, int index)
{
  // set the value of a vector element

  if (index==-1)
    {
      // just shorthand for normal scalar set
      //      cout << "shorthanding setting.\n";
      //      if (getname()!="_timerac")
      //	cout << "shorthand setting to " << valp->get_string() << ".\n";
      set(valp);
    }
  else
    {
      // for the base class, which is a scalar accessor, this results in an error
      throw CfpError("attempt to set vector element of '"+getname()+"', which is a scalar accessor.","",classerr,false,false);
    }
}

//--------------------------------------------------------------------------------
//---- CfpVAccessor:get(int index) -----------------------------------------------
//--------------------------------------------------------------------------------
CfpVal CfpVAccessor::get(int index)
{
  // get the value of a vector element
  if (index==-1)
    {
      // just shorthand for normal scalar set
      return get();
    }
  else
    {
      // for the base class, which is a scalar accessor, this results in an error
      throw CfpError("attempt to get vector element of '"+getname()+"', which is a scalar accessor.","",classerr,false,false);
    }
}

//--------------------------------------------------------------------------------
//---- CfpVAccessor:parseindex(string ) ------------------------------------------
//--------------------------------------------------------------------------------
int CfpVAccessor::parseindex(string vectoradstring)
{
  // set the value of a vector element

  // for the base class, which is a scalar accessor, this results in an error
  throw CfpError("attempt to calculate index '"+vectoradstring+"' into '"+getname()+"', which is a scalar accessor.","",classerr,false,false);
}














//--------------------------------------------------------------------------------
//---- CfpVAccessor:getstr() -----------------------------------------------------
//--------------------------------------------------------------------------------

string CfpVAccessor::getstr()
{
  if (!isdefined())
    throw CfpError("attempt to access the value of '"+getname()+"', which has not yet been assigned.","",syntactical,false,false);
  return valp->get_string();
}



//--------------------------------------------------------------------------------
//---- CfpVAccessor:getcharp() ---------------------------------------------------
//--------------------------------------------------------------------------------

const char *CfpVAccessor::getcharp()
{
  if (!isdefined())
    {
      throw CfpError("attempt to access the value of '"+getname()+"', which has not yet been assigned.","",syntactical,false,false);
    }
  return valp->get_charp();
}





void CfpVAccessor::retrieveandset(int index=-1)
{
  // what we want to do here is makesure we have the latest input, and notify our listeners of it
  // this is used from the proxy, so that after some internal class activation, the newest output variables
  //  are retrieved and then delivered.  (it's like manually retrieving sources of wires that are ext. triggered).
  if (index==-1)
    {
      CfpVal val;
      val=get (index);
      if (val.isdefined())
	changes(&val);
      //      else
      //	printf("... skipping retrieveand set because of undefined val.\n");
    }
  else
    {
      // with a vector, the target value will be retrieved inside the changes() call
      changes(index);
    }
}




//--------------------------------------------------------------------------------
//---- CfpVAccstub::listenersplice -----------------------------------------------
//--------------------------------------------------------------------------------

CfpVAccstub *CfpVAccstub::listenersplice(CfpVAccessor *inacp,int intoindex,CfpVAccessor *outacp,int outtoindex)
{
  // this is a helper function for creating filters
  // this accessor stub is held by a source as a listener, and is called when the source changes to notify our accessor.
  // what we want to do now is splice in a new filter (pair of input&output accessors) which is being passed to us
  CfpVAccstub *newstubp;

  // first we move (respawn) our old listener so that it listens to the new filter output
  newstubp=outacp->addlistener(acp,outtoindex,toindex);
  // and correct the original targets speaker list (to point to the filer output)
  acp->correctspeaker(this,newstubp);

  // now we overwrite our old listener so that it is effectively spliced in to replace the current recipient
  // since it is being spliced in, it keeps the same fromindex, but gets the new toindex and new target accessor
  acp=inacp;
  toindex=intoindex;

  // and then lastly we need to let the filter know that we are its speaker
  inacp->addspeaker(this);

  // note: filter is responsible for processing input when notified and notifying its output (which old listener now listens to)
  return newstubp;
}








//--------------------------------------------------------------------------------
//---- CfpVAccstub::retrieveandset() ---------------------------------------------
//--------------------------------------------------------------------------------

bool CfpVAccstub::retrieveandset()
{
  // get the value from the source and send it to the listener
  CfpVal val;

  val=sourceacp->get(fromindex);

  if (!(val.isdefined()))
    {
      // source value is not initialzied, so we can't very well set it
      //      cout << "Value is not initialized!\n";
      return false;
    }

  acp->set(&val,toindex);
  return true;
}









//--------------------------------------------------------------------------------
//---- CfpVAccessor::latestupdate() -----------------------------------------------
//--------------------------------------------------------------------------------

void CfpVAccessor::latestupdate()
{
  // we want to cycle through all of the sources that project to us,
  // and if we find one that tells us that its value may change without it notifying us
  //  and it is a getable source, then we are to request the latest value

  //  cout << "...  accessor "+getname()+" is in latest update.\n";

  // delete the speaker stubs which we created
  list<CfpVAccstub *>::iterator si=speakers.begin();
  list<CfpVAccstub *>::iterator send;
  send=speakers.end();
  while (si!=send)
    {
      //      cout << "... calling a speaker of "+getname()+".\n";
      (*si)->latestupdate();
      si++;
    }
}




//--------------------------------------------------------------------------------
//---- CfpVAccstub::latestupdate() -----------------------------------------------
//--------------------------------------------------------------------------------

void CfpVAccstub::latestupdate()
{
  // The target end of this stub has decided that it wants the latest value from its input source.
  // So we check if the source says that its value may change without it notifying listeners.
  //  If so, and if it is a manually getable source, then we request the latest value and send it to listener

  //  cout << "...  stub in latestupdate, from "+sourceacp->getname()+" to "+acp->getname()+".\n";

  // first we go up the chain ad tell our source to make sure IT has its latest value
  sourceacp->latestupdate();

  if (!(sourceacp->canchangesecretly()))
    {
      // source will tell us if it changed.
      return;
    }

  // source might not tell us
  if (sourceacp->getable())
    {
      // okay, we can just request the latest value

      //      cout << "... trying to get value from "+sourceacp->getname()+".\n";
      CfpVal val=sourceacp->get(fromindex);
      //      cout << "... back from trying to get value from "+sourceacp->getname()+".\n";

      if (val.isdefined())
	{
	  //	  cout << "... Got a setable value.\n";
	  acp->set(&val,toindex);
	  //	  cout << "... Using latestupdate to send "+val.get_string()+"from " + sourceacp->getname()+" to "+acp->getname()+".\n";
	  return;
	}
      
    }

  // if we are here, we might want to issue a warning, though its not nesc. a bad thing
}






//--------------------------------------------------------------------------------
//---- CfpFuncAccessor -----------------------------------------------------------
//--------------------------------------------------------------------------------

CfpFuncAccessor::CfpFuncAccessor(string inname,string argstring,CfpClass *inclassp)
  : CfpAccessor(inname+":"+argstring)
{
  // constructor
}




//--------------------------------------------------------------------------------
//---- CfpFuncAccessor -----------------------------------------------------------
//--------------------------------------------------------------------------------


string CfpFuncAccessor::getargstring()
{
  // simple accessor
  return memberp->getargstring();
}


//--------------------------------------------------------------------------------
//---- CfpFuncAccessor::invoke() -------------------------------------------------
//--------------------------------------------------------------------------------

CfpVal CfpFuncAccessor::invoke(map<string,CfpVal*> *argmap)
{
  // invoke the function
  CfpVal val;

  // it should be an error to invoke the virtual function.
  return val;
}


















//--------------------------------------------------------------------------------
//---- CfpObjFuncAcc -------------------------------------------------------------
//--------------------------------------------------------------------------------

CfpObjFuncAcc::CfpObjFuncAcc(string inname,CfpObject *inobjp,string argstring,CfpClass *inclassp,string inkey)
  :CfpFuncAccessor(inname,argstring,inclassp)
{
  // construct the objfuncaccessor
  objp=inobjp;
  keyword=inkey;
  memberp=inclassp->lookup_memberfunc(inname,argstring,true);
}


//--------------------------------------------------------------------------------
//---- CfpObjFuncAcc::invoke() ---------------------------------------------------
//--------------------------------------------------------------------------------

//--------------------------------------------------------------------------------
//---- CfpObjFuncAcc -------------------------------------------------------------
//--------------------------------------------------------------------------------

CfpVal CfpObjFuncAcc::invoke(map<string,CfpVal*> *argmap)
{
  // invoke the object-function accessor
  CfpVal val;

  // invoke the objects function handler (which is what makes this an objectfunction accessor)
  val=objp->invokef(getname(),getargstring(),keyword,argmap);

  return val;
}











// NOTE: I could not get egcs to accept these out-of-line function definitions, so
//  i had to move them all inline.  ugly i know.












//--------------------------------------------------------------------------------
//---- CfpVAccTriggerProxy
//--------------------------------------------------------------------------------


CfpVAccTriggerProxy::CfpVAccTriggerProxy(string name)
{
  // create the proxytrigger accessor which others will use to activate us
  proxytrigger=new CfpVFAMsetval<CfpVAccTriggerProxy,CfpVal>(name,NULL,this,&activate);
}

CfpVAccTriggerProxy::~CfpVAccTriggerProxy()
{
  // delete the list of accessors

  list<CfpTriggerP *>::iterator li=accessors.begin();
  list<CfpTriggerP *>::iterator lend=accessors.end();
  while (li!=lend)
    {
      delete (*li);
      li++;
    }

  // delete our trigger accessor
  delete proxytrigger;
}


void CfpVAccTriggerProxy::addaccessor(CfpVAccessor *accp,ProxyTriggerType inptype,int inindex=-1)
{
  // just add this guy to our list
  CfpTriggerP *tp=new CfpTriggerP(accp,inptype,inindex);
  accessors.push_back(tp);

  //  printf("Proxy added an accessor.\n");
}


void CfpVAccTriggerProxy::activate(CfpVal val)
{
  // we have been activated, so we step through our list of accessor and update them or fire
  // them as the case may be.
  CfpVAccessor *acp;
  int index;

  //  printf("... Proxy was activated.\n");

  list<CfpTriggerP *>::iterator li=accessors.begin();
  list<CfpTriggerP *>::iterator lend=accessors.end();
  while (li!=lend)
    {
      acp=(*li)->getaccessorp();
      index=(*li)->getindex();
      if ((*li)->isinput())
	{
	  // this is a variable that we need to make sure is up to date
	  //	  cout << "... making sure that '"+acp->getname()+"' is uptodate.\n";
	  acp->latestupdate();
	}
      else if ((*li)->isoutput())
	{
	  // this is a variable that we may have modified during our proxy firing, so we want to make sure that
	  // it tells everyone about its changes
	  //	  cout << "... proxy sending output from '"+acp->getname()+"'.\n";
	  acp->retrieveandset();
	}
      else
	{
	  // otherwise it's a trigger that we want to fire;
	  acp->set(val,index);
	}
      li++;
    }
  //  printf("... Leaving proxy activate.\n");
}















//}
// Namespace close-brace





