// axis.h
// This file defines some axis classes, both for plots and for videocameras

// forward delcarations
class Axis;
class VisibleAxis;

#ifndef axish_defined
#define axish_defined

#include <list>
#include <vector>
#include <string>




//-------------------------------------------------------------------------------
//-- Enumerations
//-------------------------------------------------------------------------------

typedef enum AxisAdjustMode
{
  noadjust,
  hardshift_mode,
  softshift_mode,
  expand_mode,
  expandsymetric_mode,
  expandplus_mode,
  expandsymetricplus_mode,
  fit_mode,
  fitplus_mode
};

typedef enum AxisScaleType
{
  linear_scale,
  log_scale
};

typedef enum TickType
{
  no_tick,
  line_tick,
  plane_tick,
  box_tick
};

typedef enum AxisLabelpos
{
  no_label,
  start_pos,
  middle_pos,
  intercept_pos,
  end_pos
};

typedef enum WallType
{
  no_wall,
  solid_wall,
  wire_wall
};







//-------------------------------------------------------------------------------
//-- Axis
//-------------------------------------------------------------------------------
//: Base Axis class

class Axis
{
protected:
  string name;								// name of axis
  string label;								// label on axis
  string enums;								// sublabels
  int discretes;							// how many bins to divy for continuousvars
  float minshow,maxshow;						// range
  bool firstval;							// is this the first val we've seen?
  AxisAdjustMode adjustmode;						// how to adjust
  AxisLabelpos labelpos;
public:
  Axis(string inname,string inlabel,string inenums);
  ~Axis() {;};
public:
  void setshifter() {adjustmode=hardshift_mode;};
  bool adjustout(float val);						// adjust range and ret true if changed
  int discreteind(float val);						// get discrete index of val
  float reverseind(int discind);					// get inverse of discrete index discind
public:
  void setrange(float min,float max) {minshow=min; maxshow=max;firstval=false;change();};	// set range
  void setrange(float minmax) {minshow=minmax; maxshow=minmax;firstval=false;change();};	// set range to singular val
public:
  string getname() {return name;};					// accessor
  string getlabel() {return label;};					// accessor
  int getdiscretes() {return discretes;};				// accessor
  void setdiscretes(int indiscretes);					// accessor
  float getmin(){return minshow;};
  float getmax(){return maxshow;};
public:
  virtual void change() {;};
};






//-------------------------------------------------------------------------------
//-- VisibleAxis
//-------------------------------------------------------------------------------
//: Derived class for graphical X,Y, and Z videocamera axes

class VisibleAxis: public Axis
{
protected:
  float minorticks,majorticks;						// how often to tick
  float spacing;							// spacing btwm datpoints (0 to 1)
  float scale;								// relative scale of axis (all 1s is cube)
  float intercept;							// intercept origin
  bool autoticks;							// automatically guess ticks
  WallType minwall,intwall,maxwall;				        // show a wall along axis?
  bool bitfonts;							// use bitfonts rather than scaled&stroked?
  bool endtick_number;						        // should endpoints be visibly numbered
  bool connect;								// should we connect neighbors (make lines)
  bool minimize_intercept;
  bool stretchsymbols;
  bool show;
protected:
  bool needsredraw;							// axes need to be redrawn
protected:
  AxisScaleType scaletype;						// linear or log
  TickType minortick_type;						// type of ticks to draw on minor points
  TickType majortick_type;						// type of ticks to draw on major points
public:
  VisibleAxis(string inname,string inlabel,string inenums);
public:
  float getminorticks(){return minorticks;};
  float getmajorticks(){return majorticks;};
  bool getendtick_number(){return endtick_number;};
  float getspacing() {return spacing;};					// accessor
  float getscale() {return scale;};					// accessor
  void setscale(float inscale) {scale=inscale;};			// accessor
  void setticks(float mjticks, float miticks=0);			// accessor
  void getaxispoints(float &sloc,float &iloc,float &eloc);		// accessor
  WallType getminwall(){return minwall;};
  WallType getmaxwall(){return maxwall;};
  WallType getintwall(){return intwall;};
  bool getstretch() {return stretchsymbols;};
  bool getshow(){return show;};
  void setshow(bool inshow){show=inshow;change();};
  void setmaxwall(int inmaxwall){maxwall=(WallType)inmaxwall;change();};
  void setspacing(float inspacing){spacing=inspacing;change();};
  float getintercept(){return intercept;};
public:
  float symbolsize();
  float symbolwidth();
  bool minint(){return minimize_intercept;};
public:
  void autotick();							// guess ticks from discretes
public:
  void change();
};

















#endif




