// cfpclass.C

#include "cfpinc.h"

//Not implemented in gcc yet (2.8.1)
//using namespace std;
//namespace Configpp{






//--------------------------------------------------------------------------------
//---- CfpClass::CfpClass() ------------------------------------------------------
//--------------------------------------------------------------------------------

CfpClass::CfpClass(CfpFactory *infactp,string inname,string inpackagename,CfpObject *inspawnerp)
{
  // simple constructor (note: spawnerp may be null)
  factp=infactp;
  name=inname;
  packagename=inpackagename;
  spawnerp=inspawnerp;

  if (inspawnerp==NULL)
    {
      // this happens when a CfpObject gets created manually rather than through the parser
      //  if we tried to perform the 'addisa' routine we could run into trouble from
      //  missing information/pointers.
      return;
    }

  // add the base class type
  addisa(inname);
}



//--------------------------------------------------------------------------------
//---- CfpClass::makenew() -------------------------------------------------------
//--------------------------------------------------------------------------------

CfpObject *CfpClass::makenew(string objname)
{
  // make a new object by calling the spawner and asking it to do it
  if (!spawnerp)
    throw CfpError("object class '" +name+ "' has no registered spawner.","",internal,true,false);

  return (spawnerp->makenew(this,objname));
}



//--------------------------------------------------------------------------------
//---- CfpClass::lookupunit() ----------------------------------------------------
//--------------------------------------------------------------------------------

CfpUnit *CfpClass::lookupUnit(string unitstring,bool throwerr)
 {
   // just call the factory version of this function
   return factp->lookupUnit(unitstring,throwerr);
 }


//--------------------------------------------------------------------------------
//---- CfpClass::lookuptype() ----------------------------------------------------
//--------------------------------------------------------------------------------

CfpType *CfpClass::lookupType(string typestring,bool throwerr)
 {
   // just call the factory version of this function
   return factp->lookupType(typestring,throwerr);
 }




//--------------------------------------------------------------------------------
//---- CfpClass::lookup_member() -------------------------------------------------
//--------------------------------------------------------------------------------

CfpMember *CfpClass::lookup_member(string inname,bool throwerr)
{
  // find the registered member and return a pointer to it
  CfpMember *memberp=NULL;
  map<string, CfpMemberVar *>::iterator pv;
  map<string, CfpMemberFunc *>::iterator pf;
  
  pv=varmanager.find(inname);
  if (pv!=varmanager.end())
    memberp=pv->second;
  else
    {
      pv=argmanager.find(inname);
      if (pv!=argmanager.end())
	memberp=pv->second;
      else
	{
	  // we couldn't find the member var or member arg
	  pf=funcmanager.find(inname);
	  if (pf!=funcmanager.end())
	    memberp=pf->second;
	  else
	    {
	      // we couldn't find the member func either
	      if (throwerr)
		throw CfpError("object member '" +inname+ "' of object class '" +name+ "' is not known.","",classerr,false,false);
	    }
	}
    }
  return memberp;
}





//--------------------------------------------------------------------------------
//---- CfpClass::lookup_memberva() -----------------------------------------------
//--------------------------------------------------------------------------------

CfpMemberVar *CfpClass::lookup_memberva(string inname,bool throwerr)
{
  // find the registered member and return a pointer to it
  CfpMemberVar *memberp=NULL;
  map<string, CfpMemberVar *>::iterator pv;

  pv=varmanager.find(inname);

  if (pv!=varmanager.end())
    memberp=pv->second;
  else
    {
      pv=argmanager.find(inname);
      if (pv!=argmanager.end())
	memberp=pv->second;
      else
	{
	  // we couldn't find the member var or arg either
	  if (throwerr)
	    {
	      throw CfpError("object member var/arg '" +inname+ "' of object class '" +name+ "' is not known.","",classerr,false,false);
	    }
	}
    }
  return memberp;
}


//--------------------------------------------------------------------------------
//---- CfpClass::lookup_membervar() ----------------------------------------------
//--------------------------------------------------------------------------------

CfpMemberVar *CfpClass::lookup_membervar(string inname,bool throwerr)
{
  // find the registered member var and return a pointer to it
  CfpMemberVar *memberp=NULL;
  map<string, CfpMemberVar *>::iterator p;
  
  p=varmanager.find(inname);
  if (p!=varmanager.end())
    memberp=p->second;
  else
    {
      // we couldn't find the member var
      if (throwerr)
	throw CfpError("object member variable '" +inname+ "' of object class '" +name+ "' is not known.","",classerr,true,false);
    }
  return memberp;
}




//--------------------------------------------------------------------------------
//---- CfpClass::lookup_memberfunc() ---------------------------------------------
//--------------------------------------------------------------------------------

CfpMemberFunc *CfpClass::lookup_memberfunc(string inname,string argstring,bool throwerr)
{
  // find the registered member var and return a pointer to it
  CfpMemberFunc *memberp=NULL;
  map<string, CfpMemberFunc *>::iterator pf;
  
  pf=funcmanager.find(inname);
  if (pf!=funcmanager.end())
    memberp=pf->second;
  else
    {
      // we couldn't find the member func
      if (throwerr)
	throw CfpError("object member function '" +inname+ "' of object class '" +name+ "' is not known.","",classerr,true,false);
    }

  return memberp;
}





//--------------------------------------------------------------------------------
//---- CfpClass::addisa() --------------------------------------------------------
//--------------------------------------------------------------------------------

void CfpClass::addisa(string newtypestr)
{
  // we want to add the list of types that we are.
  // we want to allow that typestr can be a space or comma separated list of types
  // class types may be simple names (like Robotclass) or have subtypes (like Robotclass.droidsubclass)
  // we add each leftmost as an isa (it Robotclass.droidclass would add Robotclass and Robotclass.droidclass)
  string word;

  while (newtypestr!="")
    {
      word=JRstripword(newtypestr);
      if (word!=name)
	(getcfpp())->debug("noting that class "+name+" is derived from class '"+word+"'.");
      while (word!="")
	{
	  isalist.insert(word);
	  word=JRstripleader(word);
	}
    }
}



//--------------------------------------------------------------------------------
//---- CfpClass::isa() -----------------------------------------------------------
//--------------------------------------------------------------------------------

bool CfpClass::isa(string typestr)
{
  // is class a typestr?
  // we want to allow that typenam can be a space or comma separated list of types
  set<string>::iterator p;
  string word;

  while (typestr!="")
    {
      word=JRstripword(typestr);
      if (isalist.count(word))
	return true;
    }

  return false;
}








//--------------------------------------------------------------------------------
//---- CfpClass::addfunc() -------------------------------------------------------
//--------------------------------------------------------------------------------

CfpMemberFunc *CfpClass::addfunc(string inname,string argstring,string retstring,string inlabel,string indesc)
{
  // create the func described and add it to the class
  CfpMemberFunc *funcp=NULL;

  funcp=new CfpMemberFunc(this,inname,inlabel,indesc,argstring,retstring);
  funcmanager.insert(make_pair(inname,funcp));

  return funcp;
}




//--------------------------------------------------------------------------------
//---- CfpClass::addclassfunc() --------------------------------------------------
//--------------------------------------------------------------------------------

CfpMemberFunc *CfpClass::addclassfunc(string inname,string argstring,string retstring,string inlabel,string indesc)
{
  // create the CLASS func (a static function not attached to a particular instance of
  // the class.
  CfpMemberFunc *funcp=NULL;

  return funcp;
}





//--------------------------------------------------------------------------------
//---- CfpClass::addarg() --------------------------------------------------------
//--------------------------------------------------------------------------------

CfpMemberVar *CfpClass::addarg(string inname,CfpVal ininitval,bool insscope,string inlabel,string unitstr,string typestr,string inrangestr,string indesc)
{
  // create the arg described and add it to the class
  CfpMemberVar *varp=NULL;

  if (argmanager.find(inname)!=argmanager.end())
    throw CfpError("attempt to add argument '"+inname+"' to class '"+getname()+"' when an arg of that name has already been added.","",classerr,true,false);

  varp=new CfpMemberVar(this,inname,ininitval,insscope,inlabel,unitstr,typestr,inrangestr);
  argmanager.insert(make_pair(inname,varp));

  return varp;
}


//--------------------------------------------------------------------------------
//---- CfpClass::addvarg() --------------------------------------------------------
//--------------------------------------------------------------------------------

CfpMemberVar *CfpClass::addvarg(string inname,int vecsize,CfpVal ininitval,bool insscope,string inlabel,string unitstr,string typestr,string inrangestr,string indesc)
{
  // create the VECTOR arg described and add it to the class
  CfpMemberVar *varp=NULL;

  if (argmanager.find(inname)!=argmanager.end())
    throw CfpError("attempt to add arguement '"+inname+"' to class '"+getname()+"' when an arg of that name has already been added.","",classerr,true,false);

  varp=new CfpMemberVar(this,vecsize,inname,ininitval,insscope,inlabel,unitstr,typestr,inrangestr);
  argmanager.insert(make_pair(inname,varp));

  return varp;
}





//--------------------------------------------------------------------------------
//---- CfpClass::addvar() --------------------------------------------------------
//--------------------------------------------------------------------------------

CfpMemberVar *CfpClass::addvar(string inname,CfpVal ininitval,bool insscope,string inlabel,string unitstr,string typestr,string inrangestr,string indesc)
{
  // create the var described and add it to the class
  CfpMemberVar *varp=NULL;

  if (varmanager.find(inname)!=varmanager.end())
    throw CfpError("attempt to add variable '"+inname+"' to class '"+getname()+"' when a var of that name has already been added.","",classerr,true,false);

  varp=new CfpMemberVar(this,inname,ininitval,insscope,inlabel,unitstr,typestr,inrangestr);
  varmanager.insert(make_pair(inname,varp));

  return varp;
}


//--------------------------------------------------------------------------------
//---- CfpClass::addvvar() --------------------------------------------------------
//--------------------------------------------------------------------------------

CfpMemberVar *CfpClass::addvvar(string inname,int vecsize,CfpVal ininitval,bool insscope,string inlabel,string unitstr,string typestr,string inrangestr,string indesc)
{
  // create the var described and add it to the class
  CfpMemberVar *varp=NULL;

  if (varmanager.find(inname)!=varmanager.end())
    throw CfpError("attempt to add variable '"+inname+"' to class '"+getname()+"' when a var of that name has already been added.","",classerr,true,false);

  varp=new CfpMemberVar(this,vecsize,inname,ininitval,insscope,inlabel,unitstr,typestr,inrangestr);
  varmanager.insert(make_pair(inname,varp));

  return varp;
}





//--------------------------------------------------------------------------------
//---- CfpClass::addconstructor() ------------------------------------------------
//--------------------------------------------------------------------------------

CfpMemberFunc *CfpClass::addconstructor(string argstring,string inkeyword,string inlabel,string indesc)
{
  // create the constructor func described and add it to the class
  CfpMemberFunc *funcp=NULL;
  string inname;

  inname="_constructor("+inkeyword+")_";

  funcp=new CfpMemberFunc(this,inname,inlabel,indesc,argstring,"");
  funcmanager.insert(make_pair(inname,funcp));

  return funcp;
}







//--------------------------------------------------------------------------------
//---- CfpClass::namehint() ------------------------------------------------------
//--------------------------------------------------------------------------------

string CfpClass::namehint(string inname)
{
  // display a hint about name, by searching for near matches in the var,arg,func list
  CfpMemberVar *membervp=NULL;
  CfpMemberFunc *memberfp=NULL;
  map<string, CfpMemberVar *>::iterator vp;
  map<string, CfpMemberFunc *>::iterator fp;
  string hintstring="";

  // first check member args
  vp=argmanager.begin();
  while (vp!=argmanager.end())
    {
      membervp=vp->second;
      if (JRsimiliarstr(inname,membervp->getname()))
	hintstring+="         argument '"+membervp->getname()+"' of class '"+name+"' package (" +getpackagename()+").\n";
      vp++;
    }

  // now check member vars
  vp=varmanager.begin();
  while (vp!=varmanager.end())
    {
      membervp=vp->second;
      if (JRsimiliarstr(inname,membervp->getname()))
	hintstring+="         variable '"+membervp->getname()+"' of class '"+name+"' package (" +getpackagename()+").\n";
      vp++;
    }

  // now check membar funcs
  fp=funcmanager.begin();
  while (fp!=funcmanager.end())
    {
      memberfp=fp->second;
      if (JRsimiliarstr(inname,memberfp->getname()))
	hintstring+="         function '"+memberfp->getname()+"("+memberfp->getargstring()+")' of class '"+name+"' package (" +getpackagename()+").\n";
      fp++;
    }

  return hintstring;
}







//}
// Namespace close-brace
