// cfperror.C

#include "cfpinc.h"
#include "JRutils.h"

//Not implemented in gcc yet (2.8.1)
//using namespace std;
//namespace Configpp{







//--------------------------------------------------------------------------------
//---- CfpLoc::CfpLoc() ----------------------------------------------------------
//--------------------------------------------------------------------------------

CfpLoc::CfpLoc()
{
  // A class for keeping track of where a token is in a stream, useful for error reporting.
  sisp=NULL;
  startline=0;
  startlinepos=0;
  eisp=NULL;
  endline=0;
  endlinepos=0;
  eoflag=false;
  istreamp=NULL;
  filename="";
}

void CfpLoc::startset(CfpInputSource *insisp,long insl,long inslp)
{
  // set the start loc
  sisp=insisp;
  startline=insl;
  startlinepos=inslp;
}

void CfpLoc::endset(CfpInputSource *ineisp,long inel,long inelp)
{
  // set the start loc
  eisp=ineisp;
  endline=inel;
  endlinepos=inelp;
}







void CfpLoc::show()
{
  // point to the location in the stream
  string sname,ename;
  bool sreclose=false;
  bool ereclose=false;
  long eline;
  istream *sreadstreamp,*ereadstreamp;

  // make sure there is a source associated with this token
  if (!sisp && !istreamp)
    return;

  sname="";ename="";
  if (sisp)
    {
      // reopen the stream (istreamp below) if it is closed
      sreclose=sisp->reopenclosed();
      sname=(sisp->gettypename()) + " " + (sisp->getname());
      
      // if there is a dif. endstream, get its info
      if (eisp)
	{
	  ename=(eisp->gettypename()) + " " + (eisp->getname());
	  ereclose=eisp->reopenclosed();
	}
      sreadstreamp=sisp->getstreamp();
      ereadstreamp=eisp->getstreamp();
    }
  else
    {
      // use explicit stream and name
      sname=filename;
      sreadstreamp=istreamp;
      ereadstreamp=sreadstreamp;
    }
  
  // if the location ends at eof, we tell the display func by setting endline to -1
  if (eoflag)
    eline=-1;
  else
    eline=endline;

  // now graphically point out where it is using a utility function which
  // finds the closest line and character in the file and displays the line,
  // line #, and puts carrots under the error location
  JRShowStreamPos(&cerr,sname,sreadstreamp,startline,startlinepos,ename,ereadstreamp,eline,endlinepos);

  // reclose if the file if it was closed
  if (sreclose)
    sisp->close();
  if (ereclose)
    eisp->close();
}




//--------------------------------------------------------------------------------
//---- CfpLoc::startcopy() -------------------------------------------------------
//--------------------------------------------------------------------------------

void CfpLoc::startcopy(CfpLoc *copyloc)
  {
    // copy the start info from another loc
    startline=copyloc->getstartline();
    startlinepos=copyloc->getstartlinepos();
    sisp=copyloc->getssourcep();
  }

//--------------------------------------------------------------------------------
//---- CfpLoc::endcopy() ---------------------------------------------------------
//--------------------------------------------------------------------------------

void CfpLoc::endcopy(CfpLoc *copyloc)
  {
    // copy the end info from another loc
    endline=copyloc->getendline();
    endlinepos=copyloc->getendlinepos();
    eisp=copyloc->getesourcep();
  }

//--------------------------------------------------------------------------------
//---- CfpLoc::fileposet()
//--------------------------------------------------------------------------------

void CfpLoc::fileposet(string infilename,istream *instreamp)
  {
    // alternative setting of location
    long fpos,lfpos,opos;
    int rowcount;

    filename=infilename;
    istreamp=instreamp;

    // we need to convert raw position to line num and offset
    opos=instreamp->tellg();
    istreamp->seekg(0,ios::beg);
    lfpos=0;
    for (rowcount=0;;)
      {
	// consume a line
	istreamp->ignore(1000,'\n');
	fpos=istreamp->tellg();
	if (fpos>opos)
	  break;
	if (istreamp->eof())
	  break;
	lfpos=fpos;
	++rowcount;
      }
    startline=endline=rowcount;
    startlinepos=endlinepos=opos-lfpos;

    // return file to orig pos.
    istreamp->seekg(opos,ios::beg);
  }







//--------------------------------------------------------------------------------
//---- CfpError::CfpError() ------------------------------------------------------
//--------------------------------------------------------------------------------

CfpError::CfpError()
{
  // form a default call to init
  CfpError("error","",undefined,true,false,NULL);
}

CfpError::CfpError(string indesc,string inhint,CfpErrorCode inerrorcode,bool infatalflag, Configpp *configppp)
{
  // form the call to init by assuming the error location is just the location of the last parsed token.
  CfpLoc *locp=NULL;
  if (configppp)
    locp=((configppp->getlexerp())->getcurtokenp())->getlocp();
  init(indesc,inhint,inerrorcode,infatalflag,false,locp);
}

CfpError::CfpError(string indesc,string inhint,CfpErrorCode inerrorcode,bool infatalflag, bool inexactloc,CfpLoc *incloselocp)
{
  // form the call to init
  init (indesc,inhint,inerrorcode,infatalflag,inexactloc,incloselocp);
}

CfpError::CfpError(string indesc,string inhint,CfpErrorCode inerrorcode,bool infatalflag,string fname,istream *istreamp)
{
  // create a CfpLoc item which points to the correct file location
  CfpLoc eloc;
  eloc.fileposet(fname,istreamp);
  init (indesc,inhint,inerrorcode,infatalflag,true,&eloc);
}



//--------------------------------------------------------------------------------
//---- CfpError::init() ----------------------------------------------------------
//--------------------------------------------------------------------------------

void CfpError::init(string indesc,string inhint,CfpErrorCode inerrorcode,bool infatalflag, bool inexactloc,CfpLoc *incloselocp)
{
  // create a new error entry which will be thrown
  // this does the real constructor work
  description=indesc;
  hint=inhint;
  errorcode=inerrorcode;
  fatalflag=infatalflag;
  reported=false;
  exactloc=inexactloc;
  if (incloselocp!=NULL)
    closeloc=*incloselocp;
}





//--------------------------------------------------------------------------------
//---- CfpError::report() --------------------------------------------------------
//--------------------------------------------------------------------------------

void CfpError::report(class Configpp *cfpfp)
{
  // Report an error 
  CfpLoc eloc;
  string ecodename;

  if (reported)
    return;

  // set the reported flag
  reported=true;

  // get the error code
  ecodename=getcodename();
  if (fatalflag)
    ecodename="Fatal "+ecodename;

  // now highlight it in a pretty way
  //  ecodename="\033[;33;1m"+ecodename+" error"+BOLD_RIGHT;
  ecodename=BOLD_LEFT+ecodename+" error"+BOLD_RIGHT;

  // display error description
  cerr << "-*-> " << ecodename << ": " << description << "\n";

  // if it's a complaint about too many errors, then there's no need to show where
  if (errorcode==toomanyerr)
    return;

  // First, if the creator of the error didn't tell us where the error was from,
  //  use the last one parsed, but warn the user the location is a guess.
  if (closeloc.getstartline()==0 || !exactloc)
    {
      closeloc=*((cfpfp->getcurrentokenp())->getlocp());
      if (closeloc.getssourcep())
	cfpfp->warning("error ocurred somewhere at or around:");
    }

  // Second, if the creator explicitly told us that the error location is approximate,
  // we force the prior error location bound to be Configpp::Hintloc, which is where
  //  we first starting parsing the current statement.
  if (!exactloc)
    closeloc=*(cfpfp->gethintlocp());

  // show the user where the error occurred
  closeloc.show();

  // show any hint
  if (hint.length()>0)
    {
      // there is a hint.  does user want hints?
      if (cfpfp->hintmode())
	{
	  if (hint[hint.length()-1]=='\n')
	    cerr << "     " << hint;
	  else
	    cerr << "     " << hint << "\n";
	}
    }
  cerr << "\n";
}






string CfpError::getcodename()
{
  // return the name of the error type code
  switch (errorcode)
    {
    case undefined:
      return "Undefined";
      break;
    case file:
      return "File";
      break;
    case io:
      return "I/O";
      break;
    case lexical:
      return "Parsing";
      break;
    case syntactical:
      return "Syntax";
      break;
    case internal:
      return "Internal";
      break;
    case library:
      return "Class-library";
      break;
    case sys:
      return "System";
      break;
    case classerr:
      return "Class";
      break;
    case gui:
      return "Graphical-user-interface";
      break;
    case toomanyerr:
      return "Parsing";
      break;
    }

  return "unknown";
}






// }
// Namespace close-brace
