// cfperror.h
// This file defines some error-handling classes.  Specifically, the CfpErrorInfo
//  class, which is used to store and report parsing errors.
// This is also the class that gets thrown on parsing error *exceptions*

#ifndef INC_cfperrorh
#define INC_cfperrorh

#include "cfpinc.h"

//Not implemented in gcc yet (2.8.1)
//using namespace std;
//namespace Configpp{






//--------------------------------------------------------------------------------
//---- CfpLoc --------------------------------------------------------------------
//--------------------------------------------------------------------------------

// CfpLoc is a class used for error reporting.
//  Every lexical token (CfpToken) and parser stack element contains one.
//  It keeps track of the stream and stream location where the item was encountered

class CfpLoc
{
private:
  CfpInputSource *sisp;					// pointer to the start input source
  CfpInputSource *eisp;					// pointer to the end input source
  long startline;
  long startlinepos;
  long endline;
  long endlinepos;
  istream *istreamp;
  bool eoflag;
  string filename;
public:
  CfpLoc();
public:
  void startset(CfpInputSource *insisp,long insl,long inslp);	// set the startpos
  void endset(CfpInputSource *ineisp,long inel,long inelp);	// set the endpos
  void seteoflag(bool ineoflag=true) {eoflag=ineoflag;};	// set the eof flag
  void startcopy(CfpLoc *copyloc);				// copy the start info from another loc
  void endcopy(CfpLoc *copyloc);				// copy the end info from another loc
  void fileposet(string infilename,istream *instreamp);		// alternative set
  bool geteoflag() {return eoflag;};				// accessor
  CfpInputSource *getssourcep() {return sisp;};			// accessor
  CfpInputSource *getesourcep() {return eisp;};			// accessor
  long getstartline() {return startline;};			// accessor
  long getstartlinepos() {return startlinepos;};		// accessor
  long getendline() {return endline;};				// accessor
  long getendlinepos() {return endlinepos;};			// accessor
public:
  void show();							// point to the location in stream
};





//--------------------------------------------------------------------------------
//---- CfpError ------------------------------------------------------------------
//--------------------------------------------------------------------------------

// Manages information about potential errors, and reporting them;

class CfpError
{
private:
  CfpErrorCode errorcode;					// the enumerated error code (see cfpsymbols.h)
  bool fatalflag;						// is the error a superserious fatal one?
  string description;						// verbal description of the error
  string hint;							// verbal description of a hint
  class CfpLoc closeloc;					// closest location we think is responsible
  bool exactloc;						// do we know exact location of error
  bool reported;						// has it been reported yet
  static string termname;					// for color highlighting info, etc.
public:
  CfpError();							// default constructor
  CfpError(string indesc,string inhint,CfpErrorCode inerrorcode,bool infatalflag,Configpp *configppp);
  CfpError(string indesc,string inhint,CfpErrorCode inerrorcode,bool infatalflag,bool inexactloc,CfpLoc *incloselocp=NULL);
  CfpError(string indesc,string inhint,CfpErrorCode inerrorcode,bool infatalflag,string fname,istream *istreamp);
  void init(string indesc,string inhint,CfpErrorCode inerrorcode,bool infatalflag,bool inexactloc,CfpLoc *incloselocp);
public:
  void report(Configpp *cfpfp);					// report the error, showing where it occurred
  bool isfatal() {return fatalflag;};				// is the error so bad we must stop parsing immed.?
  bool wasreported() {return reported;};			// has the error been reported yet?
  bool setfatal(bool inff=true) {fatalflag=inff;};		// set the fatal flag
  string getcodename();						// return the name of the error type code
private:
  static void setterm(string intname){termname=intname;};	// set the terminal type (for color highlights)
  static string lefthighlight() {;};				// return ascii code to start highlighting
  static string righthighlight() {;};				// return ascii code to end highlighting
  static string leftbold() {;};					// return ascii code to start highlighting
  static string rightbold() {;};				// return ascii code to end highlighting
};






// }
// Namespace close-brace
#endif
