// cfpexpr.C
// This file (along with configpp.C, cfpcomline.C, cfpparser.C) implements member fucntions for the Configpp class.
// It contains functions directly related to parsing and evaluating expressions

#include "cfpinc.h"

//Not implemented in gcc yet (2.8.1)
//using namespace std;
//namespace Configpp{








//--------------------------------------------------------------------------------
//---- Configpp::Parsefunc() ------------------------------------------------------
//--------------------------------------------------------------------------------

CfpVal Configpp::Parsefunc(string funcname)
{
  // parse a function call
  // the open paren will have been consumed already.
  // First we read expressions separated by commas, until we hit the close paren.
  // Then we figure out which object and version of the function we should invoke (if there is overloading)
  //  and invoke it.  Of course we throw an error if none match.
  list<CfpVal> argvals;
  CfpToken toke;
  CfpVal val;
  CfpVal retval;
  CfpObject *objp;
  string leaderstr;
  CfpFuncAccessor *funcp=NULL;

  // first, make sure the next symbol is an open paren
  toke=lexerp->readtoken();
  if (toke.gettype()!=openparenSym)
    throw CfpError("expected an open parentheses to begin a function call.","",internal,true,true);


  // first check that there are any args to read at all
  toke=lexerp->peektoken();
  if (toke.gettype()==closeparenSym)
    lexerp->readtoken();
  else
    {
      for (;;)
	{
	  // there is another argument (expression) to read
	  val=Parsevalexp();
	  // add the value to the arg list
	  argvals.push_back(val);
	  // now we need to consume the separator
	  toke=lexerp->readtoken();
	  if (toke.gettype()==closeparenSym)
	    {
	      // okay we got them all
	      break;
	    }
	  // make sure that it was a valid argument separator
	  if (toke.gettype()!=commaSym)
	    throw CfpError("unexpected symbol while reading function arguments.","",lexical,false,true);
	}
    }

  // okay we got all the arguments.  now we have to figure out what object function is being called

  // if they specify an object trial, look up the object
  leaderstr=JRstripleader(funcname);
  if (leaderstr!="")
    {
      // okay we have an explicit trail
      objp=lookupObjtrail(leaderstr,true);
    }

  // okay, now get the accessor for the appropriate function call
  // JESSE: this has to be modified to handle overloading functions
  if (objp)
    funcp=objp->lookupfunc(funcname,false);
  else
    funcp=lookupFunction(funcname,false);

  // form the arg list to pass
  // now call the function with the arg list
  //	retval=funcp->invokef();

  // lastly, return the return val
  return retval;
}






//--------------------------------------------------------------------------------
//---- Configpp::Parsevalexp() ----------------------------------------------------
//--------------------------------------------------------------------------------

CfpVal Configpp::Parsevalexp()
{
  // parse a value expression (ie a numeric expression like "a+b*(c+d)")
  // and convert it into the type and units of varp.
  // Stop on parsing an uninterpretable symbol (ie a comma, etc.)
  CfpVal expval;
  CfpToken toke;

  expval=Parsevalue();

  return expval;
}



//--------------------------------------------------------------------------------
//---- Configpp::Parsevalue() ----------------------------------------------------
//--------------------------------------------------------------------------------

CfpVal Configpp::Parsevalue()
{
  // parse a single value expression
  CfpVal val;
  CfpToken toke;
  string stringval="";
  CfpVAccessor *varp;
  CfpObject *objp;
  int depth=0;
  int index;

  toke=lexerp->readtoken(tokenMode);
  stringval=toke.getstr();
  switch(toke.gettype())
    {
    case stringVal:
      // a quoted string constant
      JRremquotes(stringval);
      val.set_string(stringval);
      break;
    case numericVal:
      // simple number value
      val.set_number(stringval);
      break;
    case alphaVal:
      // an object or object memember (var or func) name, or possibly even classname...
      // is it a member var, etc.?
      varp=lookupVAcc(stringval,false);
      if (varp)
	{
	  // okay we found a var/arg accessor
	  val.set_vaccesspointer(varp);
	  // now we should peek for a vector address
	  toke=lexerp->peektoken(tokenMode);
	  if (toke.gettype()==opensquareSym)
	    {
	      // okay we have a vector address specifier
	      // grab the modifier
	      depth=0;
	      stringval="";
	      for (;;)
		{
		  toke=lexerp->readtoken(tokenMode);
		  stringval.append(" "+(toke.getstr()));
		  if (toke.gettype()==opensquareSym)
		    ++depth;
		  if (toke.gettype()==closesquareSym)
		    {
		      --depth;
		      if (depth==0)
			break;
		    }
		}
	      // okay, now ask the accessor to tell us the index it corresponds to
	      index=varp->parseindex(stringval);
	      // now set the CfpVal accessor offset
	      //	      cout << "READ A VECTOR ADDRESS MODIFIER pointing to element " << JRtostr(index)<<" !\n";
	      val.set_vaccessindex(index);
	    }
	  break;
	}
      // is it an object name?
      objp=lookupObjtrail(stringval,false);
      if (objp)
	{
	  // okay we found an object pointer
	  val.set_objectpointer(objp);
	  break;
	}

      // unknown
      throw CfpError("expected to parse an expression value, but got '"+stringval+"' instead.","",syntactical,false,this);
      break;
    case opensquareSym:
      // the value is a vector value, which we treat like a string, in that we return it all as one big string
      //      cout << "GOT AN OPENSQAURE ON RIGHT SIDE OF EXPRESSION\n";
      // just grab all tokens until the close square
      depth=1;
      for (;;)
	{
	  toke=lexerp->readtoken(tokenMode);
	  stringval.append(" "+(toke.getstr()));
	  if (toke.gettype()==opensquareSym)
	    ++depth;
	  if (toke.gettype()==closesquareSym)
	    {
	      --depth;
	      if (depth==0)
		break;
	    }
	}
      // return the contents in a string
      val.set_string(stringval);
      break;
    default:
      // anything else is an error
      throw CfpError("expected to parse an expression value, but got '"+stringval+"' instead.","",syntactical,false,this);
      break;
    }

  // set the location of val (for error reporting)
  val.setloc(*(toke.getlocp()));

  return val;
}














//}
// Namespace close-brace
