// cfplexer.C

#include "cfpinc.h"

//Not implemented in gcc yet (2.8.1)
//using namespace std;
//namespace Configpp{





//--------------------------------------------------------------------------------
//---- CfpInputSource::CfpInputSource() ------------------------------------------
//--------------------------------------------------------------------------------

CfpInputSource::CfpInputSource()
{
  // base constructor, just reset pointers
  configppp=NULL;
  inputstreamp=NULL;
  streamname="uninitialized";

  positionreset();
}



//--------------------------------------------------------------------------------
//---- CfpInputSource::~CfpInputSource() -----------------------------------------
//--------------------------------------------------------------------------------

CfpInputSource::~CfpInputSource()
{
  // delete the actual stream object
}



//--------------------------------------------------------------------------------
//---- CfpInputSource::reopenclosed() --------------------------------------------
//--------------------------------------------------------------------------------

bool CfpInputSource::reopenclosed()
  {
    // reopen the stream if it has been closed

    if (!inputstreamp)
      {
	openstream();
	return true;
      }
    else
      return false;
  }




//--------------------------------------------------------------------------------
//---- CfpInputSource::positionreset() -------------------------------------------
//--------------------------------------------------------------------------------

void CfpInputSource::positionreset()
{
  // reset position info
  position=0;
  curline=1;
  curlinepos=1;
}




//--------------------------------------------------------------------------------
//---- CfpInputSource::readchar() ------------------------------------------------
//--------------------------------------------------------------------------------

int CfpInputSource::readchar()
{
  // get a char from the stream, and advance the counter
  int c;

  // are we already at eof?
  if (inputstreamp->eof())
    return EOF;

  // read the character and advance the position counter
  c=inputstreamp->get();

  // increment file position
  ++position;
  ++curlinepos;
  if (c=='\n')
    {
      // read a CR
      ++curline;
      curlinepos=1;
    }

  return c;
}




//--------------------------------------------------------------------------------
//---- CfpInputSource::prekchar() ------------------------------------------------
//--------------------------------------------------------------------------------

int CfpInputSource::peekchar()
{
  // peek a char from the stream, don;t increase pointer, etc.
  // WARNING: i'm not sure this doesn't reset the eof flags if it peeks eof.
  return inputstreamp->peek();
}

int CfpSourceFile::peekchar()
{
  // peek a char from the stream, don;t increase pointer, etc.
  // WARNING: i'm not sure this doesn't reset the eof flags if it peeks eof.
  int c;

  c=inputstreamp->peek();
  if (c==EOF)
    {
      // WIERD: if we hit eof, the error flags may get set during the peek, so we clear them
      inputstreamp->clear();
    }
  return c;
}




//--------------------------------------------------------------------------------
//---- CfpSourceFile::CfpSourceFile() --------------------------------------------
//--------------------------------------------------------------------------------

CfpSourceFile::CfpSourceFile(Configpp *inconfigppp, string filename, const string ext)
  : CfpInputSource()
{
  // create a new input stream from the file
  configppp=inconfigppp;
  streamname=filename;
  extname=ext;

  // now open it
  openstream();
}



//--------------------------------------------------------------------------------
//---- CfpSourceFile::openstream() -----------------------------------------------
//--------------------------------------------------------------------------------

void CfpSourceFile::openstream()
{
  // do the work of opening the stream
  // try to open the file, it will throw an error if it fails (so we don't have to detect)
  inputstreamp=configppp->inFilestream(streamname,streamname,extname);
  positionreset();

  configppp->debug("opened input file '" + streamname + "'.");
}




//--------------------------------------------------------------------------------
//---- CfpSourceFile::close() ----------------------------------------------------
//--------------------------------------------------------------------------------

void CfpSourceFile::close()
{
  // close any streams, but don't delete internal memory.  cfpe to call if already closed
  fstream *fsp;

  // cast inputstream pointer to a file stream, which we know it is because of where we are
  fsp=(fstream *)inputstreamp;

  if (fsp)
    {
      if (fsp->is_open())
	{
	  fsp->close();
	  configppp->debug("closed input file '" + streamname + "'.");
	}
      // now kill the stream
      delete inputstreamp;
      inputstreamp=NULL;
    }
}




//--------------------------------------------------------------------------------
//---- CfpSourceFile::destruct() -------------------------------------------------
//--------------------------------------------------------------------------------

void CfpSourceFile::destruct()
{
  // replacement destructor
  // normally we would handle in ~CfpSourceFile, but there seems to be an error
  //  in the stl delete, so we move the cleanup here
  // configppp->debug("killed file input source '" + streamname + "'.");
  close();
}




//--------------------------------------------------------------------------------
//---- CfpSourceString::CfpSourceString() ----------------------------------------
//--------------------------------------------------------------------------------

CfpSourceString::CfpSourceString(Configpp *inconfigppp, string inname, string parsestring)
  : CfpInputSource()
{
  // create a new input stream from the string

  // set base values
  configppp=inconfigppp;
  streamname=inname;

  // allocate a copy of the incoming string
  parsecopy=new char[parsestring.length()+1];
  strcpy(parsecopy,parsestring.c_str());

  openstream();
}




//--------------------------------------------------------------------------------
//---- CfpSourceString::openstream() ---------------------------------------------
//--------------------------------------------------------------------------------

void CfpSourceString::openstream()
{
  // actually open the string stream
  inputstreamp=new istringstream(parsecopy);
  if (!inputstreamp)
    throw CfpError("could not create an input stream from the string " + streamname + " '" + parsecopy + "'","",io,false,configppp);

  positionreset();
  configppp->debug("opened istringstream " +streamname+ " (" + parsecopy +").");
}




//--------------------------------------------------------------------------------
//---- CfpSourceString::close() ---------------------------------------------------
//--------------------------------------------------------------------------------

void CfpSourceString::close()
{
  // close any streams, but don't delete internal memory.  cfpe to call if already closed
  if (inputstreamp)
    {
      // kill the string stream
      configppp->debug("closed istringstream "+ streamname +".");
      delete inputstreamp;
      inputstreamp=NULL;
    }
}



//--------------------------------------------------------------------------------
//---- CfpSourceString::destruct() -----------------------------------------------
//--------------------------------------------------------------------------------

void CfpSourceString::destruct()
{
  // replacement destructor
  // normally we would handle in ~CfpSourceString, but there seems to be an error
  //  in the stl delete, so we move the cleanup here
  // configppp->debug("killed input source '" + streamname + "'.");
  close();

  // delete the char array we allocated
  delete [] parsecopy;
}










//--------------------------------------------------------------------------------
//---- CfpToken::CfpToken() ------------------------------------------------------
//--------------------------------------------------------------------------------

CfpToken::CfpToken(CfpTokenType intt,string intv)
{
  // constructor
  tokentype=intt;
  tokenval=intv;
  // should init streamloc also
}



//--------------------------------------------------------------------------------
//---- CfpToken::set() -----------------------------------------------------------
//--------------------------------------------------------------------------------

void CfpToken::set(CfpTokenType intype, string instrval)
{
  // set the token values
  tokentype=intype;
  tokenval=instrval;
}



//--------------------------------------------------------------------------------
//---- CfpToken::operator<< ------------------------------------------------------
//--------------------------------------------------------------------------------

ostream& operator<< (ostream &stream, CfpToken token)
{
  // replacement inserter, for printing debug info
  stream << "Token: '" << token.getstr() << "'.\n";
  return stream;
}





//--------------------------------------------------------------------------------
//---- CfpLexer::CfpLexer() ------------------------------------------------------
//--------------------------------------------------------------------------------

CfpLexer::CfpLexer(Configpp *inconfigppp)
{
  // constuctor
  configppp=inconfigppp;
  semihint=false;
}



//--------------------------------------------------------------------------------
//---- CfpLexer::~CfpLexer() -----------------------------------------------------
//--------------------------------------------------------------------------------

CfpLexer::~CfpLexer()
{
  // destuctor
  CfpInputSource *sisp;

  configppp->debug("Destructing the lexer.");

  // cycle through each current stack input source, and pop it off the stack.
  while (inputs.size()>0)
    popsource();

  // now the stackhistory is where we actually delete the input sources
  while (inputhistory.size()>0)
    {
      // We manually call custom destruct() because of gcc 2.8.1 BUG which fails to call the
      //  derived destructor (hopefully only with stl classes)
      (inputhistory.front())->destruct();
      // Now really delete it
      delete inputhistory.front();
      // Remove it from history stack
      inputhistory.pop_front();
    }
}




//--------------------------------------------------------------------------------
//---- CfpLexer::popsource() -----------------------------------------------------
//--------------------------------------------------------------------------------

void CfpLexer::popsource()
{
  // Close the front stream on the input stack, and remove it from our current-input stack.
  // BUT don't delete the input source object, because a copy remains on our inputhistory

  // Jesse, egcs BUG dumps core when i throw here!
  if (inputs.size()<=0)
    throw CfpError("tried to pop an input stream, but there were none to pop","",internal,true,configppp);

  // close the stream associated with the source (but don't delete source)
  (inputs.front())->close();

  // remove it from stack
  inputs.pop_front();
}




//--------------------------------------------------------------------------------
//---- CfpLexer::readchar() ------------------------------------------------------
//--------------------------------------------------------------------------------

int CfpLexer::lexreadchar()
{
  // get a char from the front stream, pop it if we hid eof, and 
  // RETURN 0 on end of all streams
  int c;

  // if we have no more streams then return special end-of-all-stream character
  if (inputs.size()<=0)
    return ENDSTREAMS_CHAR;

  // otherwise get a char from the top stream
  c=inputs.front()->readchar();
  if (c==EOF)
    {
      // we hit an eof of the current front stream, so just close it and return eof
      popsource();
    }
  else if (c==ENDSTREAMS_CHAR)
    {
      // yikes, someone is using our special character in their input file.
      //  throw an error.
      throw CfpError("illegal character","",lexical,false,configppp);
    }

  return c;
}




//--------------------------------------------------------------------------------
//---- CfpLexer::lexpeekchar() ---------------------------------------------------
//--------------------------------------------------------------------------------

int CfpLexer::lexpeekchar()
{
  // get a char from the front stream, and
  // RETURN 0 on end of all streams
  int c;

  // if we have no more streams then return special end-of-all-stream character
  if (inputs.size()<=0)
    return ENDSTREAMS_CHAR;

  // otherwise get a char from the top stream
  c=inputs.front()->peekchar();
  if (c==ENDSTREAMS_CHAR)
    {
      // yikes, someone is using our special character in their input file.
      //  throw an error.
      throw CfpError("illegal character","",lexical,false,configppp);
    }
  return c;
}




//--------------------------------------------------------------------------------
//---- CfpLexer::pushfile() ------------------------------------------------------
//--------------------------------------------------------------------------------

void CfpLexer::pushfile(string filename)
{
  // create a file stream and push it on lexer stack
  CfpSourceFile *sfile;

  sfile=new CfpSourceFile(configppp, filename, ".cfp");

  // now add pointers on permanent history stack and current parse stack
  inputhistory.push_front(sfile);
  inputs.push_front(sfile);
}



//--------------------------------------------------------------------------------
//---- CfpLexer::pushstring() ----------------------------------------------------
//--------------------------------------------------------------------------------

void CfpLexer::pushstring(string parsestring, string name)
{
  // create a string stream and push it on lexer stack
  CfpSourceString *sstring;

  sstring=new CfpSourceString(configppp, name, parsestring);

  // now add pointers on permanent history stack and current parse stack
  inputhistory.push_front(sstring);
  inputs.push_front(sstring);
}






//--------------------------------------------------------------------------------
//---- CfpLexer::readtoken() -----------------------------------------------------
//--------------------------------------------------------------------------------

CfpToken CfpLexer::readtoken(TokenReadMode rmode,bool skipping)
{
  // call internalreadtoken() until we get a non-comment token
  CfpTokenType ttype;

  // reset the semicolon hint flag
  semihint=false;

  // now get a valid token
  for (;;)
    {
      if (peektokens.size()>0)
	{
	  // there is a token to peek
	  if (rmode!=tokenMode && rmode!=tokencrMode)
	    throw CfpError("trying to recover a peeked token in a non token mode.","",internal,true,configppp);
	  curtoken=peektokens.front();
	  peektokens.pop_front();
	}
      else
	{
	  // read a new token frrom the inputs
	  internalreadtoken(rmode);
	}
      ttype=curtoken.gettype();
      if (ttype==commentVal && !skipping)
	{
	  // if it's a comment and the skipping flag is not set (ie we are executing the tokens)
	  configppp->handlecomment(curtoken.getstr(),((curtoken.getlocp())->getssourcep())->getname());
	  continue;
	}
      else if (ttype==crSym && rmode==tokenMode)
	{
	  // in tokenmode we throw away CRs
	  // BUT make note that we hit a CR so that we can hint that they may have forgotten a semicolon
	  semihint=true;
	  continue;
	}
      else
	break;
    }

  // set the end location hint for the parse to the last location
  (configppp->gethintlocp())->endcopy(curtoken.getlocp());
  return curtoken;
}



//--------------------------------------------------------------------------------
//---- CfpLexer::peektoken() -----------------------------------------------------
//--------------------------------------------------------------------------------

CfpToken CfpLexer::peektoken(TokenReadMode rmode,bool advance)
{
  // we want to look at a token, but leave it for readtoken to get it next time.
  //  we do it by really reading the tokens, but pushing them on a FIFO stack that
  //  readtoken will get next time.  that way we can peek more than one token in.
  // note: we only work in tokenMode and tokencrMode, since the other modes grab words
  //  in a way that we can't recover flexibly.
  CfpTokenType ttype;

  if (rmode!=tokenMode && rmode!=tokencrMode)
    throw CfpError("trying to peek a token in a non token mode.","",internal,true,configppp);

  // reset the semicolon hint flag
  semihint=false;

  if (advance==false && peektokens.size()>0)
    {
      // the last peek got it, and this peek again see the same peek as last time
      //      cout << "... using peek from buffer.\n";
      curtoken=peektokens.front();
      return curtoken;
    }

  // now get a valid token
  for (;;)
    {
      // read a new token from the inputs and push it on the peek stack
      internalreadtoken(tokenMode);

      peektokens.push_front(curtoken);

      ttype=curtoken.gettype();

      // new test
      if (ttype==commentVal)
	{
	  // if it's a comment and the skipping flag is not set (ie we are executing the tokens)
	  configppp->handlecomment(curtoken.getstr(),((curtoken.getlocp())->getssourcep())->getname());
	  continue;
	}
      else

      if (ttype==crSym && rmode==tokenMode)
	{
	  // in tokenmode we throw away CRs
	  semihint=true;
	  continue;
	}
      else
	break;
    }

  // set the end location hint for the parse to the last location
  (configppp->gethintlocp())->endcopy(curtoken.getlocp());
  return curtoken;
}



//--------------------------------------------------------------------------------
//---- CfpLexer::consumepeeks() --------------------------------------------------
//--------------------------------------------------------------------------------

void CfpLexer::consumepeeks()
{
  // consume any tokens that have been peeked but not read
  while (peektokens.size()>0)
    {
      // there is a token to peek
      peektokens.pop_front();
    }
}



//--------------------------------------------------------------------------------
//---- CfpLexer::readpuretoken() -----------------------------------------------------
//--------------------------------------------------------------------------------

CfpToken CfpLexer::readpuretoken()
{
  // just read the first token, skipping all CRs
  CfpToken toke;

  for (;;)
    {
      toke=readtoken(tokenMode);
      if (toke.gettype()!=crSym)
	break;
    }
  return toke;
}



//--------------------------------------------------------------------------------
//---- CfpLexer::internalreadtoken() ---------------------------------------------
//--------------------------------------------------------------------------------

void CfpLexer::internalreadtoken(TokenReadMode rmode)
{
  // Read a lexical token, storing it in the member object curtoken.
  // Note: this is the central job of the lexer, everything above is just support.
  // rmode identifies the kind of token the caller wants.
  //  rmode=tokenMode, then get a parser token, skip whitespaces, etc. treat CRs as whitespace.
  //  rmode=tokencrMode, then get a parser token, skip whitespace, etc. BUT return CR tokens.
  //  rmode=wordMode, then get everything between whitespaces, don't worry about symbols, etc.
  //  rmode=crlineMode, then get everything before a CR.
  string tstring;
  string errstring;
  int tlen;
  int c,c2,cminus2;
  CfpTokenType ttype;
  bool instring;
  bool insquote;
  bool inlinecomment;
  int inblockcomment;
  CfpInputSource *cursourcep=NULL;
  long curline,curlinepos;
  CfpLoc *curlocp;
  bool saidtoolong=false;

  // initialize the current token
  curtoken.reset();
  tstring="";
  tlen=0;
  ttype=initVal;
  instring=false;
  insquote=false;
  inlinecomment=false;
  inblockcomment=0;
  curlocp=curtoken.getlocp();


  // Get the token
  for (;;)
    {
      // Take a peek at the next character available
      c=lexpeekchar();

      // have we reached the end of all streams
      if (c==ENDSTREAMS_CHAR)
	break;

      // is it an initial whitespace?
      if (tlen==0 && (isspace(c) || c==EOF))
	{
	  if (rmode==crlineMode && c=='\n')
	    {
	      // if we are in crlineMode, an initial CR is not skipped, but consumed and a blank val returned
	      c2=lexreadchar();
	      ttype=wordVal;
	      break;
	    }
	  // jesse
	  //	  else if ((rmode==tokenMode || rmode==tokencrMode) && c=='\n')
	  //	    {
	  //	      // an initial CR is not skipped, but consumed and a blank val returned
	  //	      c2=lexreadchar();
	  //	      ttype=crSym;
	  //	      break;
	  //	    }
	  else if (rmode==tokencrMode && c=='\n')
	    {
	      // an initial CR is not skipped, but consumed and a blank val returned
	      c2=lexreadchar();
	      ttype=crSym;
	      break;
	    }
	  else if (rmode==wordMode && c=='\n')
	    {
	      // if we are in wordMode, an initial CR is not skipped, but is left on input, blank val returned
	      ttype=wordVal;
	      break;
	    }
	  else
	    {
	      // consume it, throw it away, and continue
	      c2=lexreadchar();
	      continue;
	    }
	}

      // Are we currently working on a token?
      if (tlen>0)
	{
	  // Yes, we are working on a token. Is the new character a separtor for the current working token?
	  if (c==EOF)
	    {
	      // we don't let any tokens span streams, we could change this if we wanted but it seems excessive.
	      break;
	    }
	  else if (rmode==crlineMode)
	    {
	      if (c=='\n')
		{
		  // when in crline read mode, a CR signifies token termination, AND we also want to consume
		  // the CR before we return.
		  c2=lexreadchar();
		  break;
		}
	    }
	  else if (inlinecomment)
	    {
	      // we are in a line comment
	      if (c=='\n')
		{
		  inlinecomment=false;
		  break;
		}
	    }
	  else if (inblockcomment)
	    {
	      // we are in a block comment
	      // so we don't check for seperator
	    }
	  else if (instring)
	    {
	      // we are in a string
	      // so we don't check for seperator
	    }
	  else if (insquote)
	    {
	      // we are in a singly quoted string
	      // so we don't check for seperator
	    }
	  else if (ischarseperator(ttype,c,tstring[tlen-1],tlen))
	    {
	      // yes it's a seperator, so break out, leaving the separator to be read next time.
	      break;
	    }
	}
      
      // make sure the token isn't growing out of control
      if (tlen>=MAX_TOKENLENGTH-1)
	{
	  if (inblockcomment)
	    {
	      if (!saidtoolong)
		{
		  // first time we print warning
		  configppp->debug("truncating long block comment token (not normally used for anything).");
		  saidtoolong=true;
		}
	      else
		{
		  // reduce size of comment token (but keep endparts to insure other stuff works
		  tstring.erase(4,tlen-8);
		  tlen=8;
		  //		  cout << "modified commentoken to '"<< tstring<<"'.\n";
		}
	    }
	  else
	    {
	      // first let the location struct know that it should consider the end at eof
	      curlocp->seteoflag();
	      if (instring)
		throw CfpError("run-away string is too long","",lexical,true,true,curlocp);
	      else if (insquote)
		throw CfpError("run-away quotestring is too long","",lexical,true,true,curlocp);
	      else
		throw CfpError("token has gone on too long","",lexical,true,true,curlocp);
	    }
	}

      // Add the character to the current token string
      tstring.append(1, (char)c);
      ++tlen;

      // get the current position
      cursourcep=getcursourcep();
      curline=getcurline();
      curlinepos=getcurlinepos();
      curlocp->endset(cursourcep,curline,curlinepos);
      if (tlen==1)
	{
	  // store the start location of the token
	  curlocp->startset(cursourcep,curline,curlinepos);
	}
  
      // now consume the peeked-at character
      c2=lexreadchar();
      // and check that it is the same as the peeked character (just in case?)
      if (c2!=c)
	{
	  // WHOA! the read character was somehow not equal to what peek told us it was going to be!!
	  errstring="confusion; a peeked character '";
	  errstring.append(1,(char)c);
	  errstring+="' differed from a read character '";
	  errstring.append(1,(char)c2);
	  errstring+="'.";
	  throw CfpError(errstring,"",internal,true,true,curlocp);
	}

      if (tlen==1)
	{
	  // We need to decide what ttype we think the token is, based on FIRST character
	  if (rmode==wordMode || rmode==crlineMode)
	    {
	      // these read modes don't care about strings, etc., and only type wordVal, and EOF are returned
	      ttype=wordVal;
	    }
	  else
	    {
	      ttype=toketypechar(c);
	      if (c=='\"')
		instring=true;
	      if (c=='\'')
		insquote=true;
	    }
	}
      else if (tlen>1 && !inlinecomment && rmode==tokenMode || rmode==tokencrMode)
	{
	  // It's an ongoing token.
	  // We make some special checks for strings and comments
	  cminus2=tstring[tlen-2];
	  if (tlen>3)
	    {
	      // cminus2 is used to peek at the prior character
	      //  but if the prior prior character in a string is a '\' then it's an escape char, and we ignore
	      if ((instring || insquote) && tstring[tlen-3]=='\\')
		cminus2=' ';
	    }

	  if (c=='\"' && cminus2!='\\' && instring && !inblockcomment)
	    {
	      // we've got the end of a string
	      instring=false;
	      break;
	    }
	  else if (c=='\'' && cminus2!='\\' && insquote && !inblockcomment && !instring)
	    {
	      // we've got the end of a singly quoted string
	      insquote=false;
	      break;
	    }
	  else if (c=='/' && cminus2=='*' && !instring && !insquote)
	    {
	      // we've got the end of a block comment
	      --inblockcomment;
	      if (inblockcomment<0)
		throw CfpError("unbalanced block-comment close symbol '*/'","",syntactical,true,true,curlocp);
	      if (!inblockcomment)
		break;
	    }
	  else if (c=='*' && cminus2=='/' && !instring && !insquote)
	    {
	      // we've got the start of a new (poss. hierarchical) block comment
	      //  this is still considered part of the higher comment token
	      ++inblockcomment;
	      if (inblockcomment==2)
		configppp->warning("entering nested block comment.",curlocp);
	      ttype=commentVal;
	    }
	  else if (c=='/' && cminus2=='/' && !inblockcomment && !instring && !insquote)
	    {
	      // we've got the start of a line comment
	      inlinecomment=true;
	      ttype=commentVal;
	    }
	  else if (c=='|' && cminus2=='|')
	    {
	      // normally we think of bars (|) as specifying a preproccessing textual replacement,
	      // but if we read two of them next to each other (||) then we consider it to be an C++ like "OR" token
	      ttype=conditionVal;
	    }
	}
    }


  // any clear errors?
  if (instring || inblockcomment || insquote)
    {
      if (c==EOF || c==ENDSTREAMS_CHAR)
	{
	  // we want to correct tell the location struct that it ended on eof
	  curlocp->seteoflag();
	}
      if (instring)
	throw CfpError("unterminated string","",syntactical,true,true,curlocp);
      else if (inblockcomment)
	throw CfpError("unterminated block-comment '/*'","",syntactical,true,true,curlocp);
      else if (insquote)
	throw CfpError("unterminated singly-quoted string","",syntactical,true,true,curlocp);
    }

  // any warnings to issue?
  if (cursourcep!=NULL)
    {
      if (curline!=curlocp->getstartline() && ttype==stringVal)
	{
	  // it is *probably* an error for a string to span lines
	  configppp->warning("string spans multiple lines.",curtoken.getlocp());
	}
    }

  // make sure we return a valid token
  if (ttype==initVal)
    {
      if (tlen==0)
	{
	  // this happens only on eof, so no more tokens available.
	  ttype=eofSym;
	}
      else
	{
	  // this should never happen
	  throw CfpError("failed to parse a proper token","",internal,true,true,curlocp);
	}
    }

  // replace any escapes (\n,etc.) with the real characters
  tstring=JRescaperep(tstring);

  // do any preprocessor replacement
  tstring=configppp->tokenreplacement(tstring);

  // Set the current token to what we read, and what we think it is, and return it.
  curtoken.set(ttype,tstring);

  return;
}






//--------------------------------------------------------------------------------
//---- CfpLexer::toketypechar() --------------------------------------------------
//--------------------------------------------------------------------------------

CfpTokenType CfpLexer::toketypechar(int c)
{
  // figure out token type from first char of token
  CfpTokenType rettype;

  if (c==';')
    rettype=semicolonSym;
  else if (c=='{')
    rettype=openbraceSym;
  else if (c=='}')
    rettype=closebraceSym;
  else if (c=='[')
    rettype=opensquareSym;
  else if (c==']')
    rettype=closesquareSym;
  else if (c=='(')
    rettype=openparenSym;
  else if (c==')')
    rettype=closeparenSym;
  else if (c==',')
    rettype=commaSym;
  else if (c=='#')
    {
      // the pound symbol
      rettype=poundVal;
    }
  else if (isalpha(c) || c=='|')
    {
      // it's a letter, so the rest of the token can be letters, digits, underscores
      rettype=alphaVal;
    }
  else if (isdigit(c) || c=='.')
    {
      // it's numeric, so the rest of the token must be digits, '-', '.', 'e'
      rettype=numericVal;
    }
  else if (c=='+' || c=='-' || c=='*' || c=='/' || c=='^' || c=='%')
    {
      // it's a mathematical operator
      rettype=mathVal;
    }
  else if (c=='<' || c=='>' || c=='=' || c=='&' || c=='!')
    {
      // it's a conditional operator
      rettype=conditionVal;
    }
  else if (c=='\"')
    {
      // it's the start of a string
      rettype=stringVal;
    }
  else if (c=='\'')
    {
      // it's the start of a singly quoted string
      rettype=squotestringVal;
    }
  else if (c=='\n')
    {
      // we only get this character if we are reading in a generic word in (readmode=word or line)
      rettype=wordVal;
    }
  else
    {
      // it's another symbole ($,@, etc.)
      rettype=otherSym;
    }

  return rettype;
}




//--------------------------------------------------------------------------------
//---- CfpLexer::ischarseperator() -----------------------------------------------
//--------------------------------------------------------------------------------

bool CfpLexer::ischarseperator(CfpTokenType &ttype, int c, int lc, int len)
{
  // is the char a seperator for the token?

  switch(ttype)
    {
    case semicolonSym:
    case openbraceSym:
    case closebraceSym:
    case opensquareSym:
    case closesquareSym:
    case openparenSym:
    case closeparenSym:
    case commaSym:
    case otherSym:
      return true;
      break;
    case alphaVal:
      if (isalpha(c) || isdigit(c) || c=='_' || c=='.' || c=='|')
	return false;
      else
	return true;
      break;
    case numericVal:
      if (isdigit(c))
	return false;
      else if (c=='e' || c=='E' || c=='.')
	return false;
      else if (c=='-' && (lc=='e' || lc=='E'))
	return false;
      else
	return true;
      break;
    case stringVal:
      // else where we detect string close
      return false;
      break;
    case mathVal:
      if (ttype==mathVal && isdigit(c) && len==1 && (lc=='-' || lc=='+') )
	{
	  // fix up for values that initially looked like math operators (ie -3 or +3)
	  ttype=numericVal;
	  return false;
	}
      if (c=='+' || c=='-' || c=='*' || c=='/' || c=='^' || c=='%')
	return false;
      else
	return true;
      break;
    case conditionVal:
      if (c=='<' || c=='>' || c=='=' || c=='&' || c=='|' || c=='!')
	return false;
      else
	return true;
      break;
    case commentVal:
      return false;
      break;
    case poundVal:
      if (c=='(' || c==')' || isspace(c))
	return true;
      else
	return false;
      break;
    case wordVal:
      if (isspace(c))
	return true;
      else
	return false;
      break;
    }
  
  // if we are here then something went wrong
  throw CfpError("failed to parse token character","",internal,true,configppp);
  return true;
}







//--------------------------------------------------------------------------------
//---- CfpLexer::getcurline() ----------------------------------------------------
//--------------------------------------------------------------------------------

long CfpLexer::getcurline()
{
  if (inputs.size()<=0)
    throw CfpError("attempt to get current line number but no input streams are open","",internal,true,configppp);
  return inputs.front()->getcurline();
}


//--------------------------------------------------------------------------------
//---- CfpLexer::getcurlinepos() -------------------------------------------------
//--------------------------------------------------------------------------------

long CfpLexer::getcurlinepos()
{
  if (inputs.size()<=0)
    throw CfpError("attempt to get current line position but no input streams are open","",internal,true,configppp);
  return inputs.front()->getcurlinepos();
}


//--------------------------------------------------------------------------------
//---- CfpLexer::getcursourcep() -------------------------------------------------
//--------------------------------------------------------------------------------

CfpInputSource *CfpLexer::getcursourcep()
{
  if (inputs.size()<=0)
    throw CfpError("attempt to get current line number but no input streams are open","",internal,true,configppp);
  return inputs.front();
}



string CfpLexer::getcurfilename()
{
  // return the name of the current file we are parsing
  return inputs.front()->getcurfilename();
}

string CfpSourceFile::getcurfilename()
{
  // return the name of the current file stream
  return streamname;
}


long CfpLexer::getcurpos()
{
  // return pos of current file we are parsing
  return inputs.front()->getpos();
}


CfpInputSource *CfpLexer::getcursource()
{
  // return pointer to current source
  return inputs.front();
}






void CfpLexer::showpeeks()
{
  cout << "... There are "+JRtostr((int)(peektokens.size()))+" tokens on the stack waiting.\n";
}



//}
// Namespace close-brace

