//  Jesse Reichler, June 1996
//  Fred Delcomyn
//
//  NEW contact.C
//  This file contains the class definition for the Contact class,
//   used to manage contacts & forces derived from contacts.
//  See contact.h for more info
//  
//

#include "misc.def"
#include "contact.h"
#include "physobject.h"
#include "world.h"

#include <stdio.h>
#include <math.h>
#include <stdlib.h>








Surface::Surface(percision inKx,percision inKy,percision inKz,percision inYx,percision inYy,percision inYz, percision inmusx,percision inmusy,percision inmusz, percision inmudx,percision inmudy,percision inmudz)
{
  // set the constants of the floater contact
  Kx=inKx;
  Ky=inKy;
  Kz=inKz;
  Yz=inYz;
  Yx=inYx;
  Yy=inYy;
  musx=inmusx;
  musy=inmusy;
  musz=inmusy;
  mudx=inmudx;
  mudy=inmudy;
  mudz=inmudz;
}







Attachment::Attachment(char *inname, World *inworldp, Physicalobject *inobj1, Physicalobject *inobj2, int insmartspring)
  : Contact(inobj1,inobj2,insmartspring)
{
  // construct an attachment from newconfig
  strcpy(name,inname);
  worldp=inworldp;
  
  // initialize
  attached=YES;
  xslipping=NO;
  yslipping=NO;
  zslipping=NO;

  // now ADD the contact attactment
  worldp->addcontact(this);
}










void Attachment::update(double deltat)
{
  // Compute forces from floating object
  perVectorR3 deltaloc;
  perVectorR3 deltaloc2;
  perVectorR3 rotdeltaloc;
  perVectorR3 rotdeltaloc2;
  perVectorR3 veloc;
  perVectorR3 obj2veloc;
  perMatrix transOBJ2_TO_WORLD;
  perMatrix transWORLD_TO_OBJ2;
  perVectorR3 forces;
  perVectorR3 moments;
  perVectorR3 Z0(ZEROVEC3);
 
  if (attached==NO)
    return;
  
  // Frame transformers to/from obj2 frame

  // should we get the normal at the last contact point or at current point?
  //  obj2->normalframe(contactloc,transWORLD_TO_OBJ2);
  obj2->normalframe(obj1->locationval(),transWORLD_TO_OBJ2);

  transOBJ2_TO_WORLD=~transWORLD_TO_OBJ2;

  // delta location in world and ob2 (floater) frame
  deltaloc2=(obj1->locationval())-(obj2->locationval());
  rotdeltaloc=transWORLD_TO_OBJ2*(perVector)deltaloc2;

  // Now let slip the axes of war!
  // first zero out movement of nonslipping dimensions in floater frame
  if (xslipping==NO)
    rotdeltaloc(0)=0.0;
  if (yslipping==NO)
    rotdeltaloc(1)=0.0;
  if (zslipping==NO)
    rotdeltaloc(2)=0.0;

  // now rotate back to world frame and move floater
  deltaloc2=transOBJ2_TO_WORLD* (perVector)rotdeltaloc;
  obj2->setlocation(obj2->locationval()+deltaloc2);

  // reset slipping flags
  xslipping=NO;
  yslipping=NO;
  zslipping=NO;

  // Now deltaloc from new slipped position
  deltaloc=(obj1->locationval())-(obj2->locationval());
  
  // Now convert vects to obj2 frame (perpendic to ground tile, for examp.)
  // JESSE GCC 7/1/98 (perVector)
  deltaloc=transWORLD_TO_OBJ2*(perVector)deltaloc;
  
  // FOR velocity, now we want the *relative* velocity of obj1 in obj2 frame
  //   so we subtract out the local velocity of obj2
  //   this veloc is used for computation of slipping, etc.

  veloc=transWORLD_TO_OBJ2*(perVector)(obj1->velocityval());
  obj2veloc=obj2->velocityval();

  // JESSE GCC 7/1/98 (perVector)
  veloc-=(transWORLD_TO_OBJ2*(perVector)obj2veloc);
  
  // So now deltaloc and veloc are in obj2 (floater) frame
  
  // Following Freeman Thesis, compute forces of spring/damper system
  forces=computeforces(deltaloc,veloc);
  moments=Z0;
  
  // Now add the forces to sum external forces at obj1, and opposites
  // at obj2, in *WORLD* frame, which we convert to

  // JESSE GCC 7/1/98 (perVector)
  forces=transOBJ2_TO_WORLD*(perVector)forces;
  moments=transOBJ2_TO_WORLD*(perVector)moments;

  // now add the external forces to the physical objects
  sendextforces(forces);
}




perVectorR3 Attachment::computeforces(perVectorR3 deltaloc,perVectorR3 veloc)
{
   // Compute the forces of a point on obj1 which is penetrating obj2
   // Following Freeman 1989 thesis (p.36)
   perVectorR3 forces;
   percision fx,fy,fz;
   percision fxd,fyd,fzd;
   
   forces.fill(0.0);

   // As per Freeman thesis (p.36) calculate fz, fx', fy'
   // Calculate dampers
   fxd=(Yx*veloc(0));
   fyd=(Yy*veloc(1));
   fzd=(Yz*veloc(2));

   // force=spring + damper
   fx=(-1.0*Kx)*deltaloc(0)-fxd;
   fy=(-1.0*Ky)*deltaloc(1)-fyd;
   fz=(-1.0*Kz)*deltaloc(2)-fzd;

   if (fabs(fx)>=musx)
     {
       //       printf( "... X slips. %g > %g\n",fx,musx);
       // X escapes static friciton cone and slips
       xslipping=YES;
       // apply a dynamic friction term
       fx=mudx*veloc(0);
     }
   else
     xslipping=NO;
   
   if (fabs(fy)>=musy)
     {
       //       printf( "... Y slips. %g > %g\n",fy,musy);
       // Y escapes static friciton cone and slips
       yslipping=YES;
       // apply a dynamic friction term
       fy=mudy*veloc(1);
     }
   else
     yslipping=NO;
   
   if (fabs(fz)>=musz)
     {
       //       printf( "... Z slips. %g > %g\n",fz,musz);
       // Z escapes static friciton cone and slips
       zslipping=YES;
       // apply a dynamic friction term
       fz=mudz*veloc(2);
     }
   else
     zslipping=NO;

   // Set the force vector
   forces(0)=fx; forces(1)=fy; forces(2)=fz;

   return forces;
}



void Attachment::updateconstants(perVectorR3 loc)
{
  // update sprng/damper constants and frictional coefficients
  // of the floating constrainer,
  // in the floater frame, at point loc.
  // SO most of the time these constants won't change, but you could
  //  imagine wanting to constrain an object on an arbitrary trajectory.
}
   


void Attachment::sendextforces(perVectorR3 forces)
{
  // send the contact forces to the objects
  // Forces are provided in earth coords, at contact point
  // SO we need to move them to object coords and aadjust for "comdist"
  //  the distance of the force from the COM of the objects
  perVectorR3 Z0(ZEROVEC3);

  obj1->addextforce(forces,Z0);
  obj2->addextforce(-1.0*forces,-1.0*Z0);
}








Contact::Contact()
{
  // Incomplete constructin of a contact class
  obj1=NULL;
  obj2=NULL;
  smartspring=YES;
  smartspring=NO;
  multfact=1.0;
}



Contact::Contact(Physicalobject *ob1,Physicalobject *ob2, int insmartspring)
{
  // Construct a contact class
  //  Which keeps track of possible contact interaction between
  //  object1 and another specific object2
  
  // Set the two objects
  obj1=ob1;
  obj2=ob2;
  
  // reset multiplier
  smartspring=insmartspring;
  multfact=1.0;
}










PointContact::PointContact(Physicalobject *ob1, Physicalobject *ob2, PointTrack *inpt, int insmartspring):
  Contact(ob1,ob2,insmartspring)
{
  // Object 1 is the object on whom we are checking a specific point
  
  // pointer to new trackpoint
  pt=inpt;
  
  // initialize
  contacted=NO;
  slipping=NO;
  hookloc.fill(0.0);

  //  cout << "created a new point contact. obj2\n";
  //  GroundTile *gt=(GroundTile *)obj2;
  //  cout <<" Kz="<<gt->Kz<<".\n";
}



void PointContact::update(double deltat)
{
  // Check for contact and do updating
  int iscontact;
  perVectorR3 deltaloc;
  perVectorR3 deltaloc2;
  perVectorR3 rotdeltaloc;
  perVectorR3 veloc;
  perVectorR3 obj2veloc;
  perMatrix transOBJ2_TO_WORLD(3,3);
  perMatrix transWORLD_TO_OBJ2(3,3);
  perVectorR3 forces;
  perVectorR3 moments;
  perVectorR3 Z0(ZEROVEC3);

  // was there contact?  If not, just return
  iscontact=touching(contactloc,deltat);
  if (iscontact==NO)
    {
      contacted=NO;
      slipping=NO;
      return;
    }
 
  // Frame transformers to/from obj2 frame
  obj2->normalframe(contactloc,transWORLD_TO_OBJ2);
  transOBJ2_TO_WORLD=~transWORLD_TO_OBJ2;
  
  // Update the hook position (in world frame)
  // Is this the initial contact?
  if (contacted==NO)
    {
      // Yes this is initial contact
      // Set the hook position to the previous position(see Freeman Thesis)
      hookloc=pt->prevloc;
      multfact=1.0;
    }
  else if (slipping==YES)
    {
      // If it is slipping, x&y of hook slip.
      // We want x,y component to readjust to prevloc BUT since it could be an angled tile, we want x,y component
      //  in OBJECT frame, so we take delta, convert to obj, then remove z component, and rotate back and add.
      deltaloc2=pt->curloc-hookloc;
      
      // JESSE GCC 7/1/98 (perVector)
      rotdeltaloc=transWORLD_TO_OBJ2*(perVector)deltaloc2;
      
      rotdeltaloc(2)=0.0;
      
      // JESSE GCC 7/1/98 (perVector)
      deltaloc2=transOBJ2_TO_WORLD* (perVector)rotdeltaloc;
      
      hookloc+=deltaloc2;
      slipping=NO;  // reset slipping flag
    }
  else if (slipping==NO)
    {
      // if velocity of obj2 (ground) is 0, then no update of hook needed
      // BUT if we are hooked on a tread mill, then we need to adjust the hook position.
      obj2veloc=obj2->velocityval();
      hookloc+=obj2veloc*deltat;
    }
  
  // Set contact flag
  contacted=YES;
  
  // Now deltaloc from hook position
  deltaloc=(pt->curloc)-hookloc;
  
  // Now convert vects to obj2 frame (perpendic to ground tile, for examp.)
  deltaloc=transWORLD_TO_OBJ2*(perVector)deltaloc;
  
  // FOR velocity, now we want the *relative* velocit in obj2 frame
  //   so we subtract out the local velocity of obj2
  //   this veloc is used for computation of slipping, etc.
  
  veloc=transWORLD_TO_OBJ2*((perVector)pt->velocity);
  
  obj2veloc=obj2->velocityval();
  veloc-=(transWORLD_TO_OBJ2*(perVector)obj2veloc);
 
  // So now deltaloc and veloc are in obj2 frame, with z axis perpendicular
  //  and positive into obj2 surface.
  
  // Following Freeman Thesis, compute forces of spring/damper system
  forces=computeforces(deltaloc,veloc);
  moments=Z0;
  
  //  cout << "applying forces stage1:\n";
  //  forces.print();

  // Now add the forces to sum external forces at obj1, and opposites
  // at obj2, in *WORLD* frame, which we convert to
  forces=transOBJ2_TO_WORLD*(perVector)forces;
  moments=transOBJ2_TO_WORLD*(perVector)moments;

  //  cout << "new applying forces before sending:\n";
  //  forces.print();

  // now add the external forces to the physical objects
  sendextforces(forces);
 
  // update multiplier, which is meant to heuristically increase spring contstants depending
  // on their success at retarding progress of objects
  if (smartspring==YES)
    {
      if (veloc(2)<=0.0)
	{
	  // we decrease multiplicative factor as soon as they start leaving
	  multfact/=100.0;
	  if (multfact<1.0)
	    multfact=1.0;
	}
      else
	multfact*=1.3;
    }
}





perVectorR3 PointContact::computeforces(perVectorR3 deltaloc,perVectorR3 veloc)
{
  // Compute the forces of a point on obj1 which is penetrating obj2
  // Following Freeman 1989 thesis (p.36)
  perVectorR3 forces;
  percision KZ, Kx,Ky, YZ, Yx,Yy, mus, mud;
  percision fx,fy,fz,fzd;
  percision fprimex,fprimey;
  percision fxy;
  percision xy2,mudfz;
  percision yxd,yyd;
  percision propfz;
  float ofz;
  
  forces.fill(0.0);
  
  if (obj2->thingtype()==THING_TRITILE)
    {
      GroundTile *gt=(GroundTile *)obj2;
      // get params
      KZ=gt->Kz;
      Kx=gt->Kx;
      YZ=gt->Yz;
      Yx=gt->Yx;
      Yy=gt->Yy;
      mus=gt->musx;
      mud=gt->mudx;
    }
  else if (obj2->thingtype()==THING_GROUNDMESH)
    {
      GroundMesh *gm=(GroundMesh *)(obj2);
      // get params
      KZ=gm->Kz;
      Kx=gm->Kx;
      YZ=gm->Yz;
      Yx=gm->Yx;
      Yy=gm->Yy;
      mus=gm->musx;
      mud=gm->mudx;
    }
  else
    {
      printf ("ERROR: Don't know to compute contact forces for thing '%s'.\n",obj1->name);
      return forces;
    }
      
  // As per Freeman thesis (p.36) calculate fz, fx', fy'
      
  // normal force
  propfz=(-1.0*KZ)*deltaloc(2);
  fzd=YZ*veloc(2);
  fz=propfz-fzd;
  // uncomment to restore old method
  // propfz=fz;
      
  // smart spring force multiplier
  fz*=multfact;
      
  // test 9/26/98, don't allow the ground spring to pull torwards ground
  ofz=fz;
  if (fz>0.0)
    fz=0.0;
      
  // tangentials without damper
  fprimex=(-1.0*Kx)*deltaloc(0);
  fprimey=(-1.0*Ky)*deltaloc(1);
  
  // Calculate damper
  yxd=(Yx*veloc(0));
  yyd=(Yy*veloc(1));
  
  // test 9/26/98, only allow damper to help spring not hurt it
  //  ie it is a rectified damper
  if (Jsgn(yxd)==Jsgn(fprimex))
    yxd=0.0;
  if (Jsgn(yyd)==Jsgn(fprimey))
    yyd=0.0;
  
  // now add any damper
  fprimex-=yxd;
  fprimey-=yyd;
      
  fxy=sqrt(pow(fprimex,2)+pow(fprimey,2));
  
  // Now check for slipping to calculate fx and fy, and slipping toggle
  // as per p.125 (Freeman thesis 1989)
  
  // 3/4/99 - JESSE modified. shouldn't we use non-clipped fz (ofz) here?
  //          i think that if we don't, the result is excessive slipping
  //      if (fabs(mus*ofz)>=fxy)
  //          no, what we really want is a z term that is not oscilating due to deritivative
  //          component of force, so we should use th distant (proportional) component of z force only
  // original
  //  if (fabs(mus*fz)>=fxy)
  if (fabs(mus*propfz)>=fxy)
    {
      // No slipping, so (Freeman p.59)
      fx=fprimex;
      fy=fprimey;
    }
  else
    {
      // Slipping has occurred
      //      printf( "... Slipping has occured fabs(mus*fz)<fxy    fabs(%g*%g)<%g    %g<%g.\n",mus,propfz,fxy,fabs(mus*propfz),fxy);

      // 3/4/99 - JESSE modified to take into account negative z forces which are clipped to >0 above
      //	  mudfz=mud*fabs(ofz);
      // original
      mudfz=mud*propfz;
      
      xy2=sqrt(pow(veloc(0),2)+pow(veloc(1),2));
      if (xy2==0.0)
	{
	  fx=0.0;
	  fy=0.0;
	}
      else
	{
	  fx=mudfz*( veloc(0)/xy2 );
	  fy=mudfz*( veloc(1)/xy2 );
	}
      slipping=YES;
    }
  
  // Set the force vector
  forces(0)=fx; forces(1)=fy; forces(2)=fz;
  return forces;
}




int PointContact::touching(perVectorR3 &contactloc, double deltat)
{
  // Determine if the two objects are touching, and if so
  // return YES and set the contactloc (point of contact on object2 surface)
  // otherwise return NO
  perVectorR3 locdistal;
  perVectorR3 locproximal;
  GroundTile *gt;
  GroundMesh *gm;
  perVectorR3 sizevec;
  int retv;
  
  // we only handle segment-triangletile type contacts
  if (obj2->thingtype()==THING_TRITILE)
    {
      // set pointers to the shapes
      gt=(GroundTile *)(obj2);
      
      // CONTACT METHOD #2, use foot poisition now and at time t-1
      //  and see if this segment hits the other object
      
      if (contacted==NO)
	{
	  // If not already contacting, the endpoint of test segment is the
	  //  previous location of foot.
	  hookloc=pt->prevloc;
	}
      else
	{
	  // use the hook location set in slipping forces code
	}
      
      retv=gt->linecontact(pt->curloc,hookloc, contactloc);
      return retv;
    }
  else if (obj2->thingtype()==THING_GROUNDMESH)
    {
      // check for contact at groundmesh
      // set pointers to the shapes
      gm=(GroundMesh *)(obj2);
      
      
      if (contacted==NO)
	{
	  // If not already contacting, the endpoint of test segment is the
	  //  previous location of foot.
	  hookloc=pt->prevloc;
	}
      else
	{
	  // use the hook location set in slipping forces code
	}
      
      retv=gm->linecontact(pt->curloc,hookloc, contactloc);
      return retv;
    }
  else
    return NO;
}
  
   

void PointContact::sendextforces(perVectorR3 forces)
{
  // send the contact forces to the objects
  // Forces are provided in earth coords, at contact point
  // SO we need to move them to object coords and aadjust for "comdist"
  //  the distance of the force from the COM of the objects
  perVectorR3 Z0(ZEROVEC3);

  //  if (pt->comdist(0)!=0.0 || pt->comdist(1)!=0.0 || pt->comdist(2)!=0.0)
  //    printf("$$$ CONATCT with a pointtrack not at COM.\n");
  
  obj1->addextforce(forces,Z0);
  obj2->addextforce(-1.0*forces,-1.0*Z0);
}













void PointTrack::updatepoint(perVectorR3 newloc,percision deltat)
{
  // update the location of the point being tracked, and set prevloc&velocity
 
  prevloc=curloc;
  curloc=newloc;

  if (deltat==0.0)
    velocity.fill(0.0);
  else
    velocity=(curloc-prevloc)/deltat;
}





int GroundTile::linecontact(perVectorR3 linep1, perVectorR3 linep2, perVectorR3 &contactloc)
{
  // call the triangle shape to ask
  return ((Triangle *)shape)->linecontact(linep1,linep2,contactloc);
}


GroundTile::GroundTile(const char *objname, class World *worldp, int wowned, percision KZ, percision Ktan, percision YZ, percision Ytan, percision mus, percision mud)
    : Physicalobject(objname,worldp,wowned), Surface(Ktan,KZ,Ytan,YZ,mus,mud)
{
  // for some ungodly reason calling the base Surface constructor fails to initialize
  setK(Ktan,Ktan,KZ);
  setY(Ytan,Ytan,YZ);
  setmus(mus,mus,mus);
  setmus(mud,mud,mud);
}









int GroundMesh::linecontact(perVectorR3 linep1, perVectorR3 linep2, perVectorR3 &contactloc)
{
  // call the triangle shape to ask
  return ((Mesh *)shape)->linecontact(linep1,linep2,contactloc);
}


GroundMesh::GroundMesh(const char *objname, class World *worldp, int wowned, percision KZ, percision Ktan, percision YZ, percision Ytan, percision mus, percision mud)
    : Physicalobject(objname,worldp,wowned), Surface(Ktan,KZ,Ytan,YZ,mus,mud)
{
  // for some ungodly reason calling the base Surface constructor fails to initialize
  setK(Ktan,Ktan,KZ);
  setY(Ytan,Ytan,YZ);
  setmus(mus,mus,mus);
  setmus(mud,mud,mud);
}


