//  Jesse Reichler, December 1996
//  Fred Delcomyn
//
// dynamics.C
// This file contains the new dynamics routines for the multichain mobile body
//  code.  Based on Freeman Thesis (1989)
// Contains Classes from Chain and Base
//
//

#include "matrixinc.h"
#include "misc.def"
#include "chain.h"
#include "link.h"
#include "rot.h"
#include "world.h"
#include "robot.h"

#include <stdio.h>
#include <math.h>






void Link:: Allocatevals()
{
  // Allocate and initialize the Dynamics Data Variables to proper size and type
  inertiaIH=perMatrix(3,3);
  DI=perMatrix(6,6);
  DX=perMatrix(6,6);
  DA=perMatrix(3,3);
  Dw=perVector(3,ColV);
  Dv=perVector(6,ColV);
  Do=perVector(6,ColV);
  DB=perVector(6,ColV);
  Df=perVector(3,ColV);
  DISTAR=perMatrix(6,6);
  DF=perVector(6,ColV);
  DBSTAR=perVector(6,ColV);
  DN=perMatrix(6,6);
  DHT=perVector(6,ColV);
  Da=perVector(6,ColV);
}


void Link:: Computevals()
{
  // Compute some values from others we have loaded in.
  // the 3x3 inertia matrix has been filled prior to calling
  perMatrix m00(3,3);
  perMatrix m01(3,3);
  perMatrix m10(3,3);
  perMatrix m11(3,3);
  perMatrix dtrantilde(3,3);
  perMatrix zero33(3,3);
  perMatrix E(3,3);
  perMatrix D(6,6);
  perMatrix ICM(6,6);
  perMatrix Finertia(6,6);
  perMatrix E6(6,6);
  perMatrix FDi(6,6);
  perMatrix FDit(6,6);
  perMatrix Ftmp(6,6);
  percision a1,a2,a3;

  // Off-line Calculations for Link
  // This is the DI construction routine from Freemans CODE
  //  rather than the equation above

  Dtau=0.0;

  E6.fill(0.0);
  E6(0,0)=1.0; E6(1,1)=1.0; E6(2,2)=1.0; 
  E6(3,3)=1.0; E6(4,4)=1.0; E6(5,5)=1.0; 
  
  Finertia.fill(0.0);
  Finertia(0,0)=inertia(0,0);
  Finertia(1,1)=inertia(1,1);
  Finertia(2,2)=inertia(2,2);
  Finertia(3,3)=kilograms;
  Finertia(4,4)=kilograms;
  Finertia(5,5)=kilograms;
  
  FDi=E6;
  FDi(4,2)=a1=hinge_d(0);
  FDi(3,2)=a2=-(hinge_d(1));
  FDi(3,1)=a3=hinge_d(2);
  FDi(5,1)=(-a1);
  FDi(5,0)=(-a2);
  FDi(4,0)=(-a3);
  FDi(3,0)=FDi(4,1)=FDi(5,2)=0.0;
  
  FDit=~FDi;
  
  Ftmp=Finertia*FDi;
  Finertia=FDit*Ftmp;
  DI=Finertia;

  // Extract upper left 3x3 as the Inertia w.r.t. hinge point [See Freeman p.46]
  inertiaIH=Jessup33(DI);

  // Initialize seg
  Seg=perVectorR3(a_length, d_offset*sin(alpha_twist), d_offset*cos(alpha_twist));
}



void Base:: Allocatevals()
{
  // First just initialize the Dynamics Data Variables to proper size and type
  DI=perMatrix(6,6);
  DISTAR=perMatrix(6,6);
  DBSTAR=perVector(6,ColV);
  DB=perVector(6,ColV);
  Da=perVector(6,ColV);
  Db=perVector(6,ColV);
  DaG=perVector(6,ColV);
  Dv=perVector(6,ColV);
  DEulers=perVector(3);
  DeAb=perMatrix(3,3);
  DbAe=perMatrix(3,3);
  tSUMi=perMatrix(6,6);
  tSUMb=perVector(6,ColV);
}


void Base:: Computevals()
{
  // Calculate the base Inertia Matrix stuff
  perMatrix zero33(3,3);
  perMatrix E(3,3);
  perMatrix temp33(3,3);
  percision tempf;

  // zero out external forces
  tSUMb.fill(0.0);
  tSUMi.fill(0.0);

  // Now do Off-line Calculations for I
  zero33.fill(0.0);
  E=JessE33();
  temp33=kilograms*E;
  DI=JessCompose66(inertia,zero33,zero33,temp33);

  // some zeroing out
  DEulers.fill(0.0);
  Dv.fill(0.0);
  Da.fill(0.0);
  
  // Compute the Body to Earth Transform for the base(body)
  ComputeDeAb();
}




void Chain::setbaselinks(double deltat)
{
  // set any values for base links (# 0) that need to be copied from the base, v(0) and w(0) and a(0)
  // This is NOT in Freeman thesis equations, but in thesis code. (p.120) forward_kinematics
  Base *basep;
  perVector dw(3,ColV);
  perVector tempv6c(6,ColV);
  perVector tempv3c(3,ColV);
  perVector tempv3c2(3,ColV);
  perVector tempv3c3(3,ColV);
  perVector tempv3c4(3,ColV);
  perVector tempv3c5(3,ColV);
  perVector zerov3(3,ColV);

  //  cout << "in setbaselinks.\n";

  //init any constants
  zerov3.fill(0.0);
  basep=robotp->base;
  
  // move base information to link 0 (the HIP)
  link[0]->Dv=rot.DX66(RI_HIP,RI_BODY)*(robotp->base)->Dv;
  link[0]->Dw(0)=link[0]->Dv(0);
  link[0]->Dw(1)=link[0]->Dv(1);
  link[0]->Dw(2)=link[0]->Dv(2);
  link[0]->Da=~(rot.DX66(RI_HIP,RI_BODY)*(robotp->base)->Da);
  link[0]->DI=rot.DX66(RI_HIP,RI_BODY)*(robotp->base)->DI;
  link[0]->DB.fill(0.0);

  //  cout << "leaving setbaselinks.\n";
}



void Chain::forward_kinematics(double deltat)
{
  // Propagate kinematic info from base to chain foot
  // Commented numbers are equation #s from Freeman Thesis
  int i;
  int N=links-1;
  perVector tempv6c(6,ColV);
  perVector tempv3c(3,ColV);
  perVector tempv3c2(3,ColV);
  perVector tempv3c3(3,ColV);
  perVector tempv3c4(3,ColV);
  perVector tempv3c5(3,ColV);
  perVector zerov3(3,ColV);
  percision mi;
  perVector vel(3,ColV);
  perVector velimo(3,ColV);

  //init any constants
  zerov3.fill(0.0);
  
  // now go through links
  for (i=1;i<=N;++i)
    {
      // 3.14
      link[i]->Dw(0)=0.0;
      link[i]->Dw(1)=0.0;
      link[i]->Dw(2)=link[i]->dtheta_angle;

      // 3.15
      link[i]->Dv=rot.DX66(i,i-1) * link[i-1]->Dv;
      tempv6c=JessCompose61(link[i]->Dw,zerov3);
      link[i]->Dv+=tempv6c;
      
      // we need this for first half of Dv
      vel(0)=link[i]->Dv(0);
      vel(1)=link[i]->Dv(1);
      vel(2)=link[i]->Dv(2);
      velimo(0)=link[i-1]->Dv(0);
      velimo(1)=link[i-1]->Dv(1);
      velimo(2)=link[i-1]->Dv(2);

      tempv3c=rot.DXA33(i,i-1)*velimo;
      tempv3c2=JessCross(tempv3c,link[i]->Dw);
      tempv3c3=JessCross(velimo,link[i]->hinge_c);
      tempv3c4=JessCross(velimo,tempv3c3);
      tempv3c5=rot.DXA33(i,i-1)*tempv3c4;
      link[i]->Do=JessCompose61(tempv3c2,tempv3c5);
      
      // 3.17
      tempv3c=Jessup33(link[i]->DI)*vel;
      tempv3c2=JessCross(tempv3c,vel);
      mi=link[i]->kilograms;
      tempv3c3=JessCross(vel,link[i]->hinge_d);
      tempv3c4=JessCross(tempv3c3,vel);
      link[i]->DB=JessCompose61(tempv3c2,mi*tempv3c4);

      // Db was moved to BodyAcceleration Equations, and wi go to wb
    }
}






void Chain::foot_forces(double deltat)
{
  // Process the forces and convert into freeman style
  // also sum torques at joint
  int i;
  int N=links-1;
  Physicalobject *objp;

  // link 0 is the hip
  for (i=0;i<=N;++i)
    {
      // 3.20
      objp=(Physicalobject *)link[i];
      link[i]->Df=(objp->extF)+(objp->userextF);
      
      // Now torques applied at joint
      link[i]->Dtau=link[i]->tauforce();
      link[i]->Dtau+=link[i]->tauoverextend();
      link[i]->Dtau+=link[i]->taubuiltin();
      
      // 3/12/97
      // link[i]->Dtau+=link[i]->taufriction();
      // we used to do this here, but i am trying it freemans way, and
      //  only using freeman when accelerations are made [Freeman thesis,p.136]
      // see forward_dynamics where it was moved.
    }
}










void Chain::backward_dynamics(double deltat)
{
  // Propagate dynamics from foot to hip
  int i;
  int N=links-1;
  perVector zerov3(3,ColV);
  perVector tempv6c(6,ColV);
  perVector tempv6c2(6,ColV);
  perVector tempv3a(6,ColV);
  perVector tempv3b(6,ColV);
  perMatrix temp66a(6,6);
  perMatrix temp66b(6,6);
  perMatrix temp66c(6,6);
  percision tempf;
  int footlink;

  //init any constants
  zerov3.fill(0.0);
  footlink=N+1;

  // Initialize DISTAR
  // 3.19
  link[N]->DISTAR=link[N]->DI;

  // Convert foot forces to last link
  // JESSE - note, Freemnan thesis does this wierd and manual, and omits descr.
  //  in thesis text.  The computation of normal and force components is from [Walker&Orin article 1982]
  tempv3a=(rot.DXA33(N,RI_EARTH)*link[N]->Df);
  tempv3b=JessCross(link[N]->hinge_c,tempv3a);
  link[N]->DF=JessCompose61(tempv3b,tempv3a);

  // FINALLY SET DBSTAR [3.33]
  link[N]->DBSTAR=link[N]->DB+link[N]->DF;

  // Now do the backward leg dynamics loop
  for (i=N;i>=1;i--)
    {
      // 3.29;
      fillDNmat(link[i]);
      
      // 3.28
      link[i-1]->DISTAR=(link[i-1]->DI)+rot.DXcongru(i-1,i,link[i]->DN);

      // 3.32
      link[i]->DHT=makeGamma(link[i]->DN,link[i]->Do,link[i]->Dtau,link[i]->DBSTAR,link[i]->DISTAR);

      // 3.31
      link[i-1]->DBSTAR=(link[i-1]->DB)-((rot.DX66i(i,i-1))*link[i]->DHT);

      // JESSE - adding so that contact forces can be applied at any point
      {
	tempv3a=(rot.DXA33(i-1,RI_EARTH)*link[i-1]->Df);
	tempv3b=JessCross(link[i-1]->hinge_c,tempv3a);
	link[i-1]->DF=JessCompose61(tempv3b,tempv3a);
	link[i-1]->DBSTAR+=link[i-1]->DF;
      }
    }
}




void Chain::hip_equations(double deltat)
{
  // Calculate the forces exerted on the hip at the base
  percision tempf;
  int N=links-1;

  // NOTE: freeman thesis code (p.122) has stuff different than equations,
  //       where link 0 is referred to:

  // 3.35
  // IMPORTANT - possible trouble, link 0 is hip NOT base
  link[0]->DISTAR=rot.DX66i(1,RI_BODY)*link[1]->DN*rot.DX66(1,RI_BODY);
  link[0]->DBSTAR=-1.0 * (rot.DX66i(1,RI_BODY) * link[1]->DHT);
  
  // new test - sort of a hack to propagate hip external forces
  //  (calculated earlier) to the body.  The right thing is to remove
  //  the artficial concept of a hip altogether
  link[0]->DBSTAR+=rot.DX66i(RI_HIP,RI_BODY) * link[0]->DF;
}




void Robot::iteratebase(percision deltat)
{
  // prepare base for iteration
  int i;
  Chain *chainp;

  // add forces from chains
  (base->tSUMi).fill(0.0);
  (base->tSUMb).fill(0.0);
  for (i=0;i<chains;i++)
    {
      chainp=chain[i];
      (base->tSUMi)+=(chainp->link[0])->DISTAR;
      (base->tSUMb)+=(chainp->link[0])->DBSTAR;
    }
  
  // tell base to iterate
  base->iterate(deltat);
}




void Base::iterate(percision deltat)
{
  // calc. Body Acceleration
  // Forward_kinematics and backward_dynamics has already taken place
  int i,j;
  perMatrix baserot(6,6);
  perMatrix SUMi(6,6);
  perVector SUMb(6,ColV);
  perMatrix SUMi33a(3,3);
  perMatrix SUMi33b(3,3);
  perVector SUMb3a(3,ColV);
  perVector SUMb3b(3,ColV);
  Chain *chainp;
  perVector wb(3,ColV);
  perVector tempv3c(3,ColV);
  perVector tempv3cb(3,ColV);
  perVector tempv6c(6,ColV);
  perVector basedb3(3,ColV);
  perVector zerov3(3,ColV);
  perVector Ge(3,ColV);
  perVector ab(3,ColV);
  perVector aw(3,ColV);

  //init any constants
  zerov3.fill(0.0);

  // Calculate Db(bb)
  // I moved Db (bb) to here, from forward_kinematics, as I suspect a bug
  //  And link[]->Dw changed to Dw [3.18] (modified to fix bug)
  wb(0)=Dv(0);
  wb(1)=Dv(1);
  wb(2)=Dv(2);
  
  // Compute Db (see Freeman thesis p.46)
  tempv3c=(inertia)*wb;
  basedb3=JessCross(tempv3c,wb);
  // Finally set Db
  Db=JessCompose61(basedb3,zerov3);

  // Jesse added to add external forces in earth frame to base
  //  Questionable, but I base it on methof for adding foot forces,
  //  (see Freeman thesis 1989, p.48+49)
  tempv3c=extF+userextF;
  tempv3cb=extN+userextN;
  //  we need to move the earth force to base coordinate frame,
  //  so we use the frame constructed from chain 0 rot matrix and add it to Db
  tempv6c=JessCompose61(tempv3cb,tempv3c);
  
  // New test that doesn't depend on knowing about chains
  baserot=JessmakeXi(DbAe,locationval());
  Db+=baserot*tempv6c;

  if (basetype==BASE_FIXED)
    {
      // This is for a non mobile base, we simulate the acceleration of gravity
      //  acting on the rest of the links.
      Ge(0)=0.0;  Ge(1)=0.0;  Ge(2)=homeworld->gravity();
      ab=(DbAe)*(-1.0*Ge);
      Da(3)=ab(0);  Da(4)=ab(1);  Da(5)=ab(2);
      SUMi=DI;
    }
  else
    {
      /// 3.39
      // first compute the two sums from the chains
      SUMb=tSUMb;
      SUMi=tSUMi;
      SUMb+=Db;
      SUMi+=DI;
      
      // Now get ready to solve for body accelerations
      // First apply Freman's make_sym6 which "reflects upper triangle of a 6x6 matrix to lower triangle"
      // (don't think this is nescesary any more)
      for (i=0;i<6;i++)
	{
	  for (j=i;j<6;j++)
	    SUMi(j,i)=SUMi(i,j);
	}

      if (basetype==BASE_POINTFIXED)
	{
	  // This is for a pointfixed base, which can change orientation,
	  //  but not position.  So we force the linear accel. component
	  //  to solve to zero.  This is a pair of 3x3 solves.

	  // first make the pair of 3x3s to solve
	  for (i=0;i<3;i++)
	    {
	      for (j=0;j<3;j++)
		{
		  SUMi33a(i,j)=SUMi(i,j);
		  SUMi33b(i,j)=SUMi(3+i,j);
		}
	      SUMb3a(i)=SUMb(i);
	      SUMb3b(i)=SUMb(3+i);
	    }
	  
	  // now call generalized solve routine
	  ab=Solve(SUMi33a,SUMb3a);

	  // Now set linear acceleration
	  Da(0)=ab(0);  Da(1)=ab(1);  Da(2)=ab(2);

	  //  We force the linear accel component to be the reverse gravity,
	  //   which simulates gravity on the remainder of the chain.
	  Ge(0)=0.0;  Ge(1)=0.0;  Ge(2)=homeworld->gravity();
	  ab=(DbAe)*(-1.0*Ge);
	  Da(3)=ab(0);  Da(4)=ab(1);  Da(5)=ab(2);
	}
      else
	{
	  // Run optimized cholesky_solve to find accelerations of base angles
	  Da=JessCholesky(SUMi,SUMb);
	}
    }
}













void Chain::forward_dynamics(double deltat)
{
  // propagate base accel back to chains
  // base accel has already taken place
  int i;
  int N=links-1;
  perVector aprime(6,ColV);
  perVector aprimea(6,ColV);
  percision ddq;
  perVector tempv3c(3,ColV);
  perVector tempv6c(6,ColV);
  Base *basep;
  percision tempf1,tempf2,tempf3;

  // Set a[0] to a[base]
  basep=robotp->base;
  link[0]->Da=basep->Da;
  
  for (i=1;i<=N;++i)
    {
      // 3.40
      if (i==1)
	{
	  aprime=(rot.DX66(i,RI_BODY)*basep->Da);
	  aprime+=link[i]->Do;
	}
      else
	{
	  aprime=(rot.DX66(i,i-1)*link[i-1]->Da);
	  aprime+=link[i]->Do;
	}

      // 3.41
      tempf1=1.0/link[i]->DISTAR(2,2);

      // 3/12/97 moving friction to here, [as in Freeman thesis p.136]
      tempf2=link[i]->Dtau+link[i]->taufriction()+link[i]->DBSTAR(2);

      //orig
      tempv3c=Jess3rdrow(link[i]->DISTAR);
      tempf3=tempv3c.dot(aprime);

      // now set accel. of joint
      ddq=tempf1*(tempf2-tempf3);
      
      if (link[i]->isfixed())
	ddq=0.0;

      link[i]->ddtheta_angle=ddq;
      
      // 3.42
      tempv6c.fill(0.0);
      tempv6c(2)=ddq;

      link[i]->Da=(aprime+tempv6c);
    }
}












void Base::update(double deltat)
{
  // Update velocity, position, orienatation of the base, given new accel.
  // Here we are responsible for the following:
  //  1. update Euler angle orientation of the base
  //  2. update angular&linear velocity of the base
  //  3. update the linear position in space of the base

  perVector deltaEuler(3,ColV);
  perMatrix B(3,3);
  perVector wb(3,ColV);
  perVector vb(3,ColV);
  perVector evb(3,ColV);
  percision e1,e2,e3;
  perMatrix eAb_new(3,3);
  perMatrix eAb_old(3,3);
  perMatrix bAe_new(3,3);
  perMatrix bAe_old(3,3);
  perMatrix newframeold(3,3);
  perVector Ge(3,ColV);
  perVector ab(3,ColV);
  perVector agb(3,ColV);
  perVector aw(3,ColV);
  perVectorR3 newlocation;

  if (basetype==BASE_FIXED)
    {
      // For fixed base, we don't allow it to move
      // (note that pointfixed bases are allowed to change orientation just not position).
      return;
    }

  // Set euler angles and other initializations
  e1=DEulers(0);
  e2=DEulers(1);
  e3=DEulers(2);
  wb(0)=Dv(0);
  wb(1)=Dv(1);
  wb(2)=Dv(2);
  vb(0)=Dv(3);
  vb(1)=Dv(4);
  vb(2)=Dv(5);
  aw(0)=Da(0);
  aw(1)=Da(1);
  aw(2)=Da(2);
  ab(0)=Da(3);
  ab(1)=Da(4);
  ab(2)=Da(5);

  // Compute B
  B(0,0)=0.0;  B(0,1)=sin(e3)*JessSecant(e2);  B(0,2)=cos(e3)*JessSecant(e2);
  B(1,0)=0.0;  B(1,1)=cos(e3);   B(1,2)=-1.0*sin(e3);
  B(2,0)=1.0;   B(2,1)=sin(e3)*tan(e2);  B(2,2)=cos(e3)*tan(e2);

  // *** INTEGRATE to get new euler angles
  // calculate delta Euler Angles (Freeman Thesis 1989, p.61)
  deltaEuler=B*wb;
  DEulers+=(deltat*deltaEuler);

  // Calculate new earth-body transform
  eAb_old=DeAb;
  bAe_old=DbAe;
  ComputeDeAb();
  eAb_new=DeAb;
  bAe_new=DbAe;

  // Compute frame (old body to new body) conversion matrix
  newframeold=bAe_new*eAb_old;

  // *** INTEGRATE angular velocity of body in old frame
  wb+=(deltat*aw);
  // now convert to new body frame and store
  wb=newframeold*wb;
  Dv(0)=wb(0);
  Dv(1)=wb(1);
  Dv(2)=wb(2);

  if (basetype==BASE_MOBILE)
    {
      // 4.8
      // calc. agb, the linear acceleration of body plus effects of gravity in old body frame.
      Ge(0)=0.0;
      Ge(1)=0.0;
      Ge(2)=homeworld->gravity();
      agb=ab+bAe_old*Ge;

      // *** INTEGRATE linear velocity of body (using agb) in old frame
      vb+=(deltat*agb);
      // now convert to new frame and store
      vb=newframeold*vb;

      // only strictly mobile bases can move position in space
      Dv(3)=vb(0);
      Dv(4)=vb(1);
      Dv(5)=vb(2);

      // 4.12
      // Now get the new linear velocity in ground coordinates
      evb=eAb_new*vb;
      velocity=evb;
      
      // *** INTEGRATE for body position in space
      newlocation=locationval()+(deltat*velocity);
      setlocation(newlocation);
    }

  // Store orientation
  // (we still use DbAe in dynamics, this just brings the allignment into same as rest for graphics).
  setorientation(DeAb);
}





void Chain::fillDNmat(Link *linkp)
{
  // Follow Freemans Theis (1989, p. 48) and fill the DN structure
  // 3.29
  int j,p;
  percision tempf1;

  for (j=0;j<6;++j)
    {
      for (p=0;p<6;++p)
	{
	  if (j==2 || p==2)
	    {
	      linkp->DN(j,p)=0.0;
	      linkp->DN(p,j)=0.0;
	      continue;
	    }
	  tempf1=(linkp->DISTAR(j,2)*linkp->DISTAR(2,p))/linkp->DISTAR(2,2);
	  linkp->DN(j,p)=linkp->DISTAR(j,p)-tempf1;
	}
    }
}



void Base::ComputeDeAb()
{
  // compute the body to earth transform
  
  ConvertEuler(DEulers(0),DEulers(1),DEulers(2),DeAb);
  DbAe=~DeAb;
  return;
}

































void Chain::congrutest(Link *linkp,int indexto,int indexfrom)
{
  // perform a test on the validity of our congruence transforms
  perMatrix A(3,3);
  perMatrix At(3,3);
  perVector bv(3);
  perMatrix b(3,3);
  perMatrix bt(3,3);
  perMatrix h(3,3);
  perMatrix ht(3,3);
  perMatrix Ibar(3,3);
  perMatrix Ibart(3,3);
  perMatrix M(3,3);
  perMatrix retmat1(6,6);
  perMatrix retmat2(6,6);
  perMatrix I(6,6);
  perMatrix N(6,6);
  perMatrix test33(3,3);
  perMatrix ret11(3,3);
  perMatrix ret12(3,3);
  perMatrix ret21(3,3);
  perMatrix ret22(3,3);

  // initialize
  M.fill(0.0);
  M(0,0)=linkp->kilograms;
  M(1,1)=linkp->kilograms;
  M(2,2)=linkp->kilograms;

  bv=Jessmakeci(linkp->a_length,linkp->d_offset,linkp->alpha_twist,linkp->theta_angle);
  b=JessTilde(bv);
  bt=~b;

  A=JessmakeA(linkp->a_length,linkp->d_offset,linkp->alpha_twist,linkp->theta_angle);
  At=~A;

  h=linkp->kilograms*JessTilde(linkp->hinge_d);
  ht=~h;

  I=linkp->DI;
  N=linkp->DI;

  Ibar=Jessup33(I);
  Ibart=~Ibar;

  // Now compute resulting congruence transform
  ret11=At*Ibar*A+b*At*ht*A+At*h*A*bt+b*At*M*A*bt;
  ret12=At*h*A+b*At*M*A;
  ret21=At*ht*A+At*M*A*bt;
  ret22=At*M*A;

  // now compare the 2 methods
  printf("$comparing the 2 methods:\n");
  printf("$------CALLING old method:\n");
  retmat1=rot.DXcongru(indexto,indexfrom,N);
  retmat1.print();
  printf("$------CALLING new manual method:\n");
  printf("$$$ matrix 11:\n");
  ret11.print();
  printf("$$$ matrix 12:\n");
  ret12.print();
  printf("$$$ matrix 21:\n");
  ret21.print();
  printf("$$$ matrix 22:\n");
  ret22.print();
}







