//  Jesse Reichler, January 1996
//  Fred Delcomyn
//
//  misc.h
//  This file contains miscelaneous functions not don't really belong
//   in one place.
//
//
#ifndef Misc_defined
#define Misc_defined

#include "matrixinc.h"
#include "misc.def"



// Compose a 6x6 matrix from 4 3x3 matrices (col dominant)
perMatrix JessCompose66(perMatrix m00,perMatrix m01, perMatrix m10, perMatrix m11);

// Compose a 6x1 vector from 2 3x1 vectors
perVector JessCompose61(perVector v0,perVector v1);

// Compose a 3x3 matrix from the tilde up on a 3 vector, as in Freeman thesis (1989)
perMatrix JessTilde(perVectorR3 a);
perVector JessUntilde(perMatrix mat3);

// Return a 3x3 identity matrix
perMatrix JessE33();

// Return the upper left 3x3 submatrix of a 6x6
perMatrix Jessup33(perMatrix mat66);

// Return the lower left 3x3 submatrix of a 6x6
perMatrix Jesslwr33(perMatrix mat66);

// Return the 3rd row of a 6x6 matrix
perVector Jess3rdrow(perMatrix mat66);

// Return the 3rd col of a 6x6 matrix
perVector Jess3rdcol(perMatrix mat66);

// Solve using cholesky from Freeman Thesis
perVector JessCholesky(perMatrix h,perVector b);

// Create the matrix A[i-1][i] given the 4 DH parameters
perMatrix JessmakeA(percision a_length, percision d_offset, percision alpha_twist, percision theta_angle);

// construct ci (Freeman Thesis 1989, p.54)
perVector Jessmakeci(percision a_length, percision d_offset, percision alpha_twist, percision theta_angle);

// construct  X[i][i-1] (Freeman Thesis 1989, p.27)
perMatrix JessmakeX(perMatrix A, perVector c);
perMatrix JessmakeXi(perMatrix A, perVector c);

// construct  T[i-1][i] (Freeman Thesis 1989, p.27);
perMatrix JessmakeT(perMatrix A, perVector c);

// make Gamma
perVector makeGamma(perMatrix N,perVector delta,percision tau,perVector Bstar,perMatrix Istar);

// check and make sure matrix m is symmetric
void symcheck66(perMatrix m,char *estring,int index=-99);

// normalize vector
void JessNormalize(perVectorR3 &vec);

// convert euler andles to rotation matrix
perMatrix ConvertEuler(percision e1,percision e2,percision e3,perMatrix &R);













// return sign of a float
int Jsgn(percision val);

// simple uniform random from 0 to 1
double rand01();

// return min or max value from 3
percision min3(percision a,percision b,percision c);
percision max3(percision a,percision b,percision c);

// return the secant of a value
percision JessSecant(percision x);










// Make sure a string is less than some max size
void stringcrop(char *string, int len);

// Chop off any cr at the end of a string
void crchop(char *string);

// our version of strtod
percision strnumber(char *string);

// our version of strtoupr, make a string all uppercase
void jstrupr(char *string);

// string indexing, used for config files
int stringindex(char **stringindex,char *instring);
void stringfromarray(char *string,char **stringindex);

// clean any whitespaces on front or end of a string
void cleanstring(char *string);

// My version of strdup(), since lab version of gnu c doesn't have it!
char *jstrdup(char *string);

// Is the string a valid floating point number
int isnumber(char *string);




// Fatal error handler
void fatalerr();

// experimental procedure to improve the contact impulse forces
perMatrix compute_forceinertias(perMatrix I,char *string);

// A test of euler conversion inverse
void testea(percision e1,percision e2,percision e3);



// moved below and inline
// Cross product of two 3element vectors
//perVectorR3 JessCross(perVectorR3 a, perVectorR3 b);

// Save/load vec/mat to file
//void Jvecsave(perVector &vec,FILE *moviefp, int len1);
//void Jmatsave(perMatrix &mat,FILE *moviefp,int len1,int len2);
//void Jvecload(perVector &vec,FILE *moviefp, int len1);
//void Jmatload(perMatrix &mat,FILE *moviefp,int len1,int len2);







inline perVectorR3 JessCross(perVectorR3 a, perVectorR3 b)
{
  // Our replacement, inefficient but cfpe crossproduct of 2 3dvectors
  perVectorR3 c;
  
  c(0)=(a(1)*b(2))-(a(2)*b(1));
  c(1)=(a(2)*b(0))-(a(0)*b(2));
  c(2)=(a(0)*b(1))-(a(1)*b(0));
  
  return c;
}




inline void Jvecsave(perVector &vec,FILE *moviefp,int len1)
{
  percision val;
  int count;
  
  for (count=0;count<len1;++count)
    {
      val=vec(count);
      fwrite(&val,sizeof(val),1,moviefp);
    }
}


inline void Jmatsave(perMatrix &mat,FILE *moviefp,int len1,int len2)
{
  percision val;
  int count1,count2;
  
  for (count1=0;count1<len1;++count1)
    {
      for (count2=0;count2<len2;++count2)
	{
	  val=mat(count1,count2);
	  fwrite(&val,sizeof(val),1,moviefp);
	}
    }
}


inline void Jvecload(perVector &vec,FILE *moviefp,int len1)
{
  percision val;
  int count;

  for (count=0;count<len1;++count)
    {
      fread(&val,sizeof(val),1,moviefp);
      vec(count)=val;
    }
}


inline void Jmatload(perMatrix &mat,FILE *moviefp,int len1,int len2)
{
  percision val;
  int count1,count2;

  for (count1=0;count1<len1;++count1)
    {
      for (count2=0;count2<len2;++count2)
	{
	  fread(&val,sizeof(val),1,moviefp);
	  mat(count1,count2)=val;
	}
    }
}


#endif



