// asciifilter.C
// Class methods for the ascii stream filter

#include "asciifilter.h"
#include "infactory.h"
#include "outfactory.h"
#include <cstring>



asciiStreamFilter::asciiStreamFilter(Configpp *cfpp)
  : StreamFilter(cfpp)
{
  // just initialize
  strcpy(separator," ");
  commentprefix="//";
  spanlines=false;
}



FilterReturnType asciiStreamFilter::writeheader()
{
  // Write a nice header with variable labels
  time_t curtime;
  int count;
  string desc;

  // first the opening header
  curtime=time(NULL);
  streamp->write(commentprefix+" Configpp version "+SAF_VERSIONSTRING+" using Ascii filter.\n");
  streamp->write(commentprefix+" Data output file \""+streamp->getname()+"\" created "+JRtostr(ctime(&curtime))+"\n");
  streamp->write(commentprefix+"  \n");

  // print optional user description

  desc=streamp->getdesc();
  if (desc!="")
    {
      string descline;
      streamp->write(commentprefix+" User Description:\n");
      while ((descline=JRstripline(desc))!="")
	streamp->write(commentprefix+"   "+descline+"\n");
      streamp->write(commentprefix+" \n");
    }
  
  if (dfactp)
    {
      // IF this filter is being used on a data factory, then print out a variable listing
      //now the variable list
      streamp->write(commentprefix+" Variable List\n");
      streamp->write(commentprefix+" --------------\n");
      for (count=0;count<vacs();++count)
	writeheadervar(count);
      streamp->write(commentprefix+" --------------\n");
      streamp->write(commentprefix+" \n\n\n");
    }

  return success_retval;
}



FilterReturnType asciiStreamFilter::writeheadervar(int index)
{
  // This filter is attached to a data factory
  // write a nice label for the variable/datum
  string name,label,unitstr,classname,trailstr;

  name=vac(index)->getname();
  label=vac(index)->getlabel();
  unitstr=vac(index)->getunitstr();
  trailstr=vac(index)->getobjtrail();
  classname=vac(index)->getclassname();
  if (unitstr!="")
    unitstr=" ("+unitstr+")";

  streamp->write(JRstrprintf("%s %14s    %s%s\n",commentprefix.c_str(),name.c_str(),label.c_str(),unitstr.c_str()));
  streamp->write(JRstrprintf("%s %14s     from object %s\n",commentprefix.c_str(),"",trailstr.c_str()));
  streamp->write(JRstrprintf("%s %14s     of class %s\n",commentprefix.c_str(),"",classname.c_str()));
  
  return success_retval;
}





FilterReturnType asciiStreamFilter::writetrailer()
{
  // no trailer for ascii data files
  
  return success_retval;
}




FilterReturnType asciiStreamFilter::writerow()
{
  // now write a row of dats
  int count;
  CfpVal sval;

  for (count=0;count<vacs();++count)
    {
      // display the dat
      sval=val(count);
      streamp->write(sval.get_string());
      
      if (count<vacs()-1)
	{
	  streamp->write(separator);
	}
      else
	{
	  // what to put at the end of a row
	  streamp->write("\n");
	}
    }

  // increase the rowcount
  ++rowcount;
  
  return success_retval;
}


void asciiStreamFilter::set_separator(string inseparator)
{
  // set the local char* separator (for speed)
  if (inseparator=="")
    inseparator=" ";
  if (inseparator.length()<=24)
    strcpy(separator,inseparator.c_str());
}







FilterReturnType asciiStreamFilter::readrow(int readrow)
{
  // read in a row of data and place it into the dat subindex readrow
  // be smart about ignoring comments
  int curcount,vacount;
  char iline[MAX_DF_STRLEN+1];
  char word[MAX_DF_STRLEN+1];
  FilterReturnType fretv=error_retval;

  curcount=0;
  vacount=vacs();
  do
    {
      // read in a line and remove comments
      if (streamp->ateof())
	break;
      if (!streamp->readline(iline,MAX_DF_STRLEN))
	break;
      JRcleancharp(iline,commentprefix);
      // now parse as many data elements as we can
      for (;;)
	{
	  JRstriptok(iline,word);
	  if (word[0]=='\0')
	    break;
	  if (curcount>=vacount)
	    {
	      // the line had too much stuff on it
	      streamp->error("too many items on the row of datafile.");
	    }
	  (((DatafactoryIn *)dfactp)->getivalp(curcount,readrow))->convertin(word);
	  ++curcount;
	}
      ++rowcount;
    } while (spanlines || curcount==0);

  if (curcount>0 && curcount<vacount)
    {
      // the line or file ended before all vacs/columns were found
      streamp->error("incomplete row of data encountered while reading from datafile.");
    }
  else if (curcount==0)
    {
      // we hit eof
      return end_retval;
    }

  return success_retval;
}

