// Datafactory.C


#include "datafactory.h"
#include "outfactory.h"
#include "infactory.h"
#include "JRutils.h"





// -------------------------------------------------------------------------------------
// -- Datafactory::Datafactory
// -------------------------------------------------------------------------------------

Datafactory::Datafactory(Configpp *cfpp, int insamprate)
{
  // initialize the file
  
  configppp=cfpp;
  samprate=insamprate;

  dtcounter=0;
  streamp=NULL;
  filterp=NULL;
  done=false;
}



// -------------------------------------------------------------------------------------
// -- Datafactory::~Datafactory
// -------------------------------------------------------------------------------------

Datafactory::~Datafactory()
{
  // close file and delete stream and filter

  if (!done)
    {
      // tell stream and filter to close up.
      finish();
      done=true;
    }
}


void Datafactory::start()
{
  // the filter and stream pointers should have been added, so now we do any last intitializations and checks

  if (!streamp)
    {
      // it's illegal to start a factory without a stream
      error("Datafactory started without an accompanying Datastream.\n");
      return;
    }
  if (!filterp)
    {
      // it's illegal to start a factory without a filter
      error("Datafactory started without an accompanying StreamFilter.\n");
      return;
    }

  // now we tell the filter about the stream and the factory
  filterp->setfactory(this);
  filterp->setstream(streamp);
}


void Datafactory::finish()
{
  // tell stream and filter to close up
  if (streamp)
    {
      streamp->finish();
      delete streamp;
      streamp=NULL;
    }
  if (filterp)
    {
      filterp->finish();
      delete filterp;
      filterp=NULL;
    }
  done=true;
}




void Datafactory::error(string errstring,string hintstr)
{
  // throw an error

  if (streamp)
    {
      // tell the stream to throw the error (since it can say where we are in stream)
      streamp->error(errstring,hintstr);
    }
  else
    throw CfpError(errstring,hintstr,io,false,configppp);

  done=true;
}





long Datafactory::getrowcount()
{
  return filterp->getrowcount();
};





// -------------------------------------------------------------------------------------
// StreamFilter
// -------------------------------------------------------------------------------------




StreamFilter::StreamFilter(Configpp *cfpp)
{
  // initialize the filter
  configppp=cfpp;
  dfactp=NULL;
  streamp=NULL;
  usecomments=true;
  rowcount=0;
}



void StreamFilter::start()
{
  // stream and factory pointers should be set, so check and the initialize

  if (!streamp)
    {
      // it's illegal to start a streamfilter without a stream
      error("StreamFilter started without an accompanying Datastream.\n");
      return;
    }
  if (!dfactp)
    {
      // it's illegal to start a streamfilter without a factory
      error("StreamFilter started without an accompanying DataFactory.\n");
      return;
    }
}



void StreamFilter::error(string errstring,string hintstr)
{
  // throw an error

  if (dfactp)
    {
      // tell our supervising factory to throw the error
      dfactp->error(errstring,hintstr);
    }
  else
    throw CfpError(errstring,hintstr,io,false,configppp);
}


