// swdatafactory.C

#include "cfpinc.h"
#include "swdatafactory.h"
#include "datafactory.h"
#include "datastream.h"
#include "infactory.h"
#include "outfactory.h"
#include "asciifilter.h"
#include "matlabfilter.h"




//--------------------------------------------------------------------------------
//---- SWdatafactory_obj::SWdatafactory_obj()
//--------------------------------------------------------------------------------

SWdatafactory_obj::SWdatafactory_obj(Configpp *cfpfp,CfpClass *inclassp,string oname)
  :CfpObject(cfpfp,inclassp,oname)
{
  // this is the constructor that gets called every time the user creates a new
  // instantiated object of the class.
  // This function and makenew, the next, are partners.
  dfactp=NULL;
  filterp=NULL;
}




//--------------------------------------------------------------------------------
//---- SWdatafactory_obj::makenew()
//--------------------------------------------------------------------------------

CfpObject *SWdatafactory_obj::makenew(CfpClass *inclassp,string objname)
{
  // create a new instance of ourselves, the wrapper.
  // Important: instantiating the wrapped class will come later (in function 'construct').

  return (new SWdatafactory_obj(configppp,inclassp,objname));
}






void SWdatafactory_obj::registerwrappers(CfpFactory *sfactp)
{
  // this is the constructor that gets called ONCE when the class is registered with the factory.
  CfpClass *newclassp;

  // Register Datafactory.in classes
  newclassp=sfactp->registerclass("Datafactory","Datafactory.In",this);
  newclassp->addarg("samplerate","1",true,"");
  newclassp->addarg("samplemode","step",true,"");
  newclassp->addarg("recycle","false",false,"should we restart at begining when we get to end of inputfile");
  newclassp->addarg("triggersignal","",true,"accessor to signal(timer) which activates update");

  // Register Datafactory.out classes
  newclassp=sfactp->registerclass("Datafactory","Datafactory.Out",this);
  newclassp->addarg("samplerate","1",true,"");
  newclassp->addarg("triggersignal","",true,"accessor to signal(timer) which activates update");

  // Register ascii filter classes
  newclassp=sfactp->registerclass("Datafactory","Streamfilter.Ascii",this);
  newclassp->addarg("usecomments","true",false,"","bool");
  newclassp->addarg("commentprefix","//",false,"","","string");
  newclassp->addarg("separator"," ",false,"","","string");

  // Register matlab filter classes
  newclassp=sfactp->registerclass("Datafactory","Streamfilter.Matlab",this);
  newclassp->addarg("usecomments","true",false,"","bool");

  // Register var/datum class
  newclassp=sfactp->registerclass("Datafactory","Datum.Out",this);
  newclassp->addarg("var","",false,"");
  newclassp->addarg("skip","false",false,"","bool","bool");
  newclassp=sfactp->registerclass("Datafactory","Datum.In",this);
  newclassp->addarg("var","",false,"");
  newclassp->addarg("skip","false",false,"","bool","bool");
  newclassp->addarg("samplemode","",true,"");
  newclassp=sfactp->registerclass("Datafactory","Datum.Index",this);
  newclassp->addarg("var","",false,"");
  newclassp->addarg("modulus","0",false,"","","double");
  newclassp->addarg("delay","0",false,"","","double");
}






//--------------------------------------------------------------------------------
//---- SWdatafactory_obj::destruct()
//--------------------------------------------------------------------------------

void SWdatafactory_obj::destruct()
{
  // de-instantiate the wrapped object.
  // we wrap ourselves so there is no need to delete any 'wrapped class', just set the instantiated flag to false
  instantiated=false;

  // only the factory gets killed, the others are killed by factory
  if (baseclass()=="Datafactory")
    {
      // close/end the file and then delete it
      dfactp->finish();
      delete dfactp;
    }
}
















//--------------------------------------------------------------------------------
//---- SWdatafactory_obj::construct() ------------------------------------------------
//--------------------------------------------------------------------------------

void SWdatafactory_obj::construct()
{
  // instantiate the wrapped object.  this is where we actually do late (lazy) instantiation.
  // if we are a wrapper for another class, we would allocate that target wrapped class here.
  // and save it to a local variable.  otherwise this is its own wrapper.
  // the args will have been filled by now.

  // create the wrapped object

  if (baseclass()=="Datafactory")
    construct_Datafactory();
  else if (baseclass()=="Streamfilter")
    construct_filter();
  else if (baseclass()=="Datum")
    construct_datum();
  else
    cout << "UNKNONWN CLASSA "+baseclass()+"!\n";

  // add accessors
  addaccessors();

  // set the instantiated flag
  instantiated=true;
}











//--------------------------------------------------------------------------------
//---- SWdatafactory_obj::done() ------------------------------------------------------------
//--------------------------------------------------------------------------------

CfpVal SWdatafactory_obj::done()
{
  // the object is done (ie no more children) return any return value

  if (baseclass()=="Datafactory")
    {
      // okay, now have the factory find its child stream
      Datastream *streamp= (Datastream*)((configppp->findchild(this,"Datastream",true))->getlocalp());

      //  now tell the factory about the stream
      dfactp->setstreamp(streamp);

      // the filters and datums have been added, so now tell it to start.
      // telling the factory to start will open files, etc.
      dfactp->start();
    }
  else if (baseclass()=="Streamfilter")
    ;
  else if (baseclass()=="Datum")
    ;

  return retval;
}









//--------------------------------------------------------------------------------
//---- SWdatafactory_obj::construct_Datafactory()
//--------------------------------------------------------------------------------

void SWdatafactory_obj::construct_Datafactory()
{
  // start building a Datafactory
  bool recycle;
  int samplerate;

  samplerate=(int)argval("samplerate");

  if (subclass()=="In")
    {
      // input data file
      if ((string)argval("recycle")=="true")
	recycle=true;
      else
	recycle=false;
 
      localp=dfactp=new DatafactoryIn(configppp,samplerate,recycle);
    }
  else if (subclass()=="Out")
    {
      // output data file
      localp=dfactp=new DatafactoryOut(configppp,samplerate);
    }
  else
    cout << "ERROR: Unknown subclass '"+subclass()+"' of data factory.\n";
}




//--------------------------------------------------------------------------------
//---- SWdatafactory_obj::construct_filter()
//--------------------------------------------------------------------------------

void SWdatafactory_obj::construct_filter()
{
  // build a datfile filter
  // get pointer to parent
  dfactp=(Datafactory *)((configppp->parentinstantiate(this,"Datafactory",true))->getlocalp());

  // create the appropriate filter
  StreamFilter *filterp;

  // ideally we would like to decentralize these
  if (subclass()=="Ascii")
    {
      localp=filterp=new asciiStreamFilter(configppp);
      ((asciiStreamFilter *)filterp)->set_commentprefix((string)argval("commentprefix"));
      ((asciiStreamFilter *)filterp)->set_separator((string)argval("separator"));
    }

  if (subclass()=="Matlab")
    localp=filterp=new matlabStreamFilter(configppp);

  // general args
  if (argval("usecomments")=="true")
    filterp->set_usecomments(true);
  else
    filterp->set_usecomments(false);

  // add the filter to the file
  dfactp->setfilterp(filterp);
}


//--------------------------------------------------------------------------------
//---- SWdatafactory_obj::construct_datum()
//--------------------------------------------------------------------------------

void SWdatafactory_obj::construct_datum()
{
  // build a datfile datum
  CfpVAccessor *varp;
  DatStore *dstorep;
  CfpVal val;
  bool setskip=false;
  SampleInType si;
  SampleOutType somode;
  int index;

  // get pointer to parent
  dfactp=(Datafactory *)((configppp->parentinstantiate(this,"Datafactory",true))->getlocalp());

  // get the accessor
  val=argval("var");
  varp=val.get_vaccesspointer();
  index=val.get_vaccessindex();
  
  if (subclass()=="In")
    {
      // it's an input datum
      if (dfactp->gettype()=="Out")
	throw CfpError("in an input Datafactory only Datum.in can be used (not Datum.out).","",io,false,configppp);
      if ((string)argval("samplemode")=="step")
	si=step_sampimode;
      else if ((string)argval("samplemode")=="punctuated")
	si=punctuated_sampimode;
      else
	  si=interpolate_sampimode;
      localp=dstorep=new DatInStore(varp,dfactp,si,index);
      if ((string)argval("skip")=="true")
	setskip=true;
    }
  
  else if (subclass()=="Index")
    {
      // an index datum is the one we use to index into the table, and we don't actually set in from file
      if (dfactp->gettype()=="Out")
	throw CfpError("in Datafactory.In only Datum.in can be used (not Datum.out).","",io,false,configppp);
      localp=dstorep=new DatInStore(varp,dfactp,step_sampimode,index);

      if (((DatafactoryIn*)dfactp)->getindex())
	{
	  // they have previously set an index; it's illegal to have two
	  throw CfpError("an input Datafactory can only have one index Datum.","",io,false,configppp);
	}
      ((DatafactoryIn*)dfactp)->setindex((DatInStore *)dstorep,(double)argval("modulus"),(double)argval("delay"));
      setskip=true;
    }

  else if (subclass()=="Out")
    {
      // it's an output datum
      if (dfactp->gettype()=="In")
	throw CfpError("in a Datafactory.Out only Datum.out can be used (not Datum.in)","",io,false,configppp);
      if (argeq("samplemode","current"))
	somode=current_sampomode;
      else if (argeq("samplemode","sum"))
	somode=sum_sampomode;
      else
	somode=average_sampomode;

      localp=dstorep=new DatOutStore(varp,dfactp,somode,index);
      if ((string)argval("skip")=="true")
	setskip=true;
    }

  dstorep->setskip(setskip);
      
  // add the datum to the file
  dfactp->adddat(dstorep);
}









//--------------------------------------------------------------------------------
//---- SWdatafactory_obj::addaccessors()
//--------------------------------------------------------------------------------

void SWdatafactory_obj::addaccessors()
{
  // need to install a trigger for any Datafactory, which determine when they read/write updates
  // ie for an input file, the trigger might come BEFORE any "world" updating
  //    but for an output file, the trigger might come AFTER any "world" updating
  
  if (baseclass()=="Datafactory")
    {
      // get the trigger signal that the user wants to connect to
      // note the signal must pass us the current time
      CfpVal val=argval("triggersignal");
      CfpVAccessor *triggerp=val.get_vaccesspointer();
      int index=val.get_vaccessindex();

      // now add a listener to that trigger which will call our local update signal
      Datafactory *p=(Datafactory *)localp;
      makelistener(triggerp,new CfpVFAMnotify<Datafactory>("_Datafactory_update",this,p,&Datafactory::update),-1,index);
    }

}








