// legcontroller.C

#include <stdlib.h>
#include <stdio.h>
#include <iostream.h>

#include "legcontroller.h"



extern bool NODYNAMICS;




void HCleg::initialize()
{
  // initialize the brain
  whichcyc=0;
  whereincycle=0.0;
  worldtime=0.0;
  lasttime=0.0;
  deltat=.001;
  standingmode=false;

  goal1=ang1=0;
  goal2=ang2=0;
  goal3=ang3=0;

  setconstants();
}


void HCleg::setconstants()
{
  // set cycle constants

  // normal: falls and then stands slowly
  risetime=6;
  stabletime=12;
  walktime=13.5;
  elapsedtime=.001;
  
  if (NODYNAMICS)
    {
      // test to get it moving immeditately (works only when dynamics off)
      risetime=0;
      stabletime=0;
      walktime=.1;
      elapsedtime=.1;
    }

  started=false;
  back=0.0;
  lift=.45;
  forward=.54;
  lower=.91;
  f1=-.4;
  f2=-.6;
  f3=2.10;
  fl1=.4;
  fl2=-.6;
  fl3=2.10;
  l1=.4;
  l2=-1.0;
  l3=1.90;
  lf1=-.4;
  lf2=-1.0;
  lf3=1.90;
  startcycle=.45/2.0;
}




void HCleg::setphase(float inphase)
{
  phase=inphase;
  cycleloc=inphase;
}



void HCleg::process(float curtime)
{
  // Iterate the HCleg brain instance
  float stageloc;

  worldtime=curtime;
  elapsedtime=worldtime-lasttime;
  lasttime=worldtime;

  if (!standingmode)
    {
    }
  else if (worldtime<risetime)
    {
      // up to now, let it fall on it's face (no goals)
      sang1=goal1=ang1;
      sang2=goal2=ang2;
      sang3=goal3=ang3;
      return;
    }
  else if (worldtime<=stabletime)
    {
      // from here, we make it slowly rise to standing position
      float dist=1.0-((stabletime-worldtime)/(stabletime-risetime));
      goal1=sang1+(igoal1-sang1)*dist;
      goal2=sang2+(igoal2-sang2)*dist;
      goal3=sang3+(igoal3-sang3)*dist;
      return;
    }
  else if (worldtime<walktime)
    {
      // give it time to stabilize standing up
      goal1=igoal1;
      goal2=igoal2;
      goal3=igoal3;
      return;
    }


  if (standingmode)
    {
      cycleloc+=elapsedtime/cycletime;
      while (cycleloc>1.0)
	cycleloc-=1.0;

      if (!started)
	{
	  if (worldtime<walktime)
	    {
	      // it's not time to start walking
	      return;
	    }
	  cycleloc=phase;
	  started=true;
	}
    }
  else
    {
      // if we are here then this is our first pass through and we just want to know where to start
      cycleloc=phase;
    }

  
  // now decide what functional stage we are in
  if (cycleloc>=back && cycleloc<lift)
    {
      // back swing
      stageloc=scalerange(back,lift,cycleloc);
      goal1=scaleout(f1,fl1,stageloc);
      goal2=scaleout(f2,fl2,stageloc);
      goal3=scaleout(f3,fl3,stageloc);
    }
  else if (cycleloc>=lift && cycleloc<forward)
    {
      // lift phase
      stageloc=scalerange(lift,forward,cycleloc);
      goal1=scaleout(fl1,l1,stageloc);
      goal2=scaleout(fl2,l2,stageloc);
      goal3=scaleout(fl3,l3,stageloc);
    }
  else if (cycleloc>=forward && cycleloc<lower)
    {
      // forward phase
      stageloc=scalerange(forward,lower,cycleloc);
      goal1=scaleout(l1,lf1,stageloc);
      goal2=scaleout(l2,lf2,stageloc);
      goal3=scaleout(l3,lf3,stageloc);
    }
  else if (cycleloc>=lower)
    {
      // lowerphase
      stageloc=scalerange(lower,1.0,cycleloc);
      goal1=scaleout(lf1,f1,stageloc);
      goal2=scaleout(lf2,f2,stageloc);
      goal3=scaleout(lf3,f3,stageloc);
    }

  if (!standingmode)
    {
      // memorize the starting walking position
      igoal1=goal1;
      igoal2=goal2;
      igoal3=goal3;
      goal1=ang1;
      goal2=ang2;
      goal3=ang3;
      standingmode=true;
    }

}




float HCleg::scalerange(float min,float max,float val)
{
  // a helper function, it returns a normalized (0.0 to 1.0) value
  //  indicating where in the range the value lies
  float dist;

  dist=(val-min)/(max-min);
  return dist;
}

float HCleg::scaleout(float min,float max,float dist)
{
  // a helper function.
  // Given a min and a max, and a normalized value indicating dist
  //  return a value between min and max, representing the proportional
  //  distance of 'dist' betweem the two.
  float val;

  val=min+ ( dist*(max-min) );
  return val;
}














void HCleg2::setconstants()
{
  // set cycle constants

  // normal: falls and then stands slowly
  risetime=6;
  stabletime=12;
  walktime=13.5;
  elapsedtime=.001;
  
  // fast walker
  risetime=1;
  stabletime=3;
  walktime=5;
  elapsedtime=.001;
  
  if (NODYNAMICS)
    {
      // test to get it moving immeditately (works only when dynamics off)
      risetime=0;
      stabletime=0;
      walktime=.1;
      elapsedtime=.1;
    }
  
  back=0.0;
  lift=.45;
  forward=.54;
  lower=.91;

  f1=1.5;
  fl1=1;
  l1=1;
  lf1=1.5;

  f2=-1.4;
  fl2=-1.0;
  l2=-1.5;
  lf2=-1.5;

  f3=1.50;
  fl3=1.20;
  l3=1.60;
  lf3=1.60;

  startcycle=.45/2.0;
}























void HCleg3::setconstants()
{
  // set cycle constants
  // MIDDLE LEG

  // normal: falls and then stands slowly
     risetime=6;
     stabletime=12;
     walktime=13.5;
     elapsedtime=.001;

  // fast walker
     risetime=.7;
     stabletime=3;
     walktime=6;
     elapsedtime=.001;

     if (NODYNAMICS)
       {
	 // test to get it moving immeditately (works only when dynamics off)
	 risetime=0;
	 stabletime=0;
	 walktime=.1;
	 elapsedtime=.1;
       }

  back=0.0;
  lift=.45;
  forward=.54;
  lower=.91;

  f1=1.5;
  fl1=1;
  l1=.95;
  lf1=1.4;

  f2=-1.4;
  fl2=-1.0;
  l2=-1.1;
  lf2=-1.5;

  f3=1.5;
  fl3=1;
  l3=1.1;
  lf3=1.4;

  startcycle=.45/2.0;
}




void HCleg4::setconstants()
{
  // set cycle constants
  // FRONT LEG

  // normal: falls and then stands slowly
     risetime=6;
     stabletime=12;
     walktime=13.5;
     elapsedtime=.001;

  // fast walker
     risetime=.7;
     stabletime=3;
     walktime=6;
     elapsedtime=.001;

     if (NODYNAMICS)
       {
	 // test to get it moving immeditately (works only when dynamics off)
	 risetime=0;
	 stabletime=0;
	 walktime=.1;
	 elapsedtime=.1;
       }

  back=0.0;
  lift=.45;
  forward=.54;
  lower=.91;

  f1=1.8;
  fl1=1.2;
  l1=1.1;
  lf1=1.7;

  f2=-1.4;
  fl2=-1.0;
  l2=-1.25;
  lf2=-1.45;

  f3=1.5;
  fl3=1.1;
  l3=1.2;
  lf3=1.5;

  startcycle=.45/2.0;
}












void HCleg5::setconstants()
{
  // set cycle constants
  // REAR LEG

  // normal: falls and then stands slowly
     risetime=6;
     stabletime=12;
     walktime=13.5;
     elapsedtime=.001;

  // fast walker
     risetime=.7;
     stabletime=3;
     walktime=6;
     elapsedtime=.001;

     if (NODYNAMICS)
       {
	 // test to get it moving immeditately (works only when dynamics off)
	 risetime=0;
	 stabletime=0;
	 walktime=.1;
	 elapsedtime=.1;
       }

  back=0.0;
  lift=.45;
  forward=.54;
  lower=.91;

  f1=1.3;
  fl1=.8;
  l1=.75;
  lf1=1.2;

  f2=-1.4;
  fl2=-.8;
  l2=-.85;
  lf2=-1.5;

  f3=1.5;
  fl3=.8;
  l3=.85;
  lf3=1.5;

  startcycle=.45/2.0;
}
