// SWmacro.C

#include "swmacros.h"




//--------------------------------------------------------------------------------
//---- SWmacro_obj::SWmacro_obj()
//--------------------------------------------------------------------------------

SWmacro_obj::SWmacro_obj(Configpp *cfpfp,CfpClass *inclassp,string oname)
  :CfpObject(cfpfp,inclassp,oname)
{
  // this is the constructor that gets called every time the user creates a new
  // instantiated object of the class. This function and makenew, the next, are partners.
}


//--------------------------------------------------------------------------------
//---- SWmacro_obj::makenew()
//--------------------------------------------------------------------------------

CfpObject *SWmacro_obj::makenew(CfpClass *inclassp,string objname)
{
  // create a new instance of ourselves, the wrapper.
  // Important: instantiating the wrapped class will come later (in function 'construct').

  return (new SWmacro_obj(configppp,inclassp,objname));
}



//--------------------------------------------------------------------------------
//---- SWmacro_obj()
//--------------------------------------------------------------------------------

void SWmacro_obj::registerwrappers(CfpFactory *sfactp)
{
  // this is the constructor that gets called ONCE when the class is registered with the factory.
  CfpClass *newclassp;

  // main macro class
  newclassp=sfactp->registerclass("cfpmacros","Macro",this);
  
  // now the variables and argument subclasses (when user wants to create vars and args for the macro)
  newclassp=sfactp->registerclass("cfpmacros","Macroarg",this);
  newclassp->addarg("label","a macro argument",false,"the label of the argument/variable");
  newclassp->addarg("description","",false,"the description of the argument/variable");
  newclassp->addarg("defaultvalue","",false,"the default value of the argument/variable");
  newclassp->addarg("type","",false,"the argument type (string,number,bool,pointer,etc.)");
  newclassp->addarg("range","",false,"allowable range of the argument/variable");

  newclassp=sfactp->registerclass("cfpmacros","Macrovar",this);
  newclassp->addarg("label","a macro argument",false,"the label of the argument/variable");
  newclassp->addarg("description","",false,"the description of the argument/variable");
  newclassp->addarg("defaultvalue","",false,"the default value of the argument/variable");
  newclassp->addarg("type","",false,"the argument type (string,number,bool,pointer,etc.)");
  newclassp->addarg("range","",false,"allowable range of the argument/variable");
}











//--------------------------------------------------------------------------------
//---- SWmacro_obj::destruct()
//--------------------------------------------------------------------------------

void SWmacro_obj::destruct()
{
  // de-instantiate the wrapped object and free memory

  // delete the macroclass if this is the base macro
  if (classname()=="Macro")
    delete macroclassp;
  macroclassp=NULL;

  instantiated=false;
}














//--------------------------------------------------------------------------------
//---- SWmacro_obj::construct() ------------------------------------------------
//--------------------------------------------------------------------------------

void SWmacro_obj::construct()
{
  // instantiate the wrapped object.  this is where we actually do late (lazy) instantiation.
  // if we are a wrapper for another class, we would allocate that target wrapped class here.
  // and save it to a local variable.  otherwise this is its own wrapper.
  // the args will have been filled by now.

  // create the wrapped object
  if (classname()=="Macro")
    construct_macroclass();
  else if (classname()=="Macroarg")
    construct_macroarg();
  else if (classname()=="Macrovar")
    construct_macrovar();

  // set the instantiated flag
  instantiated=true;
}


//--------------------------------------------------------------------------------
//---- SWmacro_obj::done()
//--------------------------------------------------------------------------------

CfpVal SWmacro_obj::done()
{
  // the object is done (ie no more children) return any return value

  if (classname()=="Macro")
    {
      // after a macro has been completely described, we want to pick out the codeblock
      // and then tell the macro class about it
      CfpstdInline *sip;
      sip=(CfpstdInline *)configppp->findchild(this,"Block.Code",true);
      if (!sip)
	throw CfpError("Macro "+name+" require a block of code to be specified (via Block.Code{} object).","",classerr,false,configppp);
      macroclassp->setcodeblock(sip->gulpstring());
    }
  
  return retval;
}



//--------------------------------------------------------------------------------
//---- SWmacro_obj::addaccessors()
//--------------------------------------------------------------------------------

void SWmacro_obj::addaccessors()
{
  // no accessors for a macro
}





void SWmacro_obj::construct_macroclass()
{
  // Build a macro class
  macroclassp=new CfpMacroClass(configppp,name);
}

void SWmacro_obj::construct_macroarg()
{
  // Build an argument for a macro

  // get the parent Cfpobject
  SWmacro_obj *parentmacrop=(SWmacro_obj *)(configppp->parentinstantiate(this,"Macro",true));
  macroclassp=parentmacrop->getmacroclassp();
  
  // get the parameters
  CfpVal defaultval=argval("defaultvalue","");
  string valabel=(string)argval("label");
  // now tell the macro class about the new argument
  macroclassp->addarg(name,defaultval,valabel);
}

void SWmacro_obj::construct_macrovar()
{
  // Build a variable for a macro

  // get the parent Cfpobject
  SWmacro_obj *parentmacrop=(SWmacro_obj *)(configppp->parentinstantiate(this,"Macro",true));
  macroclassp=parentmacrop->getmacroclassp();
  
  // get the parameters
  CfpVal defaultval=argval("defaultvalue","");
  string valabel=(string)argval("label");
  // now tell the macro class about the new argument
  macroclassp->addvar(name,defaultval,valabel);
}

