// cfpmovie.h
// This file defines the configpp movie classes


// forward delcarations
class cfpMovie;


#ifndef cfpmovieh_defined
#define cfpmovieh_defined

#include "cfpinc.h"

#include <iostream>
#include <time.h>
#include <sys/types.h>
#include <stdlib.h>
#include <list>
#include <string>


typedef enum MovieMode
{
  playmovie,
  recordmovie,
  invalidmovie,
  errormovie
};

typedef enum MovieCycleMode
{
  end_cycle,
  loop_cycle,
  reverse_cycle
};





//-------------------------------------------------------------------------------
//-- cfpMovie
//-------------------------------------------------------------------------------
//: this class manages saving and loading 'movies' of objects, in the loose sense
// that it is meant to save a sequence of object states.
// It's difficult to make this work in a general way with lots of different kinds of
// environments.  of particular concern is the idea that we may be creating objects
// dynamically that need to be saved.  so there is a lot of work done here to
// allow frame sizes that vary, and to at least lay the groundwork for this.
// also, for speed, we do some extra work to keep track of where neighboring frames
// start and end so we can jump around without too much trouble.


class cfpMovie
{
protected:
  MovieMode moviemode;
  MovieCycleMode cyclemode;
  string filename;
  int refreshrate;
  string usercomment;
protected:
  int refreshcount;
  FILE *filep;
  char readversionstring[255];
  list<CfpObject *>obj;
protected:
  time_t starttime;
  long framecount;
  long filestartpos;
protected:
  long curframe;
  long framepos;
  long lastframepos;
  long nextframepos;
  long firstframepos;
public:
  cfpMovie(string fname);
  ~cfpMovie();
public:
  bool loadopen();
  bool saveopen();
public:
  void addobj(CfpObject *sobj);
  void removeobj(CfpObject *sobj);
public:
  void update();
private:
  void update_load();
  void update_save();
public:
  void setrate(int newrate){refreshrate=newrate;};
  int getrate(){return refreshrate;};
  void setcomment(string incomment){usercomment=incomment;};
  string getcomment(){return usercomment;};
public:
  void gotoframe(long framenum) {skipframes(framenum-framepos);};
  bool skipframes(long frameskip);
protected:
  void write_movieheader();
  void read_movieheader();
  void write_movietrailer();
  void read_movietrailer();
  void read_frameheader();
  void write_frameheader();
  void read_frametrailer();
  void write_frametrailer();
public:
  void movieerror();
  void closefile();
public:
  virtual string writeversionstring(){return "1.0";};
};















#endif
