// pdcontroller.C

#include <stdlib.h>
#include <stdio.h>

#include "pdcontroller.h"
#include "JRutils.h"

#define JRPI  3.141592654358979
#define JRPI2 6.283185307179586



PDcontroller::PDcontroller()
{
  // default parameters
  pcomponent=0;
  dcomponent=0;
  pconstant=0;
  dconstant=0;
  output=0;
  current=goal=0;
  maxout=99999999;
  initialized=false;
  dist=0;
  multiplier=1.0;
}

void PDcontroller::initialize()
{
  // initialize using first values
  lastdist=dist;
  lastgoal=goal;
  initialized=true;
}


void PDcontroller::process()
{
  float goalchange;
  
  dist=goal-current;
  if (!initialized)
    initialize();

  if (fabs(dist)>JRPI)
    {
      // the goal and current angle must be of opposite signs, and would end up driving the arm in an ineffic. direction
      // so we point them in the more effic. direction
      if (current<0)
	current+=JRPI2;
      else
	current-=JRPI2;
      // recalc goal
      dist=goal-current;
      // now adjust lastdist
      if (fabs(dist-lastdist)>JRPI)
	{
	  if (lastdist<0)
	    lastdist+=JRPI2;
	  else
	    lastdist-=JRPI2;
	}
    }

  goalchange=goal-lastgoal;

  // PD components
  pcomponent=pconstant*dist;
  dcomponent=dconstant* (dist-lastdist);

  // Output of PD controller
  output=pcomponent+dcomponent;

  // now bound it
  if (fabs(output)>maxout)
    output=maxout*JRsgn(output);

  // now scale it with multiplier
  output*=multiplier;

  //  printf("... PDcontroller  (cur=%f goal=%f)  (P=%f D=%f)  (pcomp=%f dcomp=%f)   out=%f\n",current,goal,pconstant,dconstant,pcomponent,dcomponent,output);

  // update recordkeeping
  lastdist=dist;
  lastgoal=goal;
}








float PDcontroller::circlemod(float radval)
{
  // normalize the angle to be within -PI and +PI
  while (radval<-JRPI)
    radval+=JRPI2;
  while (radval>JRPI)
    radval-=JRPI2;
  return radval;
}
