// pointsplot.C

// Contains the derived PointsPlot class for plotting scatterplots, xyplots, etc.
//  see plot.h for header.

#include "plot.h"

#include <cmath>
#include <iostream>










//-------------------------------------------------------------------------------
//-- PointsPlot
//-------------------------------------------------------------------------------

PointsPlot::PointsPlot()
  : Plot()
{
  // constructor
  supportsinc=true;
  displaymode=scatter_display;
}



void PointsPlot::grobvidcam_draw(GrobVideoCamera *cam)
{
  // COMPLETELY redraw the plot
  Datapoint *dp;
  Grob *grb;
  int symnum;

  // first see if we need to adjust axes because a value fell out of range before the camera had a chance to redraw
  //  this is a little tricky since the axis may already be on the screen!  we assume camera will deal with this sensibly.
  //  if (needsadjust)
  //    adjustout();

  // first clear (sum axes may not be used)
  float loc[3];
  loc[0]=loc[1]=loc[2]=0.0;

  lastpoint=NULL;
  newpointer=points.begin();
  while (newpointer!=points.end())
    {
      dp=*newpointer;

      if (connect && lastpoint)
      	connectpoints(lastpoint,dp);

      dp->setloc(loc,xaxis,yaxis,zaxis);
      symnum=dp->getsourcenum();
      grb=grobsymbol[symnum];
      grb->setlocation(loc);
      cam->grobdraw(grb);

      lastpoint=dp;
      ++newpointer;
    }

  newpoints=0;
  needsredraw=false;
}




void PointsPlot::grobvidcam_incdraw(GrobVideoCamera *cam)
{
  // this is a default virtual function used to do incremental update
  Datapoint *dp;
  Grob *grb;

  // first clear (sum axes may not be used)
  float loc[3];
  loc[0]=loc[1]=loc[2]=0.0;

  while (newpointer!=points.end())
    {
      dp=*newpointer;
      if (connect && lastpoint)
	connectpoints(lastpoint,dp);

      dp->setloc(loc,xaxis,yaxis,zaxis);
      grb=grobsymbol[dp->getsourcenum()];
      grb->setlocation(loc);
      cam->grobdraw(grb);

      lastpoint=dp;

      ++newpointer;
    }

  // we succesfully did an incremental update draw
  newpoints=0;
  needsredraw=false;
}





