// swplot.C

#include "cfpinc.h"
#include "swplot.h"



//--------------------------------------------------------------------------------
//---- SWplot_obj::SWplot_obj()
//--------------------------------------------------------------------------------

SWplot_obj::SWplot_obj(Configpp *cfpfp,CfpClass *inclassp,string oname)
  :CfpObject(cfpfp,inclassp,oname)
{
  // this is the constructor that gets called every time the user creates a new
  // instantiated object of the class. This function and makenew, the next, are partners.
  plotp=NULL;
  axisp=NULL;
  psourcep=NULL;
}




//--------------------------------------------------------------------------------
//---- SWplot_obj::makenew()
//--------------------------------------------------------------------------------

CfpObject *SWplot_obj::makenew(CfpClass *inclassp,string objname)
{
  // create a new instance of ourselves, the wrapper.
  // Important: instantiating the wrapped class will come later (in function 'construct').

  return (new SWplot_obj(configppp,inclassp,objname));
}








//--------------------------------------------------------------------------------
//---- SWplot_obj()
//--------------------------------------------------------------------------------

void SWplot_obj::registerwrappers(CfpFactory *sfactp)
{
  // this is the constructor that gets called ONCE when the class is registered with the factory.
  CfpClass *newclassp;

  // Register Plot.points class (plots with indiv. datapoints, ie xyplots, 3d scatterplots)
  newclassp=sfactp->registerclass("plot","Plot.Points",this);
  addargs_plot(newclassp);

  // Register Plot.table class (plots with discrete matrix, ie barcharts, surfaceplots)
  newclassp=sfactp->registerclass("plot","Plot.Table",this);
  newclassp->addarg("storageaxis","",true,"the name of the axis to store values in (avg., sum, etc.)");
  newclassp->addarg("samplemode","sum",true,"how to treat subsampled points (sum,average,current,count)");
  newclassp->addvar("connectorstyle","3",true,"whether to draw lines connecting adjacent cells");
  addargs_plot(newclassp);

  // Register Axis class (user may make as many as they want, use names like x,y,z,color,slice)
  newclassp=sfactp->registerclass("plot","Axis",this);
  addargs_axis(newclassp);
  
  // Register Datapoint class (each plot.points has one or more datapoints, plot.matrix usually has one)
  newclassp=sfactp->registerclass("plot","Datapoint",this);
  newclassp->addarg("label","",false,"label to use on legend");
  newclassp->addarg("symbol","",true,"symbol to use for datapoint display");
  newclassp->addarg("color","",true,"color to use for symbol");

  // Register Axisval class (each Datapoint consists of multiple Axisval's, one for each axis)
  newclassp=sfactp->registerclass("plot","Axisval",this);
  newclassp->addarg("var","",false,"name of the dynamic accessor var to graph on the named axis");
  newclassp->addarg("val","",false,"numerical const. to use as axis value (var or val but not both are used)");
}






//--------------------------------------------------------------------------------
//---- SWplot_obj::destruct()
//--------------------------------------------------------------------------------

void SWplot_obj::destruct()
{
  // de-instantiate the wrapped object.
  instantiated=false;

  // kill any local, allocated object (depends what we are)
  if (baseclass()=="Plot")
    if (plotp) delete plotp;
  else if (baseclass()=="Axis")
    if (axisp) delete axisp;
  else if (baseclass()=="Datapoint")
    {
      if (psourcep)
	delete psourcep;
      if (pup)
	delete pup;
    }
  else if (baseclass()=="Axisval")
    {
      if (pval)
	delete pval;
    }
}





















//--------------------------------------------------------------------------------
//---- SWplot_obj::construct() ------------------------------------------------
//--------------------------------------------------------------------------------

void SWplot_obj::construct()
{
  // instantiate the wrapped object.  this is where we actually do late (lazy) instantiation.
  // if we are a wrapper for another class, we would allocate that target wrapped class here.
  // and save it to a local variable.  otherwise this is its own wrapper.
  // the args will have been filled by now.

  // create the wrapped object

  if (baseclass()=="Plot")
    construct_plot();
  else if (baseclass()=="Axis")
    construct_axis();
  else if (baseclass()=="Datapoint")
    construct_datapoint();
  else if (baseclass()=="Axisval")
    construct_axisval();

  // add accessors
  addaccessors();

  // set the instantiated flag
  instantiated=true;
}











//--------------------------------------------------------------------------------
//---- SWplot_obj::done()
//--------------------------------------------------------------------------------

CfpVal SWplot_obj::done()
{
  // the object is done (ie no more children) return any return value

  if (baseclass()=="Plot")
    {
      // first tell the local plot that we are done building it
      plotp->complete();
      // now complete our local pointupdaters which serve as the bridge btwn. configpp and the plots
      list<Pointupdater *>::iterator p=pups.begin();
      while (p!=pups.end())
	{
	  (*p)->complete();
	  ++p;
	}
    }
  else if (baseclass()=="Axis")
    ;
  else if (baseclass()=="Datapoint")
    ;
  else if (baseclass()=="Axisval")
    ;

  return retval;
}












//--------------------------------------------------------------------------------
//---- SWplot_obj::construct_plot()
//--------------------------------------------------------------------------------

void SWplot_obj::construct_plot()
{
  // start building a plot

  if (subclass()=="Points")
    {
      localp=plotp=new PointsPlot();
    }
  else if (subclass()=="Table")
    {
      localp=plotp=new TablePlot();
    }
}


//--------------------------------------------------------------------------------
//---- SWplot_obj::construct_axis()
//--------------------------------------------------------------------------------

void SWplot_obj::construct_axis()
{
  // build axis
  float min,max;

  // get the parent plot
  SWplot_obj *swp=(SWplot_obj *)(configppp->parentinstantiate(this,"Plot",true));
  Plot *parentplot=(Plot *)swp->getlocalp();

  // create the axis and add it to parent plot
  localp=axisp=new VisibleAxis(name,(string)argval("label"),(string)argval("enums",""));

  if (argdefined("min"))
    {
      min=argval("min");
      max=argval("max");
      axisp->setrange(min,max);
    }

  if ((string)argval("autoadjust","expand")=="shift")
    axisp->setshifter();

  parentplot->addaxis(axisp);
}


//--------------------------------------------------------------------------------
//---- SWplot_obj::construct_datapoint()
//--------------------------------------------------------------------------------

void SWplot_obj::construct_datapoint()
{
  // start building datapoint

  // get the parent plot
  SWplot_obj *swp=(SWplot_obj *)(configppp->parentinstantiate(this,"Plot",true));
  Plot *parentplot=(Plot *)swp->getlocalp();

  // create the datasource and add it to parent plot
  localp=psourcep=new PointSource((string)argval("label",name));
  parentplot->addpsource(psourcep);

  // create and add the pointupdater
  pup=new Pointupdater();
  swp->addpointup(pup);
}


//--------------------------------------------------------------------------------
//---- SWplot_obj::construct_axisval()
//--------------------------------------------------------------------------------

void SWplot_obj::construct_axisval()
{
  // start building axisval
  int axisnum;
  float constval=0;
  CfpVAccessor *vac=NULL;

  // get the parent plot
  SWplot_obj *swp=(SWplot_obj *)(configppp->parentinstantiate(this,"Plot",true));
  Plot *parentplot=(Plot *)swp->getlocalp();

  // get the parent point updater
  swp=(SWplot_obj *)(configppp->parentinstantiate(this,"Datapoint",true));
  pup=swp->pup;

  // the name of the axisval actually tells us which axis this value falls along
  axisnum=parentplot->lookupaxis(name);
  if (axisnum==-1)
    {
      // axis not found
      throw CfpError("axisval name must be a valid axis name","",syntactical,false,configppp);
    }

  // get the value (constant or dynamic var)
  if (argdefined("val"))
    constval=(float)argval("val");
  else if (argdefined("var"))
    vac=argval("var").get_vaccesspointer();

  // now create the pointval
  localp=pval=new Pointval(axisnum,vac,constval);

  // now add the pointval to the pointupdater
  pup->addpval(pval);
}







//--------------------------------------------------------------------------------
//---- SWplot_obj::addaccessors()
//--------------------------------------------------------------------------------

void SWplot_obj::addaccessors()
{
  // need to install a trigger for any plot, which determine when they read/write updates
  // ie for an input file, the trigger might come BEFORE any "world" updating
  //    but for an output file, the trigger might come AFTER any "world" updating
  CfpVAccessor *varp;

  if (baseclass()=="Plot")
    {
      Plot *p=(Plot *)getlocalp();
      // get the trigger signal that the user wants to connect to
      // note the signal must pass us the current time
      CfpVal val=argval("triggersignal");
      CfpVAccessor *triggerp=val.get_vaccesspointer();
      // now add a listener to that trigger which will call our local update signal
      makelistener(triggerp,new CfpVFAMnotify<SWplot_obj>("_plot_update",this,this,&SWplot_obj::update));

      varp=addvar(new CfpVFAMsetval<Plot,int>("colorscale",this,p,&(p->setcolorscale)));
      varp=addvar(new CfpVFAMsetval<Plot,int>("displaymode",this,p,&(p->setdisplaymode)));
      if (subclass()=="Table")
	{
	  TablePlot *pt=(TablePlot *)getlocalp();
	  varp=addvar(new CfpVFAMsetval<TablePlot,int>("connectorstyle",this,pt,&(pt->setconnectorstyle)));
	}
    }
}





//--------------------------------------------------------------------------------
//---- SWplot_obj::addargs_plot()
//--------------------------------------------------------------------------------

void SWplot_obj::addargs_plot(CfpClass *newclassp)
{
  // add the generic args for a plot
  newclassp->addarg("title","",true,"title of the plot");
  newclassp->addarg("subtitle","",true,"subtitle of the plot");
  newclassp->addarg("refreshrate","1",true,"how often to update the plot");
  newclassp->addarg("savepoints","50",true,"the number of points to keep in memory for redrawing/rescaling");
  newclassp->addarg("triggersignal","",true,"accessor to signal(timer) which activates update");
  newclassp->addarg("xaxisname","",true,"what user-named axis to use as x axis");
  newclassp->addarg("yaxisname","",true,"what user-named axis to use as y axis");
  newclassp->addarg("zaxisname","",true,"what user-named axis to use as z axis");
  newclassp->addarg("coloraxisname","",true,"what user-named axis to use for color");
  newclassp->addarg("symbolaxisname","",true,"what user-named axis to use for symbol index");
  newclassp->addvar("colorscale","1",true,"the kind of colorscale to use");
  newclassp->addvar("displaymode","0",true,"the kind of display(bars,surface) to use");
}



//--------------------------------------------------------------------------------
//---- SWplot_obj::addargs_axis()
//--------------------------------------------------------------------------------

void SWplot_obj::addargs_axis(CfpClass *newclassp)
{
  // add the arguments for an axis
  newclassp->addarg("label","",false,"label on the axis");
  newclassp->addarg("enums","",false,"optional enumerations for discrete valued axis");
  newclassp->addarg("min","",false,"minimum value on axis (origin)");
  newclassp->addarg("max","",false,"maximum value on axis");
  newclassp->addarg("minorticks","0",false,"the delta spacing for minor ticks (not labeled)");
  newclassp->addarg("majorticks","0",false,"the delta spacing for major ticks (numerical accompanying labels)");
  newclassp->addarg("autoadjust","expand",true,"when outofrange (expand,expandplus,hardshift,softshift,none)");
  newclassp->addarg("discretes","",true,"optional number of discrete slots to convert continuous value to");
  newclassp->addarg("minwall","true",true,"should we draw a wall along minimum edge of axis");
  newclassp->addarg("maxwall","false",true,"should we draw a wall along maximum edge of axis");
  newclassp->addarg("spacing",".25",true,"spacing between adjacent points (from 0 to 1, all 1 is square)");
  newclassp->addarg("size","1",false,"graphical size of axis relative to other axes (from 0 to 1)");
  newclassp->addarg("scale","linear",false,"linear or log");
  newclassp->addarg("colorscale","cool",true,"optional type of color scale to use (only used for color axes)");
  newclassp->addarg("connect","false",false,"should neighbors be connected");
}







void SWplot_obj::update()
{
  // update all updater points, then tell plot to draw
  list<Pointupdater *>::iterator p=pups.begin();

  // add all datapoints
  while (p!=pups.end())
    {
      (*p)->update(plotp);
      ++p;
    }

  // update plot
  plotp->updatedraw();
}


void SWplot_obj::addpointup(Pointupdater *pup)
{
  // add the point updater
  pups.push_back(pup);
}

























Pointval::Pointval(int inaxisnum,CfpVAccessor *invac,float inconstval)
{
  axisnum=inaxisnum;
  vac=invac;
  constval=inconstval;
}





void Pointupdater::complete()
{
  // all accessors have been added, so now reserve space for values
  datap.allocate(dims);
}

void Pointupdater::addpval(Pointval *pval)
{
  // add an accessor (dimension)
  pvals.push_back(pval);
  ++dims;
}

void Pointupdater::update(Plot *plotp)
{
  // query all point dimensions and fill in the datapoint, then add it to the chart
  pit=pvals.begin();
  Pointval *p;

  while (pit!=pvals.end())
    {
      p=*pit;
      datap.setvalat(p->getaxisnum(),p->getval());
      ++pit;
    }

  plotp->addpoint(&datap);
}






float Pointval::getval()
{
  if (vac)
    return (float)(vac->get());
  else
    return constval;
}





