/*
=============================================================================
PROGRAM sneur.c = the sneur() function
PURPOSE to generate spikes in a sensory neuron
VERSION 1998-08-28 revised 1998-09-15
PROGMER Jan Cocatre-Zilgien
LANGUAG C++ with Turbo C++ 3.0 under DOS 6.2
REMARKS works
=============================================================================
Modified to C++ by Jesse Reichler, 9/19/98
*/

#include <stdlib.h> //necessary (for rand()
#include <math.h>

#include "sneur.h"



SNEUR::SNEUR()
{
  // constructor for sneur
  defaults();
  initialize();
}

SNEUR::SNEUR(double infmax,double intau1,double intau2,double ins2,double insigma,double inregu)
{
  // constructor for sneur
  setparams(infmax,intau1,intau2,ins2,insigma,inregu);
  fmax=infmax;
  tau1=intau1;
  tau2=intau2;
  s2=ins2;
  sigma=insigma;
  regu=inregu;
  initialize();
}

void SNEUR::initialize()
{
  // clear variables
  frate=0.0;
  spikecount=0;
  deltat=.001;
}

void SNEUR::defaults()
{
  // set some default parameters
  setparams(500,.0001,.070,.0001,20,1);
}

SNEUR::~SNEUR()
{
  // destructor for sneur
}


void SNEUR::setparams(double infmax,double intau1,double intau2,double ins2,double insigma,double inregu)
{
  // constructor for sneur
  fmax=infmax;
  tau1=intau1;
  tau2=intau2;
  s2=ins2;
  sigma=insigma;
  regu=inregu;
}



double SNEUR::spikes(double S, double dt)
{
  double poisson_factor;
  double fmin;
  double df;
  double period;
  double poissonper;
  double nominalper;
  double spike;
  
  //------ catch undersampling condition
  if (dt > 2.0/fmax)
    return(-1);
  
  //------ reject negative Stimulands (in present convention)
  // AS per discussions w/Jan, 1/99-2/99, we allow negatives to be evaluated normally
  //  if (S < 0.0)
  //    return(-2);
  
  //------ calculate tonic frequency for current Suland
  if (S > s2)
    fmin = fmax * (1.0 - exp((s2-S)/sigma));
  else
    fmin = 0.0;
  
  //------ calculate instantaneous phasic-tonic frequency
  df = (fmax-f)/fmax * (S-oldS)/(tau1*dt);
  df -= (f-fmin) / tau2;
  df *= dt;
  f += df;
  
  //------ cfpeguards for extreme Sulands
  if (f > fmax)
    f = fmax;
  if (f < 0.0)
    f = 0.0;
  
  frate=f;
  
  //------ calculate time of occurrence of next spike
  if (f > 0.0)    {
    
    //------ nominal period
    nominalper = 1.0/f;
    
    //------ Poisson period
    poisson_factor = 1.0 - ((double)rand())/RAND_MAX;
    if (poisson_factor < 0.0001)
      poisson_factor =  0.0001;
    poissonper = -log(poisson_factor) * nominalper;
    
    //------ weighed average period
    period = regu * (nominalper - poissonper) + poissonper;
    
    if (period==0.0)
      frate=fmax;
    else
      frate=1/period;
    
    //------ detect occurrence of spike
    if (elapsed >= period) {
      spike = 1;
      elapsed = 0.0;
    } else {
      spike = 0;
      elapsed += dt;
    }
      
  } else {
    
    //------ neuron is silent
    spike = 0;
    elapsed += dt;
    
  }
  oldS = S;
  spikecount=spike;
  return(spike);
}



void SNEUR::reset()
{
  f = 0.0;
  oldS = 0.0;
  elapsed = 0.0;

  frate=0.0;
  spikecount=0;
}
