// swsneur.C

#include "swsneur.h"





//--------------------------------------------------------------------------------
//---- SWsneur::SWsneur()
//--------------------------------------------------------------------------------

SWsneur::SWsneur(Configpp *cfpfp,CfpClass *inclassp,string oname)
  :CfpObject(cfpfp,inclassp,oname)
{
  // this is the constructor that gets called every time the user creates a new
  // instantiated object of the class. This function and makenew, the next, are partners.
}

//--------------------------------------------------------------------------------
//---- SWsneur::makenew()
//--------------------------------------------------------------------------------

CfpObject *SWsneur::makenew(CfpClass *inclassp,string objname)
{
  // create a new instance of ourselves, the wrapper.
  // Important: instantiating the wrapped class will come later (in function 'construct').

  return (new SWsneur(configppp,inclassp,objname));
}





//--------------------------------------------------------------------------------
//---- SWsneur()
//--------------------------------------------------------------------------------

void SWsneur::registerwrappers(CfpFactory *sfactp)
{
  // this is the constructor that gets called ONCE when the class is registered with the factory.
  CfpClass *newclassp;

  newclassp=sfactp->registerclass("Sneur","Sensor.Sneur",this);
  addargs_sneur(newclassp);
}




//--------------------------------------------------------------------------------
//---- SWsneur::destruct()
//--------------------------------------------------------------------------------

void SWsneur::destruct()
{
  // de-instantiate the wrapped object.
  instantiated=false;
}














//--------------------------------------------------------------------------------
//---- SWsneur::construct() ------------------------------------------------
//--------------------------------------------------------------------------------

void SWsneur::construct()
{
  // instantiate the wrapped object.  this is where we actually do late (lazy) instantiation.
  // if we are a wrapper for another class, we would allocate that target wrapped class here.
  // and save it to a local variable.  otherwise this is its own wrapper.
  // the args will have been filled by now.

  // create the wrapped object

  if (classname()=="Sensor.Sneur")
    construct_sneur();

  // add accessors
  addaccessors();

  // set the instantiated flag
  instantiated=true;
}


//--------------------------------------------------------------------------------
//---- SWsneur::done()
//--------------------------------------------------------------------------------

CfpVal SWsneur::done()
{
  // the object is done (ie no more children) return any return value
  return retval;
}





//--------------------------------------------------------------------------------
//---- SWsneur::construct_sneur()
//--------------------------------------------------------------------------------

void SWsneur::construct_sneur()
{
  CfpVal val;
  CfpVAccessor *updateacc;

  // create a new sneur sensor
  sneurp=new SNEUR();
  lasttime=0.0;

  // use the 'stimuland' arg to tell us where to get stimuland info from
  val=argval("stimuland");
  stimulandacc=val.get_vaccesspointer();

  // we also need to build a trigger to tell us when to update the sensor
  val=argval("clock");
  updateacc=val.get_vaccesspointer();
  updateacc->addlistener(new CfpVFAMsetval<SWsneur,double>("update",this,this,&(SWsneur::update)));
}









//--------------------------------------------------------------------------------
//---- SWsneur::addargs_sneur()
//--------------------------------------------------------------------------------

void SWsneur::addargs_sneur(CfpClass *newclassp)
{
  // add the generic args for sneur
  newclassp->addarg("stimuland","",false,"accessor for getting the stimuland strength");
  newclassp->addarg("clock","",false,"accessor for getting simulated world time 9seconds)");

  newclassp->addvar("fmax","",false,"maximum firing frequency in Hz");
  newclassp->addvar("tau1","",false,"time constant for sensitivity to stimuland in s");
  newclassp->addvar("tau2","",false,"time constant for adaptation in s");
  newclassp->addvar("s2","",false,"minimum stimuland for tonic response");
  newclassp->addvar("sigma","",false,"tonic firing shape factor");
  newclassp->addvar("regu","",false,"from 0.0 (Poisson) to 1.0 (rigid pacemaker)");

  newclassp->addvar("deltat","",false,"deltat between evaluataions (in seconds)");
  newclassp->addvar("freq","",false,"current firing frequency in Hz)");
  newclassp->addvar("spikes","",false,"number of spikes occurring in last deltat interval");
}





//--------------------------------------------------------------------------------
//---- SWsneur::addaccessors()
//--------------------------------------------------------------------------------

void SWsneur::addaccessors()
{
  CfpVAccessor *varp;
  varp=addvar(new CfpVTAccessor<double>("fmax",this,&(sneurp->fmax)));
  varp=addvar(new CfpVTAccessor<double>("tau1",this,&(sneurp->tau1)));
  varp=addvar(new CfpVTAccessor<double>("tau2",this,&(sneurp->tau2)));
  varp=addvar(new CfpVTAccessor<double>("s2",this,&(sneurp->s2)));
  varp=addvar(new CfpVTAccessor<double>("sigma",this,&(sneurp->sigma)));
  varp=addvar(new CfpVTAccessor<double>("regu",this,&(sneurp->regu)));

  varp=addvar(new CfpVTAccessor<double>("deltat",this,&(sneurp->deltat)));
  varp=addvar(new CfpVTAccessor<double>("freq",this,&(sneurp->f)));
  varp=addvar(new CfpVTAccessor<double>("spikes",this,&(sneurp->spikecount)));
}



//--------------------------------------------------------------------------------
//---- SWsneur::update()
//--------------------------------------------------------------------------------

void SWsneur::update(double curtime)
{
  // this is the wrapper function which makes a well-formed call to the SNEUR function which calculates spike
  //  firing.  what we do is figure out the stimuland strength and deltat and then make the appropriate call
  //  to our wrapped SNEUR
  double deltat;
  double stimuland;

  cout << "in sneur update with curtime=" << curtime << ".\n";
  
  // calculate deltat
  deltat=curtime-lasttime;
  if (deltat==0)
    return;

  // get stimuland
  stimuland=(double)(stimulandacc->get());

  // make call
  sneurp->spikes(stimuland,deltat);

  // update lasttime
  lasttime=curtime;
}

