// cfpobject.C

#include "cfpinc.h"

//Not implemented in gcc yet (2.8.1)
//using namespace std;
//namespace Configpp{




CfpObject::CfpObject()
{
  // called ONCE per spawner, when the program loads, and caused by a global declaration.
  //  registers the spawner object with the global factory.

  name="_factory_spawning_object";
  instantiated=false;
  cfphandle_memory=true;
  cfphandle_file=true;
  configppp=NULL;
  localp=NULL;

  // now register with the factory
  classp=GlobalFactoryp()->getspawnerclassp();
}






CfpObject::CfpObject(Configpp *incfpfp,CfpClass *inclassp,string inname)
{
  // base cfpobject constructor, called every time a new user-created instance of the object is created
  configppp=incfpfp;
  name=inname;
  classp=inclassp;
  instantiated=false;
  localp=NULL;

  parentobjp=configppp->enclosingparent(this);
  // make the local argument holders
  makeargmembers();
}








void CfpObject::registerwrappers(CfpFactory *factoryp)
{
  throw CfpError("invoked the base virtual registerwrappers of object '"+prettyname()+".","",classerr,true,false);
}

CfpObject *CfpObject::makenew(CfpClass *inclassp,string objname)
{
  // instantiate a new object
  throw CfpError("invoked the base virtual makenew class of object '"+prettyname()+"'.","",classerr,true,false);
  return this;
}



string CfpObject::classname()
{
  // just return the class name
  return classp->getname();
}


string CfpObject::baseclass()
{
  // get the leader of class name (which we consider the base class)
  string str;
  str=classp->getname();
  JRstriptail(str);
  return str;
}

string CfpObject::subclass()
{
  // get the end word of class name (which we consider the sub class)
  string str;
  str=classp->getname();
  return JRstriptail(str);
}





string CfpObject::prettyname()
{
  if (name=="")
    return "anonymous obj (class "+classp->getname()+")";
  else
    return "obj "+name+" (class "+classp->getname()+")";
}

string CfpObject::prettyiname()
{
  if (name=="")
    return "anonymous";
  else
    return name;
}


string CfpObject::gettrail()
{
  if (parentobjp)
    return (parentobjp->getitrail())+prettyiname();
  else
    return prettyiname();
}

string CfpObject::getitrail()
{
  string retstr="";

  if (parentobjp)
    retstr=(parentobjp->getitrail());
  if (getname()!="" && getname()!="Configpp")
    retstr+=getname()+".";

  return retstr;
}







//--------------------------------------------------------------------------------
//---- CfpObject::addarg() -------------------------------------------------------
//--------------------------------------------------------------------------------

CfpVAccessor *CfpObject::addarg(CfpVAccessor *inaccp)
{
  // create the arg described and add it to the object
  string oname;

  oname=inaccp->getname();
  if (argmanager.find(oname)!=argmanager.end())
    throw CfpError("attempt to add argument '"+oname+"' to object '"+prettyname()+"' when an arg of that name has already been added.","",classerr,true,false);

  argmanager.insert(make_pair(oname,inaccp));
  return inaccp;
}


//--------------------------------------------------------------------------------
//---- CfpObject::addvar() -------------------------------------------------------
//--------------------------------------------------------------------------------

CfpVAccessor *CfpObject::addvar(CfpVAccessor *inaccp)
{
  // create the var described and add it to the object
  string oname;
  CfpVal defaultval;

  oname=inaccp->getname();
  if (varmanager.find(oname)!=varmanager.end())
    throw CfpError("attempt to add variable '"+oname+"' to object '"+prettyname()+"' when a var of that name has already been added.","",classerr,true,false);

  varmanager.insert(make_pair(oname,inaccp));

  if (inaccp->setable())
      {
	// if the user set up an argument with same name as accessor variable, we initialize it with the value of the argument
	if (argmanager.find(oname)!=argmanager.end())
	  {
	    defaultval=argval(oname);
	    if (defaultval.isdefined())
	      {
		if (defaultval.get_string()!="")
		  {
		    // found a class default value for the var, so set it
		    //		    cout << "...initializing the value of variable acc '"+oname+"' of object '"+getname()+"' to a found argument value '"+defaultval.get_string()+"'.\n\n";
		    inaccp->set(defaultval);
		  }
	      }
	  }
	else
	  {
	    // now, if a default value for the variable was specified in the class definition, we want to initialize accessor with default
	    CfpMemberVar *classvarp;
	    classvarp=classp->lookup_memberva(oname,true);
	    
	    //      defaultval=classvarp->getdefaultval();
	    defaultval=classvarp->getinitval();
	    
	    if (defaultval.isdefined())
	      {
		if (defaultval.get_string()!="")
		  {
		    // found a class default value for the var, so set it
		    //		    cout << "...initializing the value of variable acc '"+oname+"' to a found class default '"+defaultval.get_string()+"'.\n";
		    inaccp->set(defaultval);
		  }
	      }
	  }
      }


  return inaccp;
}


//--------------------------------------------------------------------------------
//---- CfpObject::addfunc() ------------------------------------------------------
//--------------------------------------------------------------------------------

CfpFuncAccessor *CfpObject::addfunc(CfpFuncAccessor *infaccp)
{
  // create the func described and add it to the object
  string oname;

  oname=infaccp->getname();
  funcmanager.insert(make_pair(oname,infaccp));
  return infaccp;
}


//--------------------------------------------------------------------------------
//---- CfpObject::addvarcopy()
//--------------------------------------------------------------------------------

CfpVAccessor *CfpObject::addvarcopy(string cpname,CfpVAccessor *inaccp)
{
  // create the var described and add it to the object
  // note that the accessor is created by someone else, which means we should NOT RE-delete it.

  if (varmanager.find(cpname)!=varmanager.end())
    throw CfpError("attempt to add variable '"+cpname+"' to object '"+prettyname()+"' when a var of that name has already been added.","",classerr,true,false);

  varmanager.insert(make_pair(cpname,inaccp));
  return inaccp;
}







//--------------------------------------------------------------------------------
//---- CfpObject::lookupva() -----------------------------------------------------
//--------------------------------------------------------------------------------

CfpVAccessor *CfpObject::lookupva(string inname,bool throwerr)
{
  // we want to find an appropriate argument or variable
  // we must be devious, because what we want is:
  // IF the obect is not instantiated, then we look first for a constructor argument,
  //  if we can't find an arg, and there is a matching var, then we should force the object to instantiate
  //  and then return the var.
  // IF the object IS instantiated, we look FIRST for a variable.
  CfpVAccessor *varp=NULL;
  CfpMemberVA *memberp;
  map<string, CfpVAccessor *>::iterator p;

  //  cout << "in object '" +getname()+"' asked to lookup '"+inname+"'.\n";

  // IF the obect is not instantiated, then we look first for a constructor argument.
  if (!instantiated)
    {
      p=argmanager.find(inname);
      if (p!=argmanager.end())
	{
	  // found a constructor arg
	  varp=p->second;
	  return varp;
	}
      //  if we can't find an arg, and there is a matching var, then we should force the object to instantiate
      //  and then return the var. is it an uninstantiated variable?
      memberp=classp->lookup_membervar(inname,false);
      if (memberp)
	{
	  // we found an uninstantiated variable SO what we do is force the object to instantiate
	  // then drop through to lookup now instantiated variable
	  //	  cout << "in object '" +getname()+"' asked to lookup '"+inname+"' - INSTANTIATING.\n";
	  configppp->debug("instantiating "+prettyname()+" since membervar '"+inname+"' was referenced.");
	  configppp->runconstructor(this);
	}
    }

  // object is instantiated so we look for a variable first, then an arg
  // lookup the variable
  p=varmanager.find(inname);
  if (p!=varmanager.end())
    varp=p->second;
  else
    {
      // no var, is there an arg
      p=argmanager.find(inname);
      if (p!=argmanager.end())
	varp=p->second;
    }
  
  // throw any error if appropriate
  if (!varp)
    {
      // we couldn't find the var
      //      cout << "in object '" +getname()+"' asked to lookup '"+inname+"' - RESULT WAS NEGATIVE.\n";
      if (throwerr)
	throw CfpError("Could not find member var/arg '"+inname+"' of object '"+name+"' (class "+classname()+").","",syntactical,false,false);
    }

  return varp;
}








//--------------------------------------------------------------------------------
//---- CfpObject::lookuparg() ----------------------------------------------------
//--------------------------------------------------------------------------------

CfpVAccessor *CfpObject::lookuparg(string inname,bool throwerr)
{
  CfpVAccessor *varp=NULL;
  map<string, CfpVAccessor *>::iterator p;

  p=argmanager.find(inname);
  if (p!=argmanager.end())
    varp=p->second;
  else
    {
      // we couldn't find the var
      if (throwerr)
	throw CfpError("Could not find member arguement '"+inname+"' of object '"+name+"' (class "+classname()+").","",syntactical,false,false);
    }

  return varp;
}





//--------------------------------------------------------------------------------
//---- CfpObject::lookupvar() ----------------------------------------------------
//--------------------------------------------------------------------------------

CfpVAccessor *CfpObject::lookupvar(string inname,bool throwerr)
{
  CfpVAccessor *varp=NULL;
  map<string, CfpVAccessor *>::iterator p;

  p=varmanager.find(inname);
  if (p!=varmanager.end())
    varp=p->second;
  else
    {
      // we couldn't find the var
      if (throwerr)
	throw CfpError("Could not find member variable '"+inname+"' of object '"+name+"' (class "+classname()+").","",syntactical,false,false);
    }

  return varp;
}




CfpVAccessor *CfpObject::makelistener(CfpVAccessor *sourcep,CfpVAccessor *targetp, int toindex=-1,int fromindex=-1)
{
  // attach a target listener accessor and store a pointer to it so we can delete it later
  sourcep->addlistener(targetp,fromindex,toindex);
  localaccs.push_front(targetp);
  return targetp;
}






//--------------------------------------------------------------------------------
//---- CfpObject::argdefined() ---------------------------------------------------
//--------------------------------------------------------------------------------

bool CfpObject::argdefined(string inname)
{
  CfpVAccessor *varp=lookuparg(inname,false);
  if (varp==NULL)
    return false;
  else if (!varp->isdefined())
    return false;
  return true;
}


//--------------------------------------------------------------------------------
//---- CfpObject::vardefined() ---------------------------------------------------
//--------------------------------------------------------------------------------

bool CfpObject::vardefined(string inname)
{
  if (lookupvar(inname,false)==NULL)
    return false;
  return true;
}





//--------------------------------------------------------------------------------
//---- CfpObject::argval() -------------------------------------------------------
//--------------------------------------------------------------------------------

CfpVal CfpObject::argval(string inname)
{
  CfpVAccessor *vap=lookuparg(inname,false);
  if (vap==NULL)
    {
      throw CfpError("Argument '"+inname+"' was not assigned.","",syntactical,false,false);
      return CfpVal();
    }
  else
    return vap->get();
}




//--------------------------------------------------------------------------------
//---- CfpObject::setargval() -------------------------------------------------------
//--------------------------------------------------------------------------------

void CfpObject::setargval(string inname,string valu)
{
  CfpVAccessor *vap=lookuparg(inname,false);
  if (vap==NULL)
    {
      throw CfpError("Argument '"+inname+"' was not assigned.","",syntactical,false,false);
    }
  else
    vap->set(valu);
}


//--------------------------------------------------------------------------------
//---- CfpObject::setargval() -------------------------------------------------------
//--------------------------------------------------------------------------------

void CfpObject::setargval(string inname,float valu)
{
  CfpVAccessor *vap=lookuparg(inname,false);
  if (vap==NULL)
    {
      throw CfpError("Argument '"+inname+"' was not assigned.","",syntactical,false,false);
    }
  else
    vap->set(valu);
}




//--------------------------------------------------------------------------------
//---- CfpObject::argval() -------------------------------------------------------
//--------------------------------------------------------------------------------

CfpVal CfpObject::argval(string inname,int vecindex)
{
  CfpVAccessor *vap=lookuparg(inname,false);
  if (vap==NULL)
    {
      throw CfpError("Argument '"+inname+"' was not assigned.","",syntactical,false,false);
      return CfpVal();
    }
  else
    return vap->get(vecindex);
}


//--------------------------------------------------------------------------------
//---- CfpObject::argval() -------------------------------------------------------
//--------------------------------------------------------------------------------

CfpVal CfpObject::argval(string inname,int vecindex,string defaultstr)
{
  CfpVAccessor *vap=lookuparg(inname,false);
  CfpVal val;

  if (vap==NULL)
    return CfpVal(defaultstr);

  val=vap->get(vecindex);
  if (!val.isdefined())
    val.set_string(defaultstr);
  return val;
}



//--------------------------------------------------------------------------------
//---- CfpObject::setargval() -------------------------------------------------------
//--------------------------------------------------------------------------------

void CfpObject::setargval(string inname,int vecindex,float valu)
{
  CfpVAccessor *vap=lookuparg(inname,false);
  if (vap==NULL)
    {
      throw CfpError("Argument '"+inname+"' was not assigned.","",syntactical,false,false);
    }
  else
    vap->set(valu,vecindex);
}

//--------------------------------------------------------------------------------
//---- CfpObject::setargval() -------------------------------------------------------
//--------------------------------------------------------------------------------

void CfpObject::setargval(string inname,int vecindex,string valu)
{
  CfpVAccessor *vap=lookuparg(inname,false);
  if (vap==NULL)
    {
      throw CfpError("Argument '"+inname+"' was not assigned.","",syntactical,false,false);
    }
  else
    vap->set(valu,vecindex);
}





//--------------------------------------------------------------------------------
//---- CfpObject::argval() -------------------------------------------------------
//--------------------------------------------------------------------------------

CfpVal CfpObject::argval(string inname,string defaultstr)
{
  CfpVAccessor *vap=lookuparg(inname,false);
  CfpVal val;

  if (vap==NULL)
    return CfpVal(defaultstr);

  val=vap->get();
  if (!val.isdefined())
    val.set_string(defaultstr);
  return val;
}





//--------------------------------------------------------------------------------
//---- CfpObject::varval() -------------------------------------------------------
//--------------------------------------------------------------------------------

CfpVal CfpObject::varval(string inname)
{
  CfpVAccessor *vap=lookupvar(inname,false);
  if (vap==NULL)
    {
      throw CfpError("Variable '"+inname+"' was not assigned.","",syntactical,false,false);
      return CfpVal();
    }
  else
    return vap->get();
}



//--------------------------------------------------------------------------------
//---- CfpObject::varval() -------------------------------------------------------
//--------------------------------------------------------------------------------

CfpVal CfpObject::varval(string inname,string defaultstr)
{
  CfpVAccessor *vap=lookupvar(inname,false);
  CfpVal val;

  if (vap==NULL)
    return CfpVal(defaultstr);

  val=vap->get();
  if (!val.isdefined())
      val.set_string(defaultstr);

  return val;
}









//--------------------------------------------------------------------------------
//---- CfpObject::argeq(string inname,string valstring)
//--------------------------------------------------------------------------------

bool CfpObject::argeq(string inname,string valstring)
{
  if ((string)argval(inname,"")==valstring)
    return true;
  else
    return false;
}

bool CfpObject::argeq(string inname,double valnum)
{
  if ((double)argval(inname,"")==valnum)
    return true;
  else
    return false;
}











//--------------------------------------------------------------------------------
//---- CfpObject::lookupfunc() ---------------------------------------------------
//--------------------------------------------------------------------------------

CfpFuncAccessor *CfpObject::lookupfunc(string inname,bool throwerr)
{
  CfpFuncAccessor *funcp=NULL;
  CfpMemberFunc *memberp;
  map<string, CfpFuncAccessor *>::iterator p;

  if (!instantiated)
    {
      // member functions are not accessible if the object is not yet instantiated, so if it is
      // a member func but the object is not instantiated, instantiate it.
      memberp=classp->lookup_memberfunc(inname,"",false);
      if (memberp)
	{
	  // we found an uninstantiated variable SO what we do is force the object to instantiate
	  // then drop through to lookup now instantiated variable
	  configppp->debug("instantiating "+prettyname()+" since memberfunc '"+inname+"' was referenced.");
	  configppp->runconstructor(this);
	}
    }
      
  p=funcmanager.find(inname);
  if (p!=funcmanager.end())
    funcp=p->second;
  else
    {
      // we couldn't find the func
      if (throwerr)
	throw CfpError("Could not find member function '"+inname+"' of object '"+name+"' (class "+classname()+").","",syntactical,false,false);
    }

  return funcp;
}







//--------------------------------------------------------------------------------
//---- CfpObject::invokef() ------------------------------------------------------
//--------------------------------------------------------------------------------

CfpVal CfpObject::invokef(string fname,string argstr,string keyword,map<string,CfpVal *> *argmap)
{
  // this is the catch-all, manual function caller
}


//--------------------------------------------------------------------------------
//---- CfpObject::construct() ----------------------------------------------------
//--------------------------------------------------------------------------------

void CfpObject::construct(string argstr,string keyword,map<string,CfpVal *> *argmap)
{
  // this is where the class object is *forced* to construct an instantiation of the local (wrapped) object
  // *if* it has not already done so.  ie this function supports lazy instantiation, where the wrapped object
  //  does not have to be instantiated until "enough" of it's arguments have been filled, rather than
  //  when the interpreter parses somtehing like "Robot r2d2".
  // This function gets called if the class has registered constructor functions.
}


//--------------------------------------------------------------------------------
//---- CfpObject::construct() ----------------------------------------------------
//--------------------------------------------------------------------------------

void CfpObject::construct()
{
  // this is where the class object is *forced* to construct an instantiation of the local (wrapped) object
  // *if* it has not already done so.  this version is called along with the above, in sequence, as an alternative.
}






//--------------------------------------------------------------------------------
//---- CfpObject::destruct() -----------------------------------------------------
//--------------------------------------------------------------------------------

CfpAdoptType CfpObject::adoptionquery(CfpObject *objp)
{
  // will we adopt the child? we may return:
  //  errorAdopt if it is an error that child is trying to be adopted to us
  //  rejectAdopt if we don't want to adopt the child but it's not an error
  //  acceptAdopt if we are adopting the child
  // 
  // generally, if a child you are not expecting asks for adoption you just reject it (and not return error).

  // base virtual class just returns rejectAdopt
  return rejectAdopt;
}




//--------------------------------------------------------------------------------
//---- CfpObject::isa() ----------------------------------------------------------
//--------------------------------------------------------------------------------

bool CfpObject::isa(string typestr)
{
  // is the object a derivitive of typestr?
  return classp->isa(typestr);
}


//--------------------------------------------------------------------------------
//---- CfpObject::hasa() ---------------------------------------------------------
//--------------------------------------------------------------------------------

bool CfpObject::hasa(string typenam)
{
  // does the object have a child queried?
  if (hasobj(typenam))
    return true;
  return false;
}


//--------------------------------------------------------------------------------
//---- CfpObject::hasobj() ---------------------------------------------------------
//--------------------------------------------------------------------------------

CfpObject *CfpObject::hasobj(string typenam)
{
  // does the object have a child queried?
  // go through all children and find the first that matches
  CfpObject *objp;
  list<CfpObject *>::iterator p;

  p=childmanager.begin();
  while (p!=childmanager.end())
    {
      if ((*p)->isa(typenam))
	  return *p;
      ++p;
    }
  
}






//--------------------------------------------------------------------------------
//---- CfpObject::makeargmembers() -----------------------------------------------
//--------------------------------------------------------------------------------

void CfpObject::makeargmembers()
{
  // this function is automatically called by the base constructor, whenever a new, user-created
  //  instance of a derived object is created.
  // the job of this function is to allocate and register the local member accessors for
  //  arguments.  this is nesc. since, unlike member variables, argument memory is managed
  //  automatically by the class.
  map<string,CfpMemberVar *>::iterator p;
  CfpMemberVar *memberp;
  CfpVal *valp;
  string ivalue;

  // go through each arg of the *class* and allocate a CfpVal accessor for the *object*
  p=(classp->argmanager).begin();
  while (p!=(classp->argmanager).end())
    {
      memberp=p->second;
      valp=new CfpVal(memberp->getdefaultval());

      if (memberp->getvectorsize()==1)
	addarg(new CfpValAccessor(memberp->getname(),this,valp));
      else
	{
	  // create a vector accessor and initialize using a string
	  if (valp->isdefined())
	    addarg(new CfpVecAccessor<CfpVal>(memberp->getname(),this,memberp->getvectorsize(),(string)(*valp)));
	  else
	    addarg(new CfpVecAccessor<CfpVal>(memberp->getname(),this,memberp->getvectorsize(),""));
	}

      ++p;
    }
}







//}
// Namespace close-brace
