// cfpparser.C
// This file (along with configpp.C, cfpcomline.C) implements member fucntions for the Configpp class.
// It contains functions directly related to parsing.

#include "cfpinc.h"

//Not implemented in gcc yet (2.8.1)
//using namespace std;
//namespace Configpp{





//--------------------------------------------------------------------------------
//---- Configpp::parse() ----------------------------------------------------------
//--------------------------------------------------------------------------------

bool Configpp::parse()
{
  // Parse the input streams.
  // Catch any parsing error exceptions and print them.
  // Return the number of syntax errors encountered
  debug("begining parse");
  bool stoponone=true;

  for (;;)
    {
      try
	{
	  // read a statement, break if EOF
	  if (!Parsestatement())
	    break;
	}
      catch (CfpError cfperror)
	{
	  cfperror.report(this);

	  if (cfperror.isfatal() || stoponone)
	    {
	      // it's a fatal error so we have to stop parsing
	      throw cfperror;
	    }
	}
      catch (...)
	{
	  error("unexpected exception encountered while parsing inputs.","",internal,true);
	  return false;
	}
    }
 
  if (errors>0)
    return false;
  else
    return true;
}






//--------------------------------------------------------------------------------
//---- Configpp::parseconsume() ---------------------------------------------------
//--------------------------------------------------------------------------------

CfpTokenType Configpp::parseconsume(bool retcr)
{
  // consume tokens until the end of the current (erroneous statement)
  // this is useful for continuing syntax cheking after we hit an error.
  // if retcr is true, then return after CR
  CfpToken lextoken;
  CfpTokenType ltype;
  int bracedepth=0;


  lextoken=*(lexerp->getcurtokenp());
  if (lextoken.gettype()==openbraceSym)
    {
      // if the last token (which resulted in an error is an open brace, we don't want to
      // treat it like it is part of the recovery.
      lextoken=lexerp->readtoken(tokencrMode,true);
    }

  for (;;)
    {
      ltype=lextoken.gettype();
      if (ltype==eofSym)
	{
	  // it is considered an error if we hit eof while trying to consume/recover
	  // but note that this error will never see the light of day, as the initial
	  // error will shadow it (as it should).
	  throw CfpError("unexpected end of input during error recovery.","",syntactical,true,this);
	}
      if (bracedepth==0 && ltype==semicolonSym)
	{
	  return ltype;
	  break;
	}
      else if (bracedepth==0 && ltype==crSym && retcr)
	{
	  return ltype;
	  break;
	}
      else if (bracedepth==0 && ltype==closebraceSym)
	{
	  return closebraceSym;
	  break;
	}
      if (ltype==openbraceSym)
	{
	  // do we really want to consume over this?
	  // might be nicer if we let the parser try to go into this block to check
	  //  for syntax errors...
	  ++bracedepth;
	}
      else if (ltype==closebraceSym)
	{
	  --bracedepth;
	  if (bracedepth==0)
	    return semicolonSym;
	}
      lextoken=lexerp->readtoken(tokencrMode,true);
    }

  // we should not ever end up here
  throw CfpError("confused by earlier error.","",internal,true,this);
}







//--------------------------------------------------------------------------------
//---- Configpp::Parsestatement() -------------------------------------------------
//--------------------------------------------------------------------------------

bool Configpp::Parsestatement()
{
  // parse a single statement or block of statements within braces
  // Return TRUE if we found one to parse, FALSE if we hit EOF with no statments (ie we're done).
  // we let our parent catch any errors
  CfpTokenType ttype;

  // call internalparsestatement, with a parameter which says we are at root node,
  //  any root level close braces are errors.
  ttype=internalParsestatement(false);

  // now if we get back eof, we return false, otherwise true
  if (ttype==eofSym)
    return false;
  else
    return true;
}




//--------------------------------------------------------------------------------
//---- Configpp::internalParsestatement() -----------------------------------------
//--------------------------------------------------------------------------------

CfpTokenType Configpp::internalParsestatement(bool retclosebrace)
{
  // parse a single statement or block of statements withing braces
  // this internal version will recurse to parse blocks.
  // RETURN semicolonSym on no-errors, succesful parse of a statement
  // RETURN eofSym when there are were no statements to read (and no error)
  // RETURN closebraceSym IF we encounter it as the start of a statement, AND if it is allowed via calling arg.
  // Display any errors you detect.  ALSO throw them back to parent.
  // note: on encountering a non-fatal error, this func should still try to read
  //  to the end of a statement, ie it should consume tokens until the begining of
  //  a new statement, and then re-throw the error to caller.
  CfpToken lextoken;
  CfpTokenType ttype;
  string sval;

  // Everything we do with regard to parsing a statement is done inside a try block.
  try
    {
      // get the leading token
      lextoken=lexerp->readtoken(tokenMode);
      Hintloc=*(lexerp->getlocp());
      ttype=lextoken.gettype();
      sval=lextoken.getstr();

      // handle the initial token
      switch(ttype)
	{
	case alphaVal:
	  // A normal start to a statement
	  Parsealpha(lextoken);
	  break;
	case poundVal:
	  // a pound directive
	  Parsepound(lextoken);
	  break;

	// META-PARSING
	case semicolonSym:
	  // okay a semicolon is a valid (if empty) statement
	  break;
	case eofSym:
	  // there's nothing left to get.  it's not an error, but we return false to say we are done.
	  return eofSym;
	  break;
	case openbraceSym:
	  // we got the start of a block, so we want to recursively parse all statements in the block
	  Parseblock();
	  ttype=closebraceSym;
	  break;
	case closebraceSym:
	  if (retclosebrace==false)
	    {
	      // we were told not to allow close braces (ie we are at root level), so this is an error.
	      throw CfpError("mismatched close brace '}'.","",syntactical,false,true,&Hintloc);
	    }
	  else
	    {
	      // it's okay to return a close brace (presumably we are in an open brace)
	      return closebraceSym;
	    }
	  break;
	  
	// ERRORS
	case numericVal:
	  throw CfpError("expected the start of a statement but found a number.","",syntactical,false,true,&Hintloc);
	  break;
	case stringVal:
	case squotestringVal:
	  throw CfpError("expected the start of a statement but found a string.","",syntactical,false,true,&Hintloc);
	  break;
	case mathVal:
	case conditionVal:
	case commaSym:
	case openparenSym:
	case closeparenSym:
	case opensquareSym:
	case closesquareSym:
	case otherSym:
	  // this is an error, but we first check to see if we think they meant to type a comment.
	  if (sval[0]=='/')
	    throw CfpError("unexpected token at start of statement.","was a comment intended? if so, single line comments begin with //, comment blocks are enclosed in /* and */.",syntactical,false,true,&Hintloc);
	  else
	    throw CfpError("a statement can never begin with this symbol.","",syntactical,false,true,&Hintloc);
	  break;
	default:
	  // we should never end up here
	  throw CfpError("unexpected token at start of statement.","",internal,true,true,&Hintloc);
	  break;
	}
    }

  catch (CfpError cfperror)
    {
      // we caught an error while parsing a statement.  This is where nearly every parse error displays
      // First we display the error.
      // Next we swallow tokens until we eat to the end of the current statement.
      // Note we do something fun here.  Conceivably, there could be an error while
      //  trying to consume tokens.  if so, we don't want our real error to be forgotten,
      //  so what we do is we pretend we never got the second (parseconsume) error, as
      //  it's probably caused by the first, and INSTEAD insure that we stop parsing after the
      //  first error is, by setting its fatal flag.

      // report the error
      processerror(&cfperror);

      // now figure out how to recover from it as best we can, to prepare for the next statement parse.
      if (cfperror.isfatal())
	throw cfperror;
      try
	{
	  // it's not fatal, so try to skip over the rest of the statement and then rethrow it nonfatally.
	  // (that way the caller can try to parse the next statement).
	  if (ttype==poundVal)
	    {
	      // for pound directives, a CR signifies the end of consumption
	      ttype=parseconsume(true);
	    }
	  else
	    {
	      // consume the rest of the statement (which may be multiple lines if this statement
	      // is the begining of an object instantitation).
	      ttype=parseconsume(false);
	    }
	  return ttype;
	}
      catch (...)
	{
	  // if we get an error while trying to skip over the rest of the statement, oy vey it's a doozy,
	  // so we indicate that the original error, while not initially considered fatal, is now
	  //  considered fatal.
	  cfperror.setfatal();
	  // Re-throw the (modified, so don't use "throw;") error to our caller.
	  throw cfperror;
	}
    }

  // okay we parsed a statement succesfully.  return SemicolonSym;
  return semicolonSym;
}




//--------------------------------------------------------------------------------
//---- Configpp::Parseblock() -----------------------------------------------------
//--------------------------------------------------------------------------------

void Configpp::Parseblock()
{
  CfpTokenType ttype;
  do
    {
      try
	{
	  // this is a little tricky, a try block inside of a try block...
	  // we want to catch errors that our recursive self generated, rather than letting
	  // the stack unwind, because we want to continute parsing in correct brace depth
	  //  if possible.
	  ttype=internalParsestatement(true);
	}
      catch (CfpError cfperror)
	{
	  cfperror.report(this);
	  // we got an error inside of a block. we don't want to exit immed. if it's
	  // not a fatal error.
	  if (cfperror.isfatal())
	    throw cfperror;
	}
    } while (ttype!=eofSym && ttype!=closebraceSym);
  if (ttype==eofSym)
    {
      // we hit eof in the middle of a block.  that's a fatal error.
      // set the flag of our token which signified the end of the block to extend to eof, then throw it.
      //      lextoken.seteoflag();
      throw CfpError("unexpected end of input inside a statement block.","",syntactical,true,false);
    }
}





//--------------------------------------------------------------------------------
//---- Configpp::Parsepound() -----------------------------------------------------
//--------------------------------------------------------------------------------

void Configpp::Parsepound(CfpToken stoke)
{
  // Parse a pound directive statement that begins with token 'stoke'

  // a pound directive in Configpp is really just a shorthand way of saying something.
  // Pound directives take the form
  //	#keyword\n
  // or #keyword argword\n
  // or #keyword argword remainderstring\n
  // This syntax is a holdover from the C preprocessor.  It is a bad thing, BUT
  //  we retain it for historical/habitual reasons.
  // The way we handle it is to CONVERT any pound directive into an equivelant
  //  normal statement.
  // Note that no comments are allowed after the keyword.  everything after the CR is part of the statement.
  string keywordstr;
  string argwordstr;
  string remainderstr;
  CfpToken temptoke;
  CfpObject *objp;
  CfpLoc temploc;

  // get the pound directive keyword
  keywordstr=(stoke.getstr()).substr(1);

  // parse the rest of the line (could be blank)
  temptoke=lexerp->readtoken(wordMode);
  if (temptoke.gettype()==eofSym)
    {
      // yikes, we got an eof while trying to read the first arg!
      throw CfpError("unexpected end of input while reading a pound ('#') directive","",syntactical,true,this);
    }
  else
    argwordstr=temptoke.getstr();
  temptoke=lexerp->readtoken(crlineMode);
  if (temptoke.gettype()==eofSym)
    {
      // yikes, we got an eof while trying to read the first arg!
      throw CfpError("unexpected end of input while reading a pound ('#') directive","",syntactical,true,this);
    }
  else
    remainderstr=temptoke.getstr();

  if (keywordstr.length()==0)
    {
      // It's a blank pound directive, which is an error. tell them so, and suggest that
      // maybe they intended a comment (as in perl, shellscripts, etc. which use # for comments)
      // or that maybe they just added a space by accident.
      throw CfpError("pound ('#') directive found with no keyword specified (should be \"#keyword options\").",
		     "was a comment intended? if so, single-line comments begin with // not #.",
		     syntactical,false,true,stoke.getlocp());
    }

  if (keywordstr[0]=='!')
    {
      // This is a shell script meta-control statement, used to tell the script to run
      // the file.  (ie the first line of a file might begin #!/usr/bin/configpp) which
      // would be instructions for the shell to run the configpp interpreter on the file.
      // Therefore, we ignore it.
      debug("skipping over pound directive which appears to be a shell command.");
      return;
    }

  if (!Allowpounds)
    {
      // parser is configured to not accept pound directives, for cleaner more uniform syntax
      throw CfpError("interpreter is configured to disallow pound ('#') directives.",
		     "",syntactical,false,true,stoke.getlocp());
    }

  // Now convert the shorthand into longhand
  // Keyword must be a class name, and argword must be a variable name.
  debug("processing pound directive [" +keywordstr+ "] [" +argwordstr+ "] [" +remainderstr+ "].");

  // First, substitute any aliases (as we do with all identifiers)
  keywordstr=alias(keywordstr);

  // Now invoke the proper conversion call
  if (remainderstr.length()==0)
    {
      // Given
      //	#keyword argword\n
      // We pretend they typed in the Configpp code:
      //	keyword
      //	  value="argword";
      // (there must be a member variable named value, otherwise it's an error)
      // Also, set the error hint to the keyword token location, then restore afterwards
      temploc=*(gethintlocp()); sethintloc(&stoke);
      objp=createObject(keywordstr);
      sethintloc(&temploc);
      procAssignment("value",argwordstr);
      completeObject(objp);
    }
  else
    {
      // Given
      //	#keyword argword remainderstring\n
      // We pretend they typed in the Configpp code:
      //	keyword
      //	  argword="remainderstring";
      // Also, set the error hint to the keyword token location, then restore afterwards
      temploc=*(gethintlocp()); sethintloc(&stoke);
      objp=createObject(keywordstr);
      sethintloc(&temploc);
      procAssignment(argwordstr,"\""+remainderstr+"\"");
      completeObject(objp);
    }
}




//--------------------------------------------------------------------------------
//---- Configpp::Parsealpha() -----------------------------------------------------
//--------------------------------------------------------------------------------

void Configpp::Parsealpha(CfpToken stoke)
{
  // Parse a statement that begins with token 'stoke' which is a word
  // It's either:
  //	1) object/variable declaration
  //	2) variable/object assignment
  //	3) function call
  CfpClass *classp=NULL;
  CfpObject *objp=NULL;
  CfpVAccessor *varp=NULL;
  CfpFuncAccessor *funcp=NULL;
  CfpToken toke;
  string identifier;
  string orig_identifier;
  string leaderstr;
  string vectoradstring="";

  orig_identifier=stoke.getstr();
  identifier=alias(orig_identifier);

  //  cout << "... begining parsealpha with '"+orig_identifier+"'.\n";

  // first we check if it's a function call, indicated by an open paren after the identifier
  toke=lexerp->peektoken(tokenMode);
  if (toke.gettype()==openparenSym)
    {
      // okay it's a function call.  parse it, put return value in global return value
      GlobalReturnValue=Parsefunc(identifier);
      // now there better be a semicolon afterwards
      Parsesemicolon();
      return;
    }

  // is it a declartion (signified by the first word being a class name)?
  classp=factoryp->lookupClass(identifier,false);
  if (classp)
    {
      // okay, the statement is an object declaration
      Parsedeclaration(identifier,classp);
      return;
    }


  // it must be an assignment

  // first see if the whole identifier is an object trail (ie no accompanying var or function name)
  objp=lookupObjtrail(identifier,false);
  objp=NULL; // we no longer do this, in case there is a variable which shares its name with an object (not an error)

  if (objp)
    {
      // when no variable name is specified we might assume they are talking about a default var
      throw CfpError("object '" +identifier+"' specified without accompanying member variable or function name.","",syntactical,false,true,stoke.getlocp());
    }
  else
    {
      // see if there is an explicit object trail leader
      leaderstr=JRstripleader(identifier);
      if (leaderstr!="")
	{
	  // okay we have an explicit trail
	  objp=lookupObjtrail(leaderstr,true);
	}
    }

  // is it a member argument or variable
  if (objp)
    varp=objp->lookupva(identifier,false);
  else
    varp=lookupVarArg(identifier,false);

  if (varp)
    {
      // okay, we think we have a variable assignment

      // is a specific vector element being specified?
      toke=lexerp->peektoken(tokenMode);
      if (toke.gettype()==opensquareSym)
	{
	  // yes they specified a specific vector, so read it
	  vectoradstring=Parsevectoradstring();
	  toke=lexerp->peektoken(tokenMode);
	}

      // First, let's make sure we have an assignment by consuming the '='
      if (toke.getstr()=="(")
	throw CfpError("expected assignment operator '=' because '"+identifier+"' is a variable name not a function name.","",syntactical,false,true,toke.getlocp());
      else if (toke.getstr()=="=")
	{
	  // consume the '='
	  toke=lexerp->readtoken(tokenMode);
	}
      else
	{
	  if (!allow_shorthandassignments())
	    {
	      // it used to be that we would consider this an error and print the following message:
	      throw CfpError("expected assignment operator '=' for '"+identifier+"' which is a variable.","",syntactical,false,true,toke.getlocp());
	    }
	  // but now we allow for the shorthand assignmnet "age 23" to stand in for "age=23";
	}
      Parseassignment(varp,vectoradstring);
      return;
    }

  // it's an error.  just for fun let's see if it's a function call that doesn't have an open paren.
  if (objp)
    funcp=objp->lookupfunc(identifier,false);
  else
    funcp=lookupFunction(identifier,false);
  if (funcp)
    {
      // okay, we have a function call
      // First, let's make sure we have an assignment by consuming the '('
      toke=lexerp->readtoken(tokenMode);
      if (toke.getstr()=="=")
	throw CfpError("expected parameters for a function call because '"+identifier+"' is a function name not a variable name.","",syntactical,false,true,toke.getlocp());
      else
	throw CfpError("expected parameters for '"+identifier+"' which was thought to be a function call.","",syntactical,false,true,toke.getlocp());
    }



  
  // Okay, we don't know what the hell it is (probably a spelling error)
  if (objp)
    throw CfpError("unrecognized member '" + identifier + "' of "+objp->prettyname()+".",namehint(orig_identifier),syntactical,false,true,stoke.getlocp());
  else if (identifier==orig_identifier)
    throw CfpError("unrecognized identifier '" + orig_identifier + "'; not a registered class, variable, or function name.",namehint(orig_identifier),syntactical,false,true,stoke.getlocp());
  else
    throw CfpError("Unrecognized identifier '" + identifier + "'; not a registered class, variable, or function name.",namehint(orig_identifier)+"note: '"+orig_identifier+"' was assumed to be an alias for '"+identifier+"'.",syntactical,false,true,stoke.getlocp());
}




//--------------------------------------------------------------------------------
//---- Configpp::Parsedeclartion() ------------------------------------------------
//--------------------------------------------------------------------------------

void Configpp::Parsedeclaration(string classname,CfpClass *classp)
{
  // parse an object declaration
  CfpToken toke;
  CfpObject *objp;
  CfpObject *varp;
  string objname,word;
  CfpMember *memberp;
  string pparse;

  // this is a little tricky because there are different ways that an object
  // declaration can be made:
  //	1. Classname Objectname Statementblock
  //	2. Classname Objectname Statement
  //	3. Classname Statementblock
  //	4. Classname Statement
  //	5. Classname Objectname=val;
  // one reason it's tricky is becase it may be difficult to distinguish #4 from #1&#2
  // and even more troublesome if we want to allow #5, as in the traditional variable (object) declaration.
  // ex.  RobotClass age = 23;		// #4
  // ex.  RobotClass r2d2 age = 23;	// #2
  // ex.  int age=23;			// #5
  // Note that the examples for #4 and #5 seem synactically equivelant.
  //  (in ex #4, age is a variable of RobotClass, and in #1, age IS the name of the object(variable)
  //  being declared).
  // How are we to handle this dilemna?
  // First, we will say that cases #1,2,5 will be treated as shortcuts, such that
  //  passing the Objectname could be done by saying within a statement block: name=Objectname;
  //  using an assignment on the declaration line is done by saying within a statement block: value=val;
  // So that all object declarations can be thought of as taking the form:
  //  Classname {name=xxx; value=xxx; more_statements...}
  // Of course the problem remains as to how to disambiguate the above cases...
  // Here's how we will do it.  we take a look at the first word after the class name.
  //  If this word is a local variable name or another class name, then we assume that the object has no name.
  //  a la example #4.  if not, then we know it is the name of the object.
  // if we are feeling generous then we might check for an '=' after the word if we guess it is a local
  //  variable, and if we find instead another keyword (as in ex #2), then we know the first keyword
  //  was supposed to be an object name.
  // So the solution simply means that no object may have a name which is the same as one of it's local variables
  //  or the same as a class name.
  // Here is the clean, simple way to restrict our syntax:
  //  In order to use a single-line statement in an object declaration, ie #2 & #4, the single statement
  //   is restricted to a local variable assignment.  otherwise statements must be embedded in braces.

  // no name object
  objname="";

  //  cout << "... Starting a parse declaration of class "+classname+".\n";

  // PEEK the next token
  toke=lexerp->peektoken(tokenMode);
  word=toke.getstr();

  if (toke.gettype()==alphaVal)
    {
      // here is where we make our guess about what the user wants?
      // if it's a local var name then we assume that what follows is the
      //  single statement which follows a nameless class declaration;
      //  in such a case, we just drop through and parse the statement, (leaving peeked tokens to be parsed).
      memberp=classp->lookup_member(word,false);
      if (!memberp)
	{
	  // it's not a local variable, which means it's the name of the new object
	  toke=lexerp->readtoken(tokenMode);
	  objname=word;
	  objp=createObject(classname,objname);
	  // now we want to check if there is a shorthand assignment
	  // if not, just drop through
	  toke=lexerp->peektoken(tokenMode);
	  if (toke.getstr()=="=")
	    {
	      // yes there is a shorthand assignment, so complete the object and the parse it
	      // as if it were a normal assignment
	      toke=lexerp->readtoken(tokenMode);
	      completeObject(objp);
	      Parseassignment(objp);
	      return;
	    }
	}
      else
	{
	  // it is a local variable, so it's NOT the object name (which is blank)
	  objp=createObject(classname,objname);
	}
    }
  else
    {
      // Create the object then get the rest of the declaration funny stuff
      objp=createObject(classname,objname);
    }

  // repeek the token (in case the object decided to consume the block (ie inline data)
  toke=lexerp->peektoken(tokenMode);
  if (toke.gettype()==openbraceSym)
    {
      // we got the start of a block of statements
      // note: we need to consume the token we peeked before we call parseblock)
      toke=lexerp->readtoken(tokenMode);
      // before executing the user's code which is within the scope of the object, let the object
      //  specify a block of code to execute first, within it's own scope
      pparse=objp->preparse();
      if (pparse!="")
	{
	  pushstring(" "+pparse+" } ");
	  Parseblock();
	}
      Parseblock();
    }
  else
    {
      // add any extra code the object wants to be considered as happening within its scope
      // before executing the user's code which is within the scope of the object, let the object
      //  specify a block of code to execute first, within it's own scope
      pparse=objp->preparse();
      if (pparse!="")
	{
	  pushstring(" "+pparse+" } ");
	  Parseblock();
	}
      Parsestatement();
    }

  // add any extra code the object wants to be considered as happening within its scope
  pparse=objp->postparse();
  if (pparse!="")
    {
      pushstring(" "+pparse+" } ");
      Parseblock();
    }

  // Now complete the object (leave its scope, etc.)
  completeObject(objp);
}




//--------------------------------------------------------------------------------
//---- Configpp::Parseassignment() ------------------------------------------------
//--------------------------------------------------------------------------------

void Configpp::Parseassignment(CfpVAccessor *varp,string vectoradstring="")
{
  // parse a variable/object assignment
  CfpVal expval;
  int index;
  
  // first parse an expression
  expval=Parsevalexp();

  if (vectoradstring=="")
    {
      // it's a normal scalar assignment

      // perform the assignment
      varp->set(&expval);
    }
  else
    {
      // it's a specific vector element assignment
      // first determine the index into the vector
      index=varp->parseindex(vectoradstring);
      
      // perform the assignment
      varp->set(expval,index);
    }

  // now normally there would be a semicolon here, BUT our new assignment format allows us to have
  // multiple assignments on one line (a la "Wire FROM input TO output" to stand in for "Wire {from=input;to=output;}"
  // now there better be a semicolon afterwards
  if (allow_shorthandassignments())
    {
      CfpToken toke;
      for (;;)
	{
	  toke=lexerp->peektoken(tokenMode);
	  if (toke.gettype()!=crSym)
	    break;
	  else
	    {
	      // the peektoken function will return initial CRs, which we don't care about
	      toke=lexerp->readtoken(tokenMode);
	    }
	}
      if (toke.gettype()==alphaVal)
	{
	  // looks like it is an identifier (var or object) so we are going to parse it as if it were the
	  // start of a shorthand assignment, if not it is an error and it will be detected
	  toke=lexerp->readpuretoken();
	  Parsealpha(toke);
	  return;
	}
      else if (toke.gettype()==openbraceSym)
	{
	  // looks like a new block immed. following an assignment, with no intervening semicolon
	  // so, like a series of assignments, we tread it as if it were within current scope
	  toke=lexerp->readpuretoken();
	  Parseblock();
	  return;
	}
    }

  // lok for a semicolon and throw an error if you don't find one
  Parsesemicolon();
}




//--------------------------------------------------------------------------------
//---- Configpp::Parsevectoradstring() --------------------------------------------
//--------------------------------------------------------------------------------

string Configpp::Parsevectoradstring()
{
  // read in a vector address of the form "[1 2 3]" or "[1,2,3]"
  CfpToken toke;
  string adstring="";

  // just grab all tokens until the close square
  for (;;)
    {
      toke=lexerp->readtoken(tokenMode);
      if (adstring!="")
	adstring.append(" ");
      adstring.append(toke.getstr());
      if (toke.gettype()==closesquareSym)
	break;
    }
  
  return adstring;
}




//--------------------------------------------------------------------------------
//---- Configpp::createObject() ---------------------------------------------------
//--------------------------------------------------------------------------------

CfpObject *Configpp::createObject(string classname, string objname)
{
  // create a new object instantiation and enter into the scope of that object
  CfpObject *objp=NULL;
  CfpClass *classp=NULL;
  string scopestr,fullname;
  map<string,CfpObject *>::iterator p;

  // first look up the class (it will throw an error if it can't find it).
  classp=factoryp->lookupClass(classname,true);

  // make sure that another object of the same name in the same scope does not exist
  if (objname!="")
    {
      fullname=scopestring()+"."+objname;
      p=objtable.find(fullname);
      if (p!=objtable.end())
	{
	  // object of duplicate name already exists
	  objp=p->second;
	  throw CfpError(objp->prettyname()+ " has previously been defined.","Within the same scope, all (named) objects must have unique names.",syntactical,false,false);
	  return NULL;
	}
    }

  // create a new object from the class
  debug("making a new instance of class "+classp->getname()+".");
  objp=classp->makenew(objname);
  if (!objp)
    throw CfpError("could not create object " +objname+ "[class " +classname+ "].","",library,false,false);

  // add the object to the current scope stack
  scopestack.push_back(objp);

  // add the object to the full object table using the NEW scope, add increment the idcounter
  objp->setidnum(++idcounter);
  scopestr=scopestring();
  objtable.insert(make_pair(scopestr,objp));

  // debug display current scope
  debug("entering scope " +scopestr+".");

  return objp;
}



//--------------------------------------------------------------------------------
//---- Configpp::completeObject() -------------------------------------------------
//--------------------------------------------------------------------------------

void Configpp::completeObject(CfpObject *objp)
{
  // complete the instantiation of the object, and leave the scope of that object
  CfpObject *fobjp;
  
  // first verify that we are in fact inside the scope that we are trying to leave
  if (scopestack.size()==0)
    throw CfpError("thought we were inside scope '"+objp->classname()+":"+objp->getname()+"' but are current scope is empty.","",internal,true,false);
  fobjp=scopestack.back();
  if (fobjp!=objp)
    throw CfpError("thought we were inside scope '"+objp->classname()+":"+objp->getname()+"' but we were really in scope '" +fobjp->classname()+":"+fobjp->getname()+ "'.","",internal,true,false);

  // complete the lazy instantiation of the object if nesc.
  runconstructor(objp);
  
  try
    {
      // now tell the object that we are leaving its scope (ie no more children) and get any return value
      GlobalReturnValue=objp->done();
    }
  catch (CfpError cfperror)
    {
      cfperror.report(this);
      // we got an error inside of a block. we don't want to exit immed. if it's
      // not a fatal error.
      //      destructobjtree(objp);
      throw cfperror;
    }
  catch (...)
    {
      // we got an error.  before rethrowing it we clean up
      debug("error on obj done(), so leaving scope " +scopestring()+".");
      destructobjtree(objp);
      throw;
    }


  // now leave the scope
  debug("leaving scope " +scopestring()+".");
  scopestack.pop_back();
}



//--------------------------------------------------------------------------------
//---- Configpp::runconstructor() -------------------------------------------------
//--------------------------------------------------------------------------------

void Configpp::runconstructor(CfpObject *objp)
{
  // complete the lazy instantiation of the object if nesc.
  if (objp->isinstantiated())
    return;

  debug("lazy instantiation of "+objp->prettyname()+".");

  // first we check if there are any explicit argument-based constructors for the class
  // and run any that are appropriate
  if (false)
    {
    }

  // run the default, argumentless constructor, if the object is not yet instantiated
  try
    {
      objp->construct();
    }
  catch (CfpError cfperror)
    {
      cfperror.report(this);
      // we got an error inside of a block. we don't want to exit immed. if it's
      // not a fatal error.
      destructobjtree(objp);
      throw cfperror;
    }
  catch (...)
    {
      // we got an error.  before rethrowing it we clean up
      destructobjtree(objp);
      throw;
    }

  // verfify that the object instantiated, or else throw error
  if (!objp->isinstantiated())
    {
      // we want to remove it from our table and scope, including all children.
      destructobjtree(objp);
      throw CfpError(objp->prettyname()+" did not instantiate.","",classerr,false,false);
    }
}




//--------------------------------------------------------------------------------
//---- Configpp::procAssignment() -------------------------------------------------
//--------------------------------------------------------------------------------

void Configpp::procAssignment(string varname, string varval)
{
  // perform the assignment
  debug("processing assignment [" +varname+ "='" +varval+ "'].");
}



//--------------------------------------------------------------------------------
//---- Configpp::parseSemicolon() -------------------------------------------------
//--------------------------------------------------------------------------------

void Configpp::Parsesemicolon()
{
  // consume next token which must be a semicolon
  CfpToken toke;

  toke=lexerp->peektoken();
  if (toke.gettype()!=semicolonSym)
    throw CfpError("expected a semicolon.","",syntactical,false,true,toke.getlocp());
  else
    lexerp->readtoken(tokenMode);
}













//}
// Namespace close-brace
