// cfpstdlib.h
// This file describes the Configpp standard library classes, which provide the base functionality
//  of the system

// forward declarations
class CfpstdInclude;
class CfpstdInline;
class CfpstdWire;
class CfpstdSplice;


#ifndef INC_cfpstdlib
#define INC_cfpstdlib

#include "cfpinc.h"
//#include "cfpaccessor.h"


//Not implemented in gcc yet (2.8.1)
//namespace Configpp{



//:CfpStdInclude class provides an object which handles the loading and parsing of
// configpp files which are included from other configpp files. (ie #include "filename" from C)

class CfpstdInclude: public CfpObject
{
private:
  string filename;
public:
  CfpstdInclude(bool registerme=true): CfpObject() {if (registerme) GlobalFactoryreg(this,"CfpstdInclude");};
  CfpstdInclude(Configpp *cfpfp,CfpClass *inclassp,string oname);
public:
  virtual void registerwrappers(CfpFactory *sfactp);
  virtual CfpObject *makenew(CfpClass *inclassp,string objname);		// generic makenew
  virtual void construct();							// no arg constructor of wrappedobj
  virtual void destruct();							// deinstatiate wrapped obj
};



//: CfpstdInline is used for putting stuff inside ('inline') a parsed file.
class CfpstdInline: public CfpObject
{
private:
  string filename;
  long startpos,endpos;
  CfpInputSource *sisp;
public:
  CfpstdInline(bool registerme=true) : CfpObject() {GlobalFactoryreg(this,"CfpstdInline");};
  CfpstdInline(Configpp *cfpfp,CfpClass *inclassp,string oname);
public:
  virtual void registerwrappers(CfpFactory *sfactp);
  virtual CfpObject *makenew(CfpClass *inclassp,string objname);		// generic makenew
  virtual void construct();							// no arg constructor of wrappedobj
  virtual void destruct();							// deinstatiate wrapped obj
public:
  string getfilename(){return filename;};
  long getstart(){return startpos;};
  long getend(){return endpos;};
  string gulpstring();
};





//:CfpstdWire class supports wiring together different inputs and outputs of dif. objects
class CfpstdWire: public CfpObject
{
private:
  CfpVal fromval,toval;
  CfpVAccessor *fromvarp, *tovarp;
  CfpVAccstub *laststubp;
  bool initialize;
public:
  CfpstdWire(bool registerme=true) : CfpObject() {GlobalFactoryreg(this,"CfpstdWire");};
  CfpstdWire(Configpp *cfpfp,CfpClass *inclassp,string oname);
public:
  virtual void registerwrappers(CfpFactory *sfactp);
  virtual CfpObject *makenew(CfpClass *inclassp,string objname);		// generic makenew
  virtual void construct();							// no arg constructor of wrappedobj
  virtual CfpVal done();							
  virtual void destruct();							// deinstatiate wrapped obj
  virtual void addobjmembers();							// add obj-specific members;
public:
  CfpVAccstub *getlaststubp() {return laststubp;};
  void setlaststubp(CfpVAccstub *instubp) {laststubp=instubp;};
};



//:CfpstdSplice class can be placed on wires as a black box
class CfpstdSplice: public CfpObject
{
protected:
  CfpVal inputval,outputval;
  CfpVal *triggervalp;
  CfpVAccessor *inputp, *outputp, *triggerp, *refreshinputp;
  CfpVAccstub *laststubp;
  bool manualretrieve;
public:
  CfpstdSplice(bool registerme=true) : CfpObject() {GlobalFactoryreg(this,"CfpstdSplice");};
  CfpstdSplice(Configpp *cfpfp,CfpClass *inclassp,string oname);
public:
  virtual void registerwrappers(CfpFactory *sfactp);
  virtual CfpObject *makenew(CfpClass *inclassp,string objname);		// generic makenew
  virtual void construct() {constructsetup();};							// no arg constructor of wrappedobj
  virtual void constructsetup();						// no arg constructor of wrappedobj
  virtual CfpVal done() {return donesetup();};							
  virtual CfpVal donesetup();							
  virtual void destruct();							// deinstatiate wrapped obj
  virtual void addobjmembers();							// add obj-specific members;
protected:
  virtual void process(CfpVal *valp);
  virtual void preprocessinput() {;};
  virtual void compute() {outputval=inputval;};
  virtual void postprocessoutput() {;};
  virtual bool iscomputetime() {return true;};
  virtual bool isoutputtime() {return true;};
protected:
  CfpVal refreshinput();
};




class CfpstdDatastream: public CfpObject
{
private:
  Datastream *streamp;
public:
  CfpstdDatastream(bool registerme=true): CfpObject() {GlobalFactoryreg(this,"CfpstdDatastream");};
  CfpstdDatastream(Configpp *cfpfp,CfpClass *inclassp,string oname);
public:
  virtual void registerwrappers(CfpFactory *sfactp);
  virtual CfpObject *makenew(CfpClass *inclassp,string objname);		// generic makenew;
  virtual void construct();							// no arg constructor of wrappedobj
  virtual void destruct();							// deinstatiate wrapped obj
  virtual CfpVal done();			                			// finish it;
public:
  Datastream *getstreamp(){return streamp;};
};





















//}
// Namespace close-brace
#endif
