// cfpsymbols.h
// This file holds all enumerations and defined symbols

#ifndef INC_cfpsymbolsh
#define INC_cfpsymbolsh

//Not implemented in gcc yet (2.8.1)
//using namespace std;
//namespace Configpp{





//--------------------------------------------------------------------------------
//---- Version Information -------------------------------------------------------
//--------------------------------------------------------------------------------

#define	SAF_VERSIONFLOAT	0.4
#define	SAF_VERSIONSTRING	"0.4 (jan 2000)"
#define	SAF_RCFILENAME		".cfpronrc"





//--------------------------------------------------------------------------------
//---- Error Display -------------------------------------------------------------
//--------------------------------------------------------------------------------


// These constants describe how erros are highlighted.
//  on a linux console and other vts that support color escape sequences,
//  we can do a nice color identification, in which case the
//  bool INSERT_MULTILINE_MARKS should be true, (which restores
//  the default between multi-line errors.
//  otherwise false looks better if you are using standard characters to
//  indicate location of errors.

// for pure text term
//#define BOLD_LEFT			""
//#define BOLD_RIGHT			""

// for linux and others with color
#define BOLD_LEFT			"\033[;32;1m"
#define BOLD_RIGHT			"\033[m"



//--------------------------------------------------------------------------------
//---- IO Constants --------------------------------------------------------------
//--------------------------------------------------------------------------------

// This is a fake character code we use to return the fact that there are no more 
//  input streams to read.  So this better be a char we will never see.
#define ENDSTREAMS_CHAR		1


//--------------------------------------------------------------------------------
//---- Parsing Constants ---------------------------------------------------------
//--------------------------------------------------------------------------------

// put some kind of large limit in case the person has a runaway string or comment
//  (we don't want to try to allocate 10gigs if they try to trick us).
#define MAX_TOKENLENGTH		1000

// Maximum dimensions for an array
#define MAXDIMS			10



//--------------------------------------------------------------------------------
//---- Compiler Flags/Fixes ------------------------------------------------------
//--------------------------------------------------------------------------------

//:My c++ ref book says to #include 'sstream' but gcc 2.8.1 wants 'strstream'.
#define STRSTRAM <strstream>

#ifndef stringstream
  #define stringstream strstream
#endif
#ifndef istringstream
  #define istringstream istrstream
#endif
#ifndef ostringstream
  #define ostringstream ostrstream
#endif





//--------------------------------------------------------------------------------
//---- Error Symbols -------------------------------------------------------------
//--------------------------------------------------------------------------------

//: Error codes
typedef enum CfpErrorCode
  {
    undefined,
    file,
    io,
    lexical,
    syntactical,
    internal,
    library,
    sys,
    classerr,
    gui,
    toomanyerr
  };





//--------------------------------------------------------------------------------
//---- Lexical Enumerations ------------------------------------------------------
//--------------------------------------------------------------------------------

//: Input source types
typedef enum CfpInputSourceType
  {
    filetype=1,
    stringtype
  };

//: Kinds of lexical and syntactical tokens
typedef enum CfpTokenType
{
  initVal,
  commentVal,
  eofSym,
  numericVal,
  semicolonSym,
  stringVal,
  squotestringVal,
  mathVal,
  conditionVal,
  commaSym,
  openparenSym,
  closeparenSym,
  opensquareSym,
  closesquareSym,
  otherSym,
  poundVal,
  alphaVal,
  openbraceSym,
  closebraceSym,
  wordVal,
  crSym,
};

typedef enum TokenReadMode
{
  tokenMode,
  tokencrMode,
  wordMode,
  crlineMode  
};


//--------------------------------------------------------------------------------
//---- Parser Enumerations -------------------------------------------------------
//--------------------------------------------------------------------------------

//: Parser modes
typedef enum CfpParserModeType
{
  undefinedMode,
  syntaxcheckMode,
  runMode
};


//: Value types
typedef enum CfpValType
{
  undefinedVType,
  boolVType,
  numberVType,
  stringVType,
  voidpointerVType,
  classpointerVType,
  objectpointerVType,
  vaccesspointerVType
};



//: Adoption return tyoes
typedef enum CfpAdoptType
{
  ownAdopt,
  acceptAdopt,
  rejectAdopt,
  errorAdopt
};




//--------------------------------------------------------------------------------
//---- Running Enumerations
//--------------------------------------------------------------------------------


//: Timer types
typedef enum CfpTimerType
{
  wallclockType,
  cycleclockType,
  simclockType
};








// }
// Namespace close-brace
#endif
