// cfpwiresplice.C

#include "cfpinc.h"

//Not implemented in gcc yet (2.8.1)
//namespace Configpp{










//--------------------------------------------------------------------------------
//---- CfpSpliceNoise -------------------------------------------------------------
//--------------------------------------------------------------------------------



CfpSpliceNoise::CfpSpliceNoise(Configpp *cfpfp,CfpClass *inclassp,string oname)
  :CfpstdSplice(cfpfp,inclassp,oname)
{
  // this is the constructor that gets called every time the user creates a new
  // instantiated object of the class.
  // This function and makenew, the next, are partners.
}


CfpObject *CfpSpliceNoise::makenew(CfpClass *inclassp,string objname)
{
  // create a new instance of ourselves, the wrapper.
  // Important: instantiating the wrapped class will come later (in function 'construct').
  CfpObject *newp=NULL;

  newp=new CfpSpliceNoise(configppp,inclassp,objname);
  return newp;
}



void CfpSpliceNoise::registerwrappers(CfpFactory *sfactp)
{
  // this is the constructor that gets called ONCE when the class is registered with the factory.
  CfpClass *newclassp;

  newclassp=sfactp->registerclass("CfpStd","Splice.Noise",this);

  // register an optional trigger (if not specified, info will process and propagate when source changes)
  newclassp->addarg("trigger","",false,"the trigger source to cause info to propagate through splice");

  // register the noise arguments
  newclassp->addvar("mean","0.0",false,"mean value of the gaussian noise");
  newclassp->addvar("stddev","1.0",false,"standard deviation of the gaussian noise");
}


void CfpSpliceNoise::construct()
{
  // instantiate the wrapped object.  this is where we actually do late (lazy) instantiation.
  // if we are a wrapper for another class, we would allocate that target wrapped class here.
  // and save it to a local variable.  otherwise this is its own wrapper.

  // run the parent construct
  constructsetup();
  
  // now any local setup
  addaccessors();
}


void CfpSpliceNoise::addaccessors()
{
  // add the accessors for the splice
  CfpVAccessor *varp;
  varp=addvar(new CfpVTAccessor<double>("mean",this,&noisemean));
  varp=addvar(new CfpVTAccessor<double>("stddev",this,&noisestddev));
}


CfpVal CfpSpliceNoise::done()
{
  // children have been added, so now we can finalize the splice

  // run the parent done()
  donesetup();

  // now our local sutff
  return retval;
}


void CfpSpliceNoise::compute()
{
  // noiseify the input to get the output
  double dval=inputval.get_number();
  double noise;

  //  cout << "Input of noisy splice was '"+inputval.get_string()+"'.\n";

  // okay corrupt it with noise
  if (noisestddev!=0)
    {
      noise=(configppp->jrandhelper).RandomGaussian(noisemean,noisestddev);
      dval+=noise;
    }

  outputval=CfpVal(dval);
  //  cout << "Output of noisy splice was '"+outputval.get_string()+"'.\n";
}








//--------------------------------------------------------------------------------
//---- CfpSpliceDelay ------------------------------------------------------------
//--------------------------------------------------------------------------------



CfpSpliceDelay::CfpSpliceDelay(Configpp *cfpfp,CfpClass *inclassp,string oname)
  :CfpstdSplice(cfpfp,inclassp,oname)
{
  // this is the constructor that gets called every time the user creates a new
  // instantiated object of the class.
  // This function and makenew, the next, are partners.

  // initialize
  alwaysoutput=false;
  shouldoutput=false;
  shouldcompute=true;
  stepcounter=0;
  deltatcounter=0.0;
  lasttime=0;
  buffersize=0;
}


CfpObject *CfpSpliceDelay::makenew(CfpClass *inclassp,string objname)
{
  // create a new instance of ourselves, the wrapper.
  // Important: instantiating the wrapped class will come later (in function 'construct').
  CfpObject *newp=NULL;

  newp=new CfpSpliceDelay(configppp,inclassp,objname);
  return newp;
}




void CfpSpliceDelay::registerwrappers(CfpFactory *sfactp)
{
  // this is the constructor that gets called ONCE when the class is registered with the factory.
  CfpClass *newclassp;

  newclassp=sfactp->registerclass("CfpStd","Splice.Delay",this);

  // register an optional trigger (if not specified, info will process and propagate when source changes)
  newclassp->addarg("trigger","",false,"the trigger source to cause info to propagate through splice");

  // register the delay arguments

  // the following say how often a trigger sets off processing (only one should be set by user)
  newclassp->addvar("deltatrate","-1",false,"sample rate in triggering time deltat");
  newclassp->addvar("samplerate","0",false,"delay steps for triggering");

  // the following say how old a value gets passed to output source on firing (only one should be set by user)
  newclassp->addvar("timedelay","-1",false,"how old (in trigger deltat) the output value is");
  newclassp->addvar("stepdelay","0",false,"how old (in trigger occurences) the output value is");

  // a flag for saying whether, despite sampling rates, the output should be generated every trigger
  newclassp->addvar("alwaysoutput","false",false,"flag for saying whether the output should be generated every trigger");
}


void CfpSpliceDelay::construct()
{
  // instantiate the wrapped object.  this is where we actually do late (lazy) instantiation.
  // if we are a wrapper for another class, we would allocate that target wrapped class here.
  // and save it to a local variable.  otherwise this is its own wrapper.

  // run the parent construct
  constructsetup();
  
  // now our local setup
  addaccessors();
}


void CfpSpliceDelay::addaccessors()
{
  // add the accessors for the splice
  CfpVAccessor *varp;
  varp=addvar(new CfpVTAccessor<double>("deltatrate",this,&deltatrate));
  varp=addvar(new CfpVTAccessor<int>("samplerate",this,&samplerate));
  varp=addvar(new CfpVTAccessor<double>("timedelay",this,&timedelay));
  varp=addvar(new CfpVTAccessor<int>("stepdelay",this,&stepdelay));
  varp=addvar(new CfpVTAccessor<bool>("alwaysoutput",this,&alwaysoutput));
}


CfpVal CfpSpliceDelay::done()
{
  // children have been added, so now we can finalize the splice

  // run the parent done()
  donesetup();

  // now our local sutff
  //  shouldoutput=alwaysoutput;

  return retval;
}


void CfpSpliceDelay::destruct()
{
  // free any allocated space
}




bool CfpSpliceDelay::iscomputetime()
{
  // is it time to compute a new output
  // a litte tricky since there are two ways to specify the proccessing rate

  if (samplerate>0 || deltatrate<=0)
    {
      // user has specified that we should activate very 'samplerate' trigger occurences
      if (++stepcounter>samplerate)
	{
	  // yes, it's one of those times to fire
	  stepcounter=0;
	  shouldcompute=true;
	}
    }
  else if (deltatrate>0)
    {
      // user has specified that we should activate very 'deletarate' steps in the input trigger value
      double thistime=triggervalp->get_number();
      double thisdeltat=thistime-lasttime;
      if (thisdeltat>deltatrate)
	{
	  // a large enought deltat has passed, so we should activate
	  shouldcompute=true;
	  lasttime=thistime;
	}
    }
}


bool CfpSpliceDelay::isoutputtime()
{
  // should we send an output
  if (shouldoutput)
    {
      // if we are not always outputing, then we want to reset our output flag
      if (!alwaysoutput)
	shouldoutput=false;
      return true;
    }
  return false;
}




void CfpSpliceDelay::compute()
{
  // Delay the input to get the output
  double thistime;

  //  cout << "... IN DELAY SPLICE CALCULATE.\n";
  // what we want to do now is send an output from our buffer which represents
  //  an old value, whose age is determined by timedelay or stepdelay.

  if (!shouldcompute)
    {
      // since we didn't compute a new value, we are just sending an old output because we are set to alwaysoutput
      // so just send through out old output
    }
  else
    {
      // the way we simulate a delay is by creating a pipeline.  When a new value is sampled, we push it in one
      //  end and take the last value out of the other end.
      // the size of the pipeline is governed by the delay magnitude, and the samplingrate.
      // This is made tricky because 1) if delay is specified in deltat, then we have to figure out how big the pipeline should be.
      //  and 2) we want the user to be able to modify the delay online (yikes!), thus the pipeline size should dynamically adjust.

      if (stepdelay>0 || timedelay==-1)
	{
	  // user has specified a fixed number of trigger steps that the output should be delayed
	  if (stepdelay==0)
	    {
	      // no delay, so just pass through the input, no need to buffer anything
	      outputval=inputval;
	      shouldoutput=true;
	    }
	  else
	    {
	      if (stepdelay>buffersize || buffersize==0)
		{
		  // we haven't accumulated enough past values, so no output
		}
	      else if (stepdelay<buffersize)
		{
		  // it must be that someone has reduced the step delay, since now we have
		  // too many old values in the buffer, so we so get rid of them
		  while (stepdelay<=buffersize)
		    pop(&outputval);
		  shouldoutput=true;
		}
	      else
		{
		  // grab the oldest value which we have previously buffered
		  pop(&outputval);
		  // since we computed a new output, set the flag to trigger an output
		  shouldoutput=true;
		}
	      // buffer the current input
	      if (triggervalp->isdefined())
		thistime=triggervalp->get_number();
	      else
		thistime=0;
	      push(&inputval,thistime);
	    }
	}
      else
	{
	  // user has specified a time delay
	  if (timedelay<=0)
	    {
	      // no delay, so just pass through the input, no need to buffer anything
	      outputval=inputval;
	      shouldoutput=true;
	    }
	  else
	    {
	      double thistime=triggervalp->get_number();
	      //	      cout << "TIME="+JRtostr(thistime)+"  (buffersize="+JRtostr(buffersize)+").\n";

	      if (buffersize>0)
		{
		  double valtime=peektime();
		  double thisdeltat=thistime-valtime;

		  //		  cout << "thistime="+JRtostr(thistime)+"  valtime="+JRtostr(valtime)+"  thisdeltat="+JRtostr(thisdeltat)+"  timedelay="+JRtostr(timedelay)+".\n";

		  if (thisdeltat<timedelay)
		    {
		      // the oldest stored value is not old enough, so send out nothing
		    }
		  else
		    {
		      shouldoutput=true;
		      // okay, we know that we have a data item that is old enough
		      // BUT it is possible that the user has changed the timedelay on us
		      //  such that the last value stored is too old
		      // so we look for the buffered which is closest to and older than the delay
		      do
			{
			  pop(&outputval);
			  valtime=peektime();
			  thisdeltat=thistime-valtime;
			} while (buffersize>0 && thisdeltat>=timedelay);
		    }
		}
	      // buffer the current input
	      push(&inputval,thistime);
	    }
	}

    }

  // reset the shouldcompute flag
  shouldcompute=false;
}









void CfpSpliceDelay::push(CfpVal *valp,double timeval)
{
  // push the current value onto our pipeline (lifo) stack
  CfpandTime *st=new CfpandTime(valp,timeval);
  pipeline.push_front(st);
  ++buffersize;
}

void CfpSpliceDelay::pop(CfpVal *valp)
{
  // pop the oldest value into valp
  if (buffersize==0)
    {
      cout << "INTERNAL ERROR: ATTEMPT TO pop off a cfpandtime element when buffer is empty.\n";
      return;
    }

  CfpandTime *st=pipeline.back();
  *valp=st->val;

  // now remove it from pipeline and free it
  pipeline.pop_back();
  delete st;
  --buffersize;
  }

double CfpSpliceDelay::peektime()
{
  // return the time stamp of the last (oldest) element in the pipeline
  if (buffersize==0)
    return 0;

  CfpandTime *st=pipeline.back();
  return st->timestamp;
}





//}
// Namespace close-brace
