// configpp.C

#include "cfpinc.h"
#include "JRutils.h"
#include "JRrand.h"


// forward declarations for library objects.  these and associated static loading
// need to be replaced with a dll or dll-like, or global factory locater
//#include "cfplibs/fltk/swfltk.h"
//#include "cfplibs/biobot/swbiobot.h"
//#include "cfplibs/datafactory/swdatafactory.h"
//#include "cfplibs/plot/swplot.h"
//#include "cfplibs/movie/swmovie.h"
//#include "cfplibs/hexcontrol/swhexcontrol.h"
//#include "cfplibs/pdcontrol/swpdcontrol.h"
//#include "cfplibs/macros/swmacros.h"





//Not implemented in gcc yet (2.8.1)
//using namespace std;
//namespace Configpp{











CfpFactory *globalfpointer=NULL;

void GlobalFactoryreg(CfpObject *objp,string packagename)
{
  if (globalfpointer==NULL)
    globalfpointer=new CfpFactory;
  globalfpointer->registerspawner(objp,packagename);
}

CfpFactory *GlobalFactoryp()
{
  if (globalfpointer==NULL)
    globalfpointer=new CfpFactory;
  return globalfpointer;
}





Configpp::Configpp(string inappname,CfpFactory *infactoryp, int argc, char **argv)
{
  // Constructor/initialization
  appname=inappname;
  errors=0;
  diddone=false;
  idcounter=0;
  factoryp=infactoryp;
  lexerp=NULL;
  timermanagerp=NULL;
  cfpwrapp=NULL;

  // Set the default debug&warning display modes, etc.
  Debugmode=false;
  Warningmode=false;
  Hintmode=true;
  Parsermode=runMode;
  Allowpounds=true;
  //  Maxerrors=10;
  Maxerrors=1;

  // initialize our random number generator
  jrandhelper.TimeRandomize();

  try
    {
      // Process command line options which may overide the above.
      //  (we do this before loading rc files so that we can go to debug mode right away, if desired)

      // Initialize all wrapperspawers with the configpp pointer (this will cause them to register their classes)
      factoryp->setconfigppp(this);

      if (argc>1)
	commandline(argc,argv);
      
      // Init the search directories
      addsearchdir("");
      addsearchdir(".");
      // the following are built-in guesses about where configpp files are
      //  there is no harm in these directories not existing.  but we might want to move them to
      //  a file where the installer might change them...
      addsearchdir("/usr/local/lib/configpp");
      addsearchdir("/usr/local/include/configpp");
      
      // Load any config info from environment or .rc file
      loadrc();

      // initialize the interpreter
      debug("Initializing interpreter, loading config info from environment, .cfpronrc file");

      // reprocess command line options again, now that rc files have been loaded, to overide them if desired
      //      if (argc>1)
      //	commandline(argc,argv);

      // Create lexer
      lexerp=new CfpLexer(this);

      // Create the timer manager
      timermanagerp=new CfpTimerManager(this);
      
      // Create and register foundation objects that are nesc. for all others
      foundations(&argc,argv);

      // Load any library packages
      Loadlibs(&argc,argv);

      // all libs are loaded and initialized, now tell them we are about to start parsing
      factoryp->preinterpret();

      // maually add an instance of the CfpWrap class as our root object underwhich all interpretation takes place
      // this object wraps some internal state information and functions that users can then refer to.

      cfpwrapp=(CfpWrap *)createObject(" _Configpp","Configpp");
      cfpwrapp->construct();

      // save local copy of commandline args
      largc=argc;
      largv=argv;
    }
  catch (CfpError cfperror)
    {
      cfperror.report(this);
      exit(EXIT_FAILURE);
    }
}



Configpp::~Configpp()
{
  // Destructor: kill lexer and factory

  debug("Destructing the main configpp class.");

  // FIRST delete the objectable
  delete_objects();

  // now destruct helpers
  if (factoryp)
    delete factoryp;
  if (lexerp)
    delete lexerp;
  if (timermanagerp)
    delete timermanagerp;

  debug("Done cleaning up after Configpp.");
}


    
void Configpp::delete_objects()
{
  // delete all objects in objtable that we are responsible for
  map<string,CfpObject *>::iterator p;
  CfpObject *objp;

  debug("Clearing object table.");

  p=objtable.begin();
  while (p!=objtable.end())
    {
      objp=p->second;
      // if we manage the objects memory (default) then delete it
      if (objp)
	{
	  debug("deleting "+objp->prettyname());
	  objp->destruct();
	  delete objp;
	  objp=NULL;
	}
      ++p;
    }

  // now remove all objects
  objtable.clear();
  scopestack.clear();
}




void Configpp::welcome()
{
  //  Announce the Configpp interpreter and print version info, etc.
  //  C o p y r i g h t 1 9 9 6 - 2 0 0 0 j e s s e r e i c h l e r f r e d d e l c o o m y n
  if (appname == "not initialized yet")
    {
      warning("Configpp class should be initialized() before welcoming the user.\n");
      cout <<"\n"<< "Unitialized Configpp Interpreter, version " << SAF_VERSIONSTRING << ".\n\n\n";
    }
  else if (appname == "")
    cout <<"\n"<< "Configpp Interpreter, version " << SAF_VERSIONSTRING << ".\n\n\n";
  else
    cout <<"\n"<< appname << " using the Embedded Configpp Interpreter version " << SAF_VERSIONSTRING << ".\n\n\n";
}



void Configpp::foundations(int *argcp,char **argv)
{
  // create and register base factory classes needed
  CfpObject *objp;
  CfpUnit *unitp;
  
  // create the blank units that are used for variables/args that don't care about units
  factoryp->registerunit(new CfpUnit(factoryp,""));
  // create some common units
  factoryp->registerunit(new CfpUnit(factoryp,"seconds","time","second"));
  factoryp->registerunit(new CfpUnit(factoryp,"meters","length","meter"));
  factoryp->registerunit(new CfpUnit(factoryp,"radians","angle","radian"));
  factoryp->registerunit(new CfpUnit(factoryp,"newtons","force","newton"));
  factoryp->registerunit(new CfpUnit(factoryp,"cycles","time","cycle"));
  factoryp->registerunit(new CfpUnit(factoryp,"bool","truth","bool"));

  // create the blank type that is used for variables/args that don't care about types
  factoryp->registertype(new CfpType(factoryp,""));
  // cretate some types
  factoryp->registertype(new CfpType(factoryp,"double"));
  factoryp->registertype(new CfpType(factoryp,"int"));
  factoryp->registertype(new CfpType(factoryp,"string"));
  factoryp->registertype(new CfpType(factoryp,"enum"));
  factoryp->registertype(new CfpType(factoryp,"bool"));

  // create a Configpp spawner wrapper which provides language access to interpreter internals
  // most other spawners are created statically, but for this one we allocate here so we
  // can keep a pointer to it.
  cfpwrapp=new CfpWrap();
}



//--------------------------------------------------------------------------------
//---- Configpp::Loadlibs() -------------------------------------------------------
//--------------------------------------------------------------------------------

void Configpp::Loadlibs(int *argcp,char **argv)
{
  debug("Loading libraries");
  // here is where we would want to automatically discover and dynamically link in any libraries
}




//--------------------------------------------------------------------------------
//---- Configpp::addsearchdirs() --------------------------------------------------
//--------------------------------------------------------------------------------

void Configpp::addsearchdir(string dir)
{
  // add a directory to the search-dir list

  // Add a backslash onto dir if it isn't there already
  if (dir.length()>0)
    {
      if (dir[dir.length()-1]!='/')
	dir+="/";
    }

  // now add it to the *back* of the list (ie will be search after others before it).
  searchdirs.push_back(dir);
  debug("adding directory '" + dir + "' to file search path");
}
























bool Configpp::parsefile(string filename)
{
  // push a string onto the file stack, then parse it
  bool retv;
  int numerrors;

  retv=pushfile(filename);
  if (retv)
    retv=parse();

  return retv;
}


bool Configpp::parsestring(string parsestring,string name)
{
  // push a string onto the file stack, then parse it
  bool retv;

  debug("parsing an explicit string " + name + " \""+parsestring+"\"" );

  retv=pushstring(parsestring,name);
  if (retv)
    retv=parse();

  return retv;
}




bool Configpp::pushfile(string filename)
{
  // push a file on inputsource stack

  // All we do is tell the lexer to push the file and catch any errors
  try
    {
      lexerp->pushfile(filename);
    }
  catch (CfpError cfperror)
    {
      cfperror.report(this);
      if (cfperror.isfatal())
	throw cfperror;
      return false;
    }
  catch (...)
    {
      error("unexpected exception encountered while trying to push file '" +filename+ "' onto input stack.","",internal,true);
      return false;
    }

  return true;
}


bool Configpp::pushstring(string parsestring,string name)
{
  // push a string on inputsource stack

  // All we do is tell the lexer to push the string and catch any errors
  try
    {
      lexerp->pushstring(parsestring,name);
    }
  catch (CfpError cfperror)
    {
      cfperror.report(this);
      if (cfperror.isfatal())
	throw cfperror;
      return false;
    }
  catch (...)
    {
      error("unexpected exception encountered while trying to push string '" +name+ "' onto input stack.","",internal,false);
      return false;
    }

  return true;
}





inline void Configpp::debug(string debugstring)
  {
    // If in debug mode then print string, else nop
    // We make this inline, even though we aren't really worried about speed for this
    //  interpreter.

    if (Debugmode)
      {
	cout << "---> debug: " << debugstring << "\n";
	cout.flush();
      }
  }


void Configpp::announce(string annstring)
  {
    // print the announcement
    cerr << "---> " << annstring << "\n";
  }


void Configpp::warning(string warnstring, CfpLoc *wloc)
  {
    // If in warning mode then print string, else nop

    // print the warning text
    if (Warningmode || Debugmode)
      {
	cerr << "---> warning: " << warnstring << "\n";

	// show the token location as a pointer
	if (wloc!=NULL)
	  wloc->show();
      }
  }


void Configpp::error(string errstring,string hintstr,CfpErrorCode etype, bool fatalflag)
  {
    // display an error message using the CfpError class
    CfpError *se;

    se=new CfpError(errstring,hintstr,etype,fatalflag,this);
    processerror(se);

    delete se;
  }




void Configpp::processerror(CfpError *serr)
  {
    // first check if we have already reported this error.
    // this happes if it is an error which ocurred during parsing but which got passed out.
    // if so, then don't report it again.

    //    if (serr->wasreported())
    //      return;

    // report the error
    serr->report(this);

    // keep track of how many errors we have read so far
    if (++errors>=Maxerrors && !(serr->isfatal())) 
      {
	// even though the error is not intrinsically fatal, it is fatal too have too many of them.
	serr->setfatal();
	if (Maxerrors>1)
	  error("overwhelmed by errors.","",toomanyerr,true);
      }
    
    if (!(serr->isfatal()))
      {
	// it's a non-fatal error, so we don't exit.
	// go into syntax checking mode (ie no more executing)
	gosyntaxcheckmode();
      }
    else
      {
	// jesse: i can't get the code to properly rethrow errors up a chain once they are caught the
	//  first time, which is nescesary.  so we have to do this here to prevent segmentation faults
	cout << "Exiting due to errors...\n";
	exit(EXIT_FAILURE);
      }
  }






fstream *Configpp::inFilestream(string filename, const string ext)
{
  // just call the big one
  string tempstring;
  return inFilestream(filename,tempstring,ext);
}

fstream *Configpp::inFilestream(string filename, string &outfilename, const string ext)
{
  // Find and open the file stream for input, be smart about where to look, and about the extension.
  //  if we can't open it, throw an error.
  // Subtlety: if we do find it someplace unusual, change filename to the actual location
  fstream *isp;

  // what we do is we search our list of directories, checking for first the exact filename passed,
  // and then the filename+extension.
  isp=JRSearchFileOpen(filename,&searchdirs,outfilename,ios::in | ios::binary,ext);
  if (!isp)
    throw CfpError("could not open file '" + filename + "' for input.","",file,false,false);

  // succesful open
  return isp;
}

fstream *Configpp::outFilestream(string filename, string &outfilename, const string ext)
{
  // Find and open the file stream for output, be smart about where to look, and about the extension.
  //  if we can't open it, throw an error.
  // Subtlety: if we do find it someplace unusual, change filename to the actual location
  fstream *isp;

  // what we do is we search our list of directories, checking for first the exact filename passed,
  // and then the filename+extension.
  isp=JRSearchFileOpen(filename,&searchdirs,outfilename,ios::out | ios::trunc | ios::binary,ext);
  if (!isp)
    throw CfpError("could not open file '" + filename + "' for output.","",file,false,false);

  // succesful open
  return isp;
}

fstream *Configpp::outappFilestream(string filename, string &outfilename, const string ext)
{
  // Find and open the file stream for output appending, be smart about where to look, and about the extension.
  //  if we can't open it, throw an error.
  // Subtlety: if we do find it someplace unusual, change filename to the actual location
  fstream *isp;

  // what we do is we search our list of directories, checking for first the exact filename passed,
  // and then the filename+extension.
  isp=JRSearchFileOpen(filename,&searchdirs,outfilename,ios::app | ios::binary,ext);
  if (!isp)
    throw CfpError("could not open file '" + filename + "' for appending output.","",file,false,false);

  // succesful open
  return isp;
}




//--------------------------------------------------------------------------------
//---- Configpp::gosyntaxcheckmode() ----------------------------------------------
//--------------------------------------------------------------------------------

void Configpp::gosyntaxcheckmode()
{
  // go into syntax check mode (ie don't allow execution, but continue parsing)
  // and warn the user that we won't be executing further statements, just checking them.
  if (Parsermode!=syntaxcheckMode)
    warning("switching into syntax-check mode (further commands will not be executed).");
  Parsermode=syntaxcheckMode;
}



//--------------------------------------------------------------------------------
//---- Configpp::handlecomment() --------------------------------------------------
//--------------------------------------------------------------------------------

void Configpp::handlecomment(string commentstr,string sourcename)
{
  // this function is invoked when a comment is parsed.  normally we just ignore
  // comments, but we might want to print them or do something else with them.
  if (commentstr.length()>2)
    {
      if (commentstr[2]=='p')
	{
	  // we treat a 'p' after the comment start= (as in '//p ...' or '/*p ... */')
	  //  as a directive to print out the comment on stderr (useful for debugging)
	  if (commentstr[commentstr.length()-1]!='\n')
	    commentstr.append("\n");
	  cerr << "---> " << sourcename << " comment:" << commentstr.substr(3);
	}
    }
}



//--------------------------------------------------------------------------------
//---- Configpp::alias() ----------------------------------------------------------
//--------------------------------------------------------------------------------

string Configpp::alias(string wordstr)
{
  // return what wordstr is an alias of, just my querying our Alias map
  return aliases.alias(wordstr);
}


//--------------------------------------------------------------------------------
//---- Configpp::addalias() -------------------------------------------------------
//--------------------------------------------------------------------------------

void Configpp::addalias(string aliastr,string keyword,bool casesensitive=true)
{
  // add a new alias to the list of aliases, just call our Alias map
  string tempstr;

  // first check if there is already an alias of this type
  tempstr=aliases.alias(aliastr);
  if (tempstr!=aliastr)
    warning("replacing existing alias '" + aliastr + "' which previously mapped to '" + tempstr + "'; now maps to '" +keyword+ "'.");

  // then add the alias
  aliases.addalias(aliastr,keyword,casesensitive);
  if (casesensitive)
    debug("adding '" +aliastr+ "' as an alias of '" +keyword+ "'.");
  else
    debug("adding '" +aliastr+ "' as a case-insensitive alias of '" +keyword+ "'.");
}


















//--------------------------------------------------------------------------------
//---- Configpp::lookupVarArg() ---------------------------------------------------
//--------------------------------------------------------------------------------

CfpVAccessor *Configpp::lookupVarArg(string identifier,bool complain=true,bool skipcurrent=false)
{
  // we want to find an appropriate argument or variable
  // we must be devious, because what we want is:
  // IF the obect is not instantiated, then we look first for a constructor argument,
  //  if we can't find an arg, and there is a matching var, then we should force the object to instantiate
  //  and then return the var.
  // IF the object IS instantiated, we look FIRST for a variable.
  CfpVAccessor *varp=NULL;
  deque<CfpObject *>::reverse_iterator p;

  // we have no explicit object, so procede from deepest object in scope outwards, return first match we find
  p=scopestack.rbegin();

  // skip out of current scope first?
  if (skipcurrent && p!=scopestack.rend())
    ++p;

  while (p!=scopestack.rend())
    {
      // ask the object if it knows about the variable
      varp=(*p)->lookupva(identifier,false);
      if (varp)
	break;
      ++p;
    }

  if (!varp && complain)
    throw CfpError("could not find variable or argument '" +identifier+ "' in current scope.","",syntactical,false,false);
  return varp;
}












//--------------------------------------------------------------------------------
//---- Configpp::lookupVariable() -------------------------------------------------
//--------------------------------------------------------------------------------

CfpVAccessor *Configpp::lookupVariable(string varname,bool complain)
{
  // lookup the variable within the current scope and return pointer to accessor
  // note: the variable name may be specific to an explicit object and have leading objects
  //   (ie rootobject.subobject.varname)
  // OR may just be a varname where we must find the first matching varname in current scope.
  CfpVAccessor *varp=NULL;
  deque<CfpObject *>::reverse_iterator p;

  // we have no explicit object, so procede from deepest object in scope outwards, return first match we find
  p=scopestack.rbegin();
  while (p!=scopestack.rend())
    {
      // ask the object if it knows about the variable
      varp=(*p)->lookupvar(varname,false);
      if (varp)
	break;
      ++p;
    }

  if (!varp && complain)
    throw CfpError("could not find variable '" +varname+ "' in current scope.","",syntactical,false,false);
  return varp;
}



//--------------------------------------------------------------------------------
//---- Configpp::lookupFunction() -------------------------------------------------
//--------------------------------------------------------------------------------

CfpFuncAccessor *Configpp::lookupFunction(string funcname,bool complain)
{
  // lookup the cfp function within the current and return pointer to accessor
  CfpFuncAccessor *funcp=NULL;
  deque<CfpObject *>::reverse_iterator p;

  // we have just funcname, so procede from deepest object in scope outwards, returning first match we find
  p=scopestack.rbegin();
  while (p!=scopestack.rend())
    {
      // ask the object if it knows about the variable
      funcp=(*p)->lookupfunc(funcname,false);
      if (funcp)
	break;
      ++p;
    }

  if (!funcp && complain)
    throw CfpError("could not find function '" +funcname+ "' in current scope.","",syntactical,false,false);
  return funcp;
}





//--------------------------------------------------------------------------------
//---- Configpp::lookupObjtrail() -------------------------------------------------
//--------------------------------------------------------------------------------

CfpObject *Configpp::lookupObjtrail(string objtrail,bool complain)
{
  // look up the object specified by the objtrail passed
  // the object trail may be relative, starting from anywhere in current scope
  // ie if our current scope is (r2d2.leftleg.foot), then a valid trail is
  //  foot.length, or leftleg.foot, or r2d2.leftleg.foot, or r2d2, or r2d2.leftleg)
  CfpObject *objp=NULL;
  map<string,CfpObject *>::reverse_iterator p;
  map<string,CfpObject *>::reverse_iterator pend;
  map<string,CfpObject *>::iterator p2;
  string fullobjectname;
  // if the ambiguityerror flag is true, then we just look for the nearest (in terms of scope) match
  //  if false, then we check all objects so far and throw an error if we find more than one that matches
  // if ambiguitywarn==true then we check all and warn if more than one matches
  bool ambiguityerror;
  bool ambiguitywarn;
  string scopestr;
  int len,len2,len3;
  string objleader;
 
  ambiguityerror=false;
  ambiguitywarn=getwarningmode() || getdebugmode();

  // We also want to let the user avoid specifying all details of a trail, providing the address
  //  is unique.
  // ie given some user trail "b.c.e.g.j" we want to infer that they meant object "a.b.c.d.e.f.g.j"
  //  our lookup rules are: 1) the last object on the trail must be specified. 2) the trail must be unique within scope
  //  ie this means the user can skip initial and intermediate objects on a trail when the subsequent ids uniquely identify it.
  //   (note we only require/check for uniqueness within the nearest scope as we proceed outwards).
  
  // we make up to three passes through our object table.
  // on the first pass we look for explicit path names within current scope
  // on the second pass we look smartly at objects that are within current scope.
  // on the third pass we look at all of them



  // start at the current deepest scope and proceed outward, with first match winning (this is the fast one)
  if (!ambiguityerror && !ambiguitywarn)
    {
      // since we don't care about multiple matches we can do this fast check and if we find it leave
      scopestr=scopestring()+".dummy";
      while (scopestr!="")
	{
	  scopestr=JRstripleader(scopestr);
	  if (scopestr=="")
	    fullobjectname=objtrail;
	  else
	    fullobjectname=scopestr+"."+objtrail;
	  p2=objtable.find(fullobjectname);
	  if (p2!=objtable.end())
	    {
	      objp=p2->second;
	      if (objp)
		break;
	    }
	}
      if (objp)
	return objp;
    }


  // second pass, less restrictive
  scopestr=scopestring();
  len=scopestr.length();
  p=objtable.rbegin();
  pend=objtable.rend();
  while (p!=pend)
    {
      if (!(p->second)->scopesearchable())
	{
	  ++p;
	  continue;
	}

      fullobjectname=p->first;
      objleader=fullobjectname;
      JRstriptail(objleader);
      len2=objleader.length();
      if (len2>len)
	{
	  ++p;
	  continue;
	}
      if (len2>0 && objleader!=scopestr.substr(0,len2))
	{
	  // this means that the current object is not within the current scope (ie not a parent of current scope)
	  //  so we skip it this pass
	  ++p;
	  continue;
	}
      //	  cout << "... checking candidate '"+objtrail+"'  against fullobject trail '"+fullobjectname+"'.\n";
      if (compatibletrail(fullobjectname,objtrail))
	{
	  // okay we found a compatible trail
	  //	  cout << "... candidate '"+objtrail+"' matches object '"+fullobjectname+"' in pass1.\n";
	  if (objp!=NULL)
	    {
	      // BUT it is the second one we've found and so is ambiguous.  we throw it out but also:
	      if (ambiguityerror)
		throw CfpError("ambiguous reference '" +objtrail+ "' matches multiple objects.",namehint(objtrail),syntactical,false,false);
	      else if (ambiguitywarn)
		warning("ambiguous reference '" +objtrail+ "' matches multiple objects.\n");
	    }
	  else
	    {
	      objp=p->second;
	      // let's warn them if we are inferring a non-obvious object
	      len2=fullobjectname.length();
	      len3=objtrail.length();
	      if (len3>len2)
		warning("performing the non-obvious inference that '"+objtrail+"' refers to object '"+fullobjectname+".\n");
	      else if (objtrail!=fullobjectname.substr(len2-len3,len3))
		{
		  warning("performing the non-obvious inference that '"+objtrail+"' refers to object '"+fullobjectname+".\n");
		  cout << "... the substr was '"+fullobjectname.substr(len2-len3,len3)+"'.\n";
		}
	      if (!ambiguityerror && !ambiguitywarn)
		{
		  // since we don't care about multiple matches we can leave now
		  return objp;
		}
	    }
	}
      ++p;
    }
  

  // second pass, check against all objects
  p=objtable.rbegin();
  pend=objtable.rend();
  while (p!=pend)
    {
      if ((p->second)->scopesearchable())
	{
	  fullobjectname=p->first;
	  //	  cout << "... checking candidate '"+objtrail+"'  against fullobject trail '"+fullobjectname+"'.\n";
	  if (compatibletrail(fullobjectname,objtrail))
	    {
	      // okay we found a compatible trail
	      //	      cout << "... candidate '"+objtrail+"' matches object '"+fullobjectname+"' in pass2.\n";
	      if (objp!=NULL && objp!=p->second)
		{
		  // BUT it is the second one we've found and so is ambiguous.  we throw it out but also:
		  if (ambiguityerror)
		    throw CfpError("ambiguous reference '" +objtrail+ "' matches multiple objects.",namehint(objtrail),syntactical,false,false);
		  else if (ambiguitywarn)
		    warning("ambiguous reference '" +objtrail+ "' matches multiple objects.\n");
		}
	      else if (objp==NULL)
		{
		  objp=p->second;
		  warning("performing the non-obvious inference that '"+objtrail+"' refers to object '"+fullobjectname+".\n");

		  if (!ambiguityerror && !ambiguitywarn)
		    {
		      // since we don't care about multiple matches we can leave now
		      return objp;
		    }
		}
	    }
	}
      ++p;
    }


  
  if (!objp && complain)
    throw CfpError("could not find object '" +objtrail+ "'.",namehint(objtrail),syntactical,false,false);
  
  return objp;
}



bool Configpp::compatibletrail(string fulltrail,string candidatetrail)
{
  // given a complete trail 'fulltrail', and a candidate trail with possibly missing entries 'candidatetrail'
  //  return true if they are compatible.
  // ie given some user trail "b.c.e.g.j" we want to infer that they meant object "a.b.c.d.e.f.g.j"
  //  our lookup rules are: 1) the last object on the trail must be specified. 2) the trail must be unique within scope
  //  ie this means the user can skip initial and intermediate objects on a trail when the subsequent ids uniquely identify it.
  //   (note we only require/check for uniqueness within the nearest scope as we proceed outwards).
  string fulltoken,candidatetoken;

  // the way we check for compatibility is simple if inefficient.
  // to be compatible, every token on the candidate trail must be accounted for on the full trail in order,
  //  but not vice versa.  and additionally, the last token on both must match, since the user must specify the end object target.

  // first check that the last token on both are identical
  fulltoken=JRstriptail2(fulltrail);
  candidatetoken=JRstriptail2(candidatetrail);
  if (fulltoken!=candidatetoken)
    return false;
      
  // now procede through rest of candidate trail
  while (candidatetrail!="")
    {
      candidatetoken=JRstriptail2(candidatetrail);
      while (fulltrail!="")
	{
	  fulltoken=JRstriptail2(fulltrail);
	  if (fulltoken==candidatetoken)
	    break;
	}
      if (fulltoken!=candidatetoken)
	{
	  // we exhausted the full trail without finding a match for the latest token in the candidate trail
	  //  so this object fails
	  return false;
	}
    }

  // if we end up here then we consumed the entire candidate trail and we are still compatible, so this is a match
  return true;
}



//--------------------------------------------------------------------------------
//---- Configpp::lookupVal() ------------------------------------------------------
//--------------------------------------------------------------------------------

CfpVal Configpp::lookupVal(string identifier,bool complain)
{
  // search the scope stack for a non-undefined value for the identifier
  CfpVal val;
  CfpVAccessor *varp=NULL;
  deque<CfpObject *>::reverse_iterator p;

  p=scopestack.rbegin();
  while (p!=scopestack.rend())
    {
      // ask the object if it knows about the variable
      varp=(*p)->lookupva(identifier,false);
      if (varp)
	{
	  // we found a match, but is it defined
	  val=varp->get();
	  if (val.isdefined())
	    break;
	}
      ++p;
    }

 if (!(val.isdefined()) && complain)
    throw CfpError("could not find a value for identifier '" +identifier+ "' in current scope.","",syntactical,false,false);
  return val;
}








//--------------------------------------------------------------------------------
//---- Configpp::lookupVAcc() -----------------------------------------------------
//--------------------------------------------------------------------------------

CfpVAccessor *Configpp::lookupVAcc(string identifier,bool complain)
{
  // look up the var accessosor that the identifier refers to
  CfpVAccessor *varp=NULL;
  CfpObject *objp=NULL;
  string leaderstr;

  leaderstr=JRstripleader(identifier);
  if (leaderstr=="Parent")
    {
      // if a value is specified as Parent.varargname, then we treat this specially.
      // we treat this as the request for a variable belonging to the parent tree,
      //  and search for the variable from the parent and up.
      // This can be very useful when one is in a macro or other situation where one wants to
      //  assign a variable using a value which is now shadowed by the variable.
      //  ie it allows you to define a macro with arguments the same as the arguments of an object
      //   in the macro and still access the macro argument.
      //  For example, if the object Box takes arguments x and y, then a macro might be set up as follows:
      //  Macro SuperBox
      //    {
      //    Macroarg x;
      //    Macroarg y;
      //    Inline
      //      {
      //      Box
      //         {
      //         label="a superbox macro built box.";
      //         style="Super";
      //         x=Parent.x;
      //         y=Parent.y;
      //         }
      //      }
      //    }
      varp=lookupVarArg(identifier,complain,true);
      return varp;
    }
  else if (leaderstr!="")
    {
      // okay we have an explicit trail
      objp=lookupObjtrail(leaderstr,complain);
    }
  
  // is it a member argument or variable
  if (objp)
    varp=objp->lookupva(identifier,complain);
  else
    varp=lookupVarArg(identifier,complain,false);
  
  return varp;
}







//--------------------------------------------------------------------------------
//---- Configpp::scopestring() ----------------------------------------------------
//--------------------------------------------------------------------------------

string Configpp::scopestring()
{
  // make a string out of the current scope
  string scopestr="";
  string ostring;
  string idstring;
  deque<CfpObject *>::iterator p;
  int count=0;

  p=scopestack.begin();
  while (p!=scopestack.end())
    {
      ++count;
      ostring=(*p)->getname();
      if (ostring=="")
	{
	  idstring=JRdtostr((*p)->getidnum());
	  ostring=(*p)->classname()+"("+idstring+")";
	}
      if (scopestr.length()>0)
	scopestr+="." + ostring;
      else
	scopestr=ostring;
      ++p;
    }

  return scopestr;
}





//--------------------------------------------------------------------------------
//---- Configpp::semihint() -------------------------------------------------------
//--------------------------------------------------------------------------------

string Configpp::semihint()
{
  // this is called when we get an error reading a token, and we want to inform them if we think they may have
  //  forgotten to use a semicolon at the last CR.  we can give this hint because althouth we treat CRs
  //  as whitespace when parsing, most of the time a CR signifies the end of a statement and it preceded
  //  by a semicolon.
  if (lexerp->getsemihint())
    return "was there supposed to be a semicolon on the previous line?";
  else
    return "";
}



//--------------------------------------------------------------------------------
//---- Configpp::namehint() -------------------------------------------------------
//--------------------------------------------------------------------------------

string Configpp::namehint(string inname)
{
  // display a hint about name
  // this function is called when the user refers to an unknown class or variable in some scope.
  //  BUT even though the name is not known in scope, the user may be refering to soemthing out of scope,
  //  so we want to give them a hint about this.
  string hintstring="";
  CfpObject *objp=NULL;


  // note: if we were nice we would do a case *insensitive* search for close matches

  // first check if it matches any object names in current scope
  objp=lookupObjtrail(inname,false);
  if (objp)
    hintstring+="         object '"+inname+"' (class "+objp->classname()+").\n";

  // now check if it matches any classes or class members
  hintstring+=factoryp->namehint(inname);

  if (hintstring!="")
    hintstring="Remember that identifiers are case sensitive.  Closest matches:\n"+hintstring;
  else
    hintstring="Remember that identifiers are case sensitive.\n";

  return hintstring;
}



//--------------------------------------------------------------------------------
//---- Configpp::adoptobj() -------------------------------------------------------
//--------------------------------------------------------------------------------

CfpObject *Configpp::adoptobj(CfpObject *cobjp)
{
  // the object wants to be adopted
  // so we query the objects in the current scope and see if any want to be its parent
  // return a pointer to the first object that adopted it.  or return NULL if no one wants it.
  CfpObject *objp=NULL;
  deque<CfpObject *>::reverse_iterator p;
  
  p=scopestack.rbegin();
  while (p!=scopestack.rend())
    {
      // ask the object if it wants to adopt
      //      if ((*p)->isa(typestr))
      if (false)
	{
	  objp=*p;
	  break;
	}
      ++p;
    }
  
  return objp;
}





//--------------------------------------------------------------------------------
//---- Configpp::inscopefinda() ---------------------------------------------------
//--------------------------------------------------------------------------------

CfpObject *Configpp::inscopefinda(string typestr)
{
  // find the nearest obj within the current scope which is of type typestr
  // this function might be called by a new object which wants to find a higher level encompassing 
  // object of a specific type.  
  CfpObject *objp=NULL;
  deque<CfpObject *>::reverse_iterator p;
  
  p=scopestack.rbegin();
  while (p!=scopestack.rend())
    {
      // ask the object if it is the type
      if ((*p)->isa(typestr))
	{
	  objp=*p;
	  break;
	}
      ++p;
    }
  
  return objp;
}


//--------------------------------------------------------------------------------
//---- Configpp::finda() ----------------------------------------------------------
//--------------------------------------------------------------------------------

CfpObject *Configpp::finda(string typestr)
{
  // find any instantiated object of the specific type.
  // this function might be called by a new object which wants to find a higher level encompassing 
  // object of a specific type.  or by an object which can only be instantiated once.
  CfpObject *objp=NULL;
  map<string,CfpObject *>::reverse_iterator p;
  
  p=objtable.rbegin();
  while (p!=objtable.rend())
    {
      // ask the object if it is the type
      if ((p->second)->isa(typestr))
	{
	  objp=p->second;
	  break;
	}
      ++p;
    }
  
  return objp;
  }





//--------------------------------------------------------------------------------
//---- Configpp::findchild()
//--------------------------------------------------------------------------------

CfpObject *Configpp::findchild(CfpObject *pobjp,string typestr,bool complain=true)
{
  // find the first child of the parent which matches typestr.
  // note that the child may not be on current scope stack, having been completed already
  CfpObject *objp=NULL;
  CfpObject *cobjp;
  map<string,CfpObject *>::reverse_iterator p;
  
  //  cout << "... looking for a child of type '"+typestr+"'.\n";
  p=objtable.rbegin();
  while (p!=objtable.rend())
    {
      // ask the object if it is the type, and if its parent is pobjp
      cobjp=(p->second);
      if (cobjp->isa(typestr))
	{
	  if (cobjp->getparentobjp()==pobjp)
	    {
	      objp=cobjp;
	      break;
	    }
	}
      ++p;
    }
  
  if (objp==NULL && complain)
    throw CfpError("could not find child object of type '" +typestr+ "' in current scope.","",classerr,false,false);

  return objp;
  }





//--------------------------------------------------------------------------------
//---- Configpp::enclosingparent()
//--------------------------------------------------------------------------------

CfpObject *Configpp::enclosingparent(CfpObject *childobjp)
{
  // find the nearest enclosing parent object on the current stack
  // note: we don't start looking for parent until we see the child
  CfpObject *objp=NULL;
  deque<CfpObject *>::reverse_iterator p;
  bool skipping=true;
  
  p=scopestack.rbegin();
  while (p!=scopestack.rend())
    {
      // ask the object if it is the type
      if (objp==NULL)
	objp=*p;

      if (!skipping)
	{
	  objp=*p;
	  break;
	}

      if (*p==childobjp)
	{
	  skipping=false;
	  objp=NULL;
	}
      ++p;
    }

  return objp;
}




//--------------------------------------------------------------------------------
//---- Configpp::parentinstantiate() ----------------------------------------------
//--------------------------------------------------------------------------------

CfpObject *Configpp::parentinstantiate(CfpObject *childobjp,string typestr,bool complain)
{
  // find and instantiate parent object in scope. throw error if we can't find one
  // note: we don't start looking for parent until we see the child
  // also note that typestr can be a space separated list of acceptable parent types
  // we return the first parent that matches any of the types in typestr.
  CfpObject *objp=NULL;
  deque<CfpObject *>::reverse_iterator p;
  bool skipping=true;
  
  p=scopestack.rbegin();
  while (p!=scopestack.rend())
    {
      // ask the object if it is the type
      if ((*p)->isa(typestr) && !skipping)
	{
	  objp=*p;
	  break;
	}
      if (*p==childobjp)
	skipping=false;
      ++p;
    }

  if (!objp && complain)
    throw CfpError("child '"+childobjp->prettyname()+"' needs a parent of type '" +typestr+ "'.","",syntactical,false,false);

  if (objp)
    {
      // make sure it is instantiated
      runconstructor(objp);
    }

  return objp;
}





//--------------------------------------------------------------------------------
//---- Configpp::run()
//--------------------------------------------------------------------------------

bool Configpp::run()
{
  // run any registered simulation/GUI

  // first, if user didn't call done() then we call it, to tell all classes we are done with main parsing
  try
    {
      if (!diddone)
	done();
      
      // now run any simulations
      timermanagerp->takeover();
    }
  catch (CfpError cfperror)
    {
      cfperror.report(this);
      if (cfperror.isfatal())
	throw cfperror;
      return false;
    }
  catch (...)
    {
      error("unexpected exception encountered while running configpp.","",internal,true);
      return false;
    }
  return true;
}




//--------------------------------------------------------------------------------
//---- Configpp::done() -----------------------------------------------------------
//--------------------------------------------------------------------------------

void Configpp::done()
{
  // this is called when parsing is done, before we delete and leave
  // now to postinterpret
  factoryp->postinterpret();
  diddone=true;
}






//--------------------------------------------------------------------------------
//---- Configpp::quit_request()
//--------------------------------------------------------------------------------

void Configpp::quit_request()
{
  // someone has requested that we quit WHILE running the simulation (timers)
  // so abide by that request by sending the timermanager a single to quit
  timermanagerp->kill();
}







void Configpp::destructobjtree(CfpObject *objp)
{
  // remove obj and all its children from the table and scope, then delete it
  CfpObject *cobjp=NULL;
  deque<CfpObject *>::iterator p=scopestack.begin();
  bool deleting=false;
  
  while (p!=scopestack.end())
    {
      cobjp=*p;
      // ask the object if it is the type
      if (cobjp==objp)
	deleting=true;
      if (deleting)
	{
	  p=scopestack.erase(p);
	  tableremove(cobjp);
	  cobjp->destruct();
	  delete cobjp;
	  continue;
	}
      ++p;
    }

  if (!deleting)
    {
      // we didn't find it in the current scope, so we look for it in the objtable
      tableremove(objp);
      objp->destruct();
      delete objp;
    }
}


void Configpp::tableremove(CfpObject *objp)
{
  // locate the object in the object table and remove it
  map<string,CfpObject *>::iterator p=objtable.begin();
  CfpObject *cobjp;

  while (p!=objtable.end())
    {
      cobjp=p->second;
      if (objp==cobjp)
	{
	  objtable.erase(p);
	  break;
	}
      ++p;
    }
}



CfpVAccessor *Configpp::findvaccessor(string name,int &index)
{
  // look up the accessor and return it (along with
  CfpVAccessor *vap;

  // ideally we would like to parse out any vector address
  string identifier=name;

  // now get the main accessor pointer
  vap=lookupVAcc(identifier,false);

  // not lookup ay index
  index=-1;

  return vap;
}



string Configpp::tokenreplacement(string instring)
{
  // we allow all tokens to specify a kind of preprocessing textual replacement.
  //  all occurences within a token of a word (without spaces) surrounded by bars (|)
  //  will be replaced, prior to interpretation, by the value of the variable
  //  with name indicated by word.
  // ie  if the variable index was currently == 23, then the expression "Linkone.|index|.angle" 
  //  would be translated into "Linkone.23.angle".
  int slen,sloc,eloc;
  string identifier;
  string ivalue;
  bool bretv;
  CfpVal isval;
  int spos;

  sloc=0;
  slen=instring.length();
  while ((sloc=instring.find('|',sloc))>=0)
    {
      if (sloc==slen-1)
	{
	  //	  it's the last character, so def. no good.
	  break;
	}
      eloc=instring.find('|',sloc+1);
      if (eloc<0)
	{
	  // no closing bar, so just exit
	  break;
	}
      else if (eloc==sloc+1)
	{
	  // it's a double bar (||), which we skip over
	  sloc=eloc+1;
	}

      // okay we have a potential replacement, so get the identifier
      identifier=instring.substr(sloc+1,eloc-(sloc+1));
      
      if ((spos=identifier.find(' ',0))>=0)
	{
	  // false alarm, there is a space inbetween the bars
	  sloc=eloc+1;
	  continue;
	}
      // okay we think we found one
      //      cout << "... Attempting to find replacement for '"+identifier+"'.\n";
      bretv=expressionval(identifier,isval);
      if (bretv==false)
	{
	  // couldn't find the identifier named
	  warning("Attempting to do preprocessor textual replacement on string '"+instring+"' but could not find any identifier by the name of '"+identifier+"'.");
	  break;
	}
      // found it so now replace it
      ivalue=isval.get_string();
      instring=instring.substr(0,sloc)+ivalue+instring.substr(eloc+1,slen-(eloc+1));
      slen=instring.length();
      // and loop to look for the next one
    }

  return instring;
}







bool Configpp::expressionval(string identifier,CfpVal &val)
{
  // find the value of the identifier specified, if succesful, set 'val' to it and return true
  //  return false on failure to find it.
  string vecstr;
  int sloc,eloc;
  int slen;
  int index;
  CfpVAccessor *varp;

  // first split off any vector address
  slen=identifier.length();
  sloc=identifier.find('[',0);
  if (sloc>0)
    {
      if (sloc==slen-1)
	return false;
      eloc=slen-1;
      if (eloc!=slen-1 || eloc==sloc+1)
	return false;
      vecstr=identifier.substr(sloc+1,(slen-1)-(sloc+1));
      identifier=identifier.substr(0,sloc);
    }
  else
    vecstr="";
  
  //  cout << "... In expressionval with ("+identifier+","+vecstr+").\n";

  // lookup identifier
  varp=lookupVAcc(identifier,false);
  if (!varp)
    return false;

  // okay we found a var/arg accessor
  val.set_vaccesspointer(varp);

  // now handle the vector address if any
  if (vecstr!="")
    {
      // okay, now ask the accessor to tell us the index it corresponds to
      index=varp->parseindex(vecstr);
      val.set_vaccessindex(index);
    }

  return true;
}







// }
// Namespace close-brace
