/***********************************************************************

 LinearAlgebra.c
 Tools for solving Linear Systems of Equations
 Copyright (C) 1993
        Lyle Hall 1993/94
        Eric C. Welton 1993/94
        Beckman Institute - Urbana, Illinois
        For Neuronal Pattern Analysis Group and Hexapod Project

        These libraries are free and may be redistributed and/or modified,
        provided the above notice and this paragraph are included.
        These libraries may not be sold and are distributed without
        any warranty or implication of fitness for a purpose.

Jesse Reichler 2/28/96
adding double version - should be done with templates later
memory leakage errors fixed 2/96
***********************************************************************/



#include <stdio.h>
#include <math.h>
#include "LinearAlgebra.h"

FVector Solve(const FMatrix& mat,const FVector& vect) {
FVector returnvec;
int i,j,k,kk;
int tmp;
float ftmp;

 // check to see that the matrix is square
 if(mat.rows() != mat.columns()) {
	fprintf(stderr,"Solve::Matrix is not square.\n");
 	return FVector();
	}

 // check to see that the vector is as tall as the matrix wide
 if(vect.size() != mat.columns()) {
	fprintf(stderr,"Solve::Vector does not match Matrix dimensions.\n");
 	return FVector();
	}

 // 
 const int N = mat.rows();              // matrix dimension
 FMatrix LU(mat);                        // a copy of A to compute LU in
 int* pi = new int[N];                  // a permutation of rows of A and b (P)


 // First compute the LUP decomposition (CLR p.759).
      // Put L under LU's diagonal, U on and over it, and P implicitly in pi.
for (i=0; i<N; ++i) pi[i] = i;

for (k=0; k<N-1; ++k)
{ // columns 0 .. n-1
   float p = 0;
   // rows k .. n
   // find max in column (on|under diag)
	for (i=k; i<N; ++i)
                if(fabs(LU(i,k)) > p) {
			p = fabs(LU(i,k));
			kk = i;
			}
	if( p == 0 ) {
		fprintf(stderr,"ERRORa1 in ::solve -> matrix A is singular.\n");
		return FVector();
		}
	for (i=0; i<N; ++i) {			// swap row k and kk (ensuring that each
		ftmp=LU(kk,i);			// elt on diag is > anything under it)
		LU(kk,i)=LU(k,i);
		LU(k,i)=ftmp;
		}
	tmp=pi[kk];				// and record the swap in pi
	pi[kk]=pi[k];
	pi[k]=tmp;

	for (i=k+1; i<N; ++i) {
		LU(i,k) = LU(i,k) / LU(k,k);  // divide elts under max by max
		for (j=k+1; j<N; ++j)        // Schur compl elts under/right of max
		LU(i,j) = LU(i,j) - LU(i,k) * LU(k,j);
		}
	}


 // Now solve for y and finally x (CLR p.753).
float* y = new float[N];
       for (i=0; i<N; ++i){
	   float sum = 0;
           for (j=0; j<i; ++j)
              sum += LU(i,j) * y[j];
   	   y[i] = vect(pi[i]) - sum;
	}

  float* x = new float[N];
  for (i=N-1; i>=0; --i) {
    float sum = 0;
    for (j=i+1; j<N; ++j) sum += LU(i,j) * x[j];
    if ( isnan(x[i]) )
    {
      puts("ERRORb1 in ::solve -> matrix A is singular.");
      return FVector();
    }
    x[i] = (y[i] - sum) / LU(i,i);
  }

// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// JESSE ADDED 2/12/96 NIGHTMARE!!!!
//  MEMORY LEAKAGE BUGS

returnvec=FVector(mat.rows(),x,ColV);
delete [] pi;
delete [] x;
delete [] y;

return returnvec;
}








// DOUBLE VERSION 2/29/96
// 10/1/96 Getting vector to high error even on simple test


DVector Solve(const DMatrix& mat,const DVector& vect) {
DVector returnvec;
int i,j,k,kk;
int tmp;
double dtmp;

 // check to see that the matrix is square
 if(mat.rows() != mat.columns()) {
	fprintf(stderr,"Solve::Matrix is not square.\n");
 	return DVector();
	}

 // check to see that the vector is as tall as the matrix wide
 if(vect.size() != mat.columns()) {
	fprintf(stderr,"Solve::Vector does not match Matrix dimensions.\n");
 	return DVector();
	}

 // 
 const int N = mat.rows();              // matrix dimension
 DMatrix LU(mat);                        // a copy of A to compute LU in
 int* pi = new int[N];                  // a permutation of rows of A and b (P)

 // First compute the LUP decomposition (CLR p.759).
 // Put L under LU's diagonal, U on and over it, and P implicitly in pi.
 for (i=0; i<N; ++i) pi[i] = i;

 for (k=0; k<N-1; ++k) { // columns 0 .. n-1
	double p = 0;

	// rows k .. n
	// find max in column (on|under diag)
	for (i=k; i<N; ++i)
		if(fabs(LU(i,k)) > p) {
			p = fabs(LU(i,k));
			kk = i;
			}
		if( p == 0 ) {
			fprintf(stderr,"ERRORa2 in ::solve -> matrix A is singular.\n");
			return DVector();
			}
		for (i=0; i<N; ++i) {			// swap row k and kk (ensuring that each
			dtmp=LU(kk,i);			// elt on diag is > anything under it)
			LU(kk,i)=LU(k,i);
			LU(k,i)=dtmp;
			}
		tmp=pi[kk];					// and record the swap in pi
		pi[kk]=pi[k];
		pi[k]=tmp;

		for (i=k+1; i<N; ++i) {
			LU(i,k) = LU(i,k) / LU(k,k);  // divide elts under max by max
			for (j=k+1; j<N; ++j)        // Schur compl elts under/right of max
				LU(i,j) = LU(i,j) - LU(i,k) * LU(k,j);
			}
	}

 // Now solve for y and finally x (CLR p.753).
 double* y = new double[N];
 for (i=0; i<N; ++i) {
	double sum = 0;
	for (j=0; j<i; ++j) sum += LU(i,j) * y[j];
	y[i] = vect(pi[i]) - sum;
	}

  double* x = new double[N];
  for (i=N-1; i>=0; --i) {
    double sum = 0;
    for (j=i+1; j<N; ++j) sum += LU(i,j) * x[j];
    if ( isnan(x[i]) )
    {
      puts("ERRORb2 in ::solve -> matrix A is singular.");
      return DVector();
    }
    x[i] = (y[i] - sum) / LU(i,i);
  }

// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// JESSE ADDED 2/12/96 NIGHTMARE!!!!
//  MEMORY LEAKAGE BUGS

returnvec=DVector(mat.rows(),x,ColV);
delete [] pi;
delete [] x;
delete [] y;

return returnvec;
}




