/***********************************************************************

 TransformationDMatrix.h
 Transformation DMatrix classes including Scaling, Rotation, and Translation
 Copyright (C) 1993
        Eric C. Welton 1993/94
        Beckman Institute - Urbana, Illinois
        For Neuronal Pattern Analysis Group and Hexapod Project

        These libraries are free and may be redistributed and/or modified,
        provided the above notice and this paragraph are included.
        These libraries may not be sold and are distributed without
        any warranty or implication of fitness for a purpose.

 General Information

        The DMatrix library provides fundamental matrix operations.

***********************************************************************/

#ifndef TRANSFORMATION_MATRIX_HDR
#define TRANSFORMATION_MATRIX_HDR

#include <math.h>
#include <DMatrix.h>
#include <FMatrix.h>


class TDMatrixR2:public DMatrix {
public:
 TDMatrixR2():DMatrix(2,2) { ident(); }
 TDMatrixR2& clear() { fill(0.0); return (*this); }
 TDMatrixR2& ident() { fill(0.0); register int max=width*height;
   register int inc=width+1; for(register int i=0;i<max;i+=inc) data[i]=1.0; return (*this); }
 };

class TDMatrixHR2:public DMatrix {
public:
 TDMatrixHR2():DMatrix(3,3) { ident(); }
 TDMatrixHR2& clear() { fill(0.0); data[width*height-1]=1.0; return (*this); }
 TDMatrixHR2& ident() { fill(0.0); register int max=width*height;
   register int inc=width+1; for(register int i=0;i<max;i+=inc) data[i]=1.0; return (*this); }
 };


class TDMatrixR3:public DMatrix {
public:
 TDMatrixR3():DMatrix(3,3) { ident(); }
 TDMatrixR3& clear() { fill(0.0); return (*this); }
 TDMatrixR3& ident() { fill(0.0); register int max=width*height;
   register int inc=width+1; for(register int i=0;i<max;i+=inc) data[i]=1.0; return (*this); }
 };

class TDMatrixHR3:public DMatrix {
public:
 TDMatrixHR3():DMatrix(4,4) { ident(); }
 TDMatrixHR3& clear() { fill(0.0); data[width*height-1]=1.0; return (*this); }
 TDMatrixHR3& ident() { fill(0.0); register int max=width*height;
   register int inc=width+1; for(register int i=0;i<max;i+=inc) data[i]=1.0; return (*this); }
 };

class TFMatrixR2:public FMatrix {
public:
 TFMatrixR2():FMatrix(2,2) { ident(); }
 TFMatrixR2& clear() { fill(0.0); return (*this); }
 TFMatrixR2& ident() { fill(0.0); register int max=width*height;
   register int inc=width+1; for(register int i=0;i<max;i+=inc) data[i]=1.0; return (*this); }
 };

class TFMatrixHR2:public FMatrix {
public:
 TFMatrixHR2():FMatrix(3,3) { ident(); }
 TFMatrixHR2& clear() { fill(0.0); data[width*height-1]=1.0; return (*this); }
 TFMatrixHR2& ident() { fill(0.0); register int max=width*height;
   register int inc=width+1; for(register int i=0;i<max;i+=inc) data[i]=1.0; return (*this); }
 };


class TFMatrixR3:public FMatrix {
public:
 TFMatrixR3():FMatrix(3,3) { ident(); }
 TFMatrixR3& clear() { fill(0.0); return (*this); }
 TFMatrixR3& ident() { fill(0.0); register int max=width*height;
   register int inc=width+1; for(register int i=0;i<max;i+=inc) data[i]=1.0; return (*this); }
 };

class TFMatrixHR3:public FMatrix {
public:
 TFMatrixHR3():FMatrix(4,4) { ident(); }
 TFMatrixHR3& clear() { fill(0.0); data[width*height-1]=1.0; return (*this); }
 TFMatrixHR3& ident() { fill(0.0); register int max=width*height;
   register int inc=width+1; for(register int i=0;i<max;i+=inc) data[i]=1.0; return (*this); }
 };



extern TDMatrixR2 ScalingDMatrixR2(double x,double y);
extern TDMatrixR2 ScalingDMatrixR21(double x);
extern TDMatrixR2 ScalingDMatrixR22(double y);
inline TDMatrixR2 ScalingDMatrixR2(float x,float y)
   { return ScalingDMatrixR2((double)x,(double)y); }
inline TDMatrixR2 ScalingDMatrixR21(float x)
   { return ScalingDMatrixR21((double)x); }
inline TDMatrixR2 ScalingDMatrixR22(float y)
   { return ScalingDMatrixR22((double)y); }

extern TDMatrixR3 ScalingDMatrixR3(double x,double y,double z);
extern TDMatrixR3 ScalingDMatrixR31(double x);
extern TDMatrixR3 ScalingDMatrixR32(double y);
extern TDMatrixR3 ScalingDMatrixR33(double z);
inline TDMatrixR3 ScalingDMatrixR3(float x,float y,float z)
   { return ScalingDMatrixR3((double)x,(double)y,(double)z); }
inline TDMatrixR3 ScalingDMatrixR31(float x)
   { return ScalingDMatrixR31((double)x); }
inline TDMatrixR3 ScalingDMatrixR32(float y)
   { return ScalingDMatrixR32((double)y); }
inline TDMatrixR3 ScalingDMatrixR33(float z)
   { return ScalingDMatrixR33((double)z); }

extern TDMatrixHR2 ScalingDMatrixHR2(double x,double y);
extern TDMatrixHR2 ScalingDMatrixHR21(double x);
extern TDMatrixHR2 ScalingDMatrixHR22(double y);
inline TDMatrixHR2 ScalingDMatrixHR2(float x,float y)
   { return ScalingDMatrixHR2((double)x,(double)y); }
inline TDMatrixHR2 ScalingDMatrixHR21(float x)
   { return ScalingDMatrixHR21((double)x); }
inline TDMatrixHR2 ScalingDMatrixHR22(float y)
   { return ScalingDMatrixHR22((double)y); }

extern TDMatrixHR3 ScalingDMatrixHR3(double x,double y,double z);
extern TDMatrixHR3 ScalingDMatrixHR31(double x);
extern TDMatrixHR3 ScalingDMatrixHR32(double y);
extern TDMatrixHR3 ScalingDMatrixHR33(double z);
inline TDMatrixHR3 ScalingDMatrixHR3(float x,float y,float z)
   { return ScalingDMatrixHR3((double)x,(double)y,(double)z); }
inline TDMatrixHR3 ScalingDMatrixHR31(float x)
   { return ScalingDMatrixHR31((double)x); }
inline TDMatrixHR3 ScalingDMatrixHR32(float y)
   { return ScalingDMatrixHR32((double)y); }
inline TDMatrixHR3 ScalingDMatrixHR33(float z)
   { return ScalingDMatrixHR33((double)z); }

extern TDMatrixHR2 TranslationDMatrixHR2(double x,double y);
extern TDMatrixHR2 TranslationDMatrixHR21(double x);
extern TDMatrixHR2 TranslationDMatrixHR22(double y);
inline TDMatrixHR2 TranslationDMatrixHR2(float x,float y)
   { return TranslationDMatrixHR2((double)x,(double)y); }
inline TDMatrixHR2 TranslationDMatrixHR21(float x)
   { return TranslationDMatrixHR21((double)x); }
inline TDMatrixHR2 TranslationDMatrixHR22(float y)
   { return TranslationDMatrixHR22((double)y); }

extern TDMatrixHR3 TranslationDMatrixHR3(double x,double y,double z);
extern TDMatrixHR3 TranslationDMatrixHR31(double x);
extern TDMatrixHR3 TranslationDMatrixHR32(double y);
extern TDMatrixHR3 TranslationDMatrixHR33(double z);
inline TDMatrixHR3 TranslationDMatrixHR3(float x,float y,float z)
   { return TranslationDMatrixHR3((double)x,(double)y,(double)z); }
inline TDMatrixHR3 TranslationDMatrixHR31(float x)
   { return TranslationDMatrixHR31((double)x); }
inline TDMatrixHR3 TranslationDMatrixHR32(float y)
   { return TranslationDMatrixHR31((double)y); }
inline TDMatrixHR3 TranslationDMatrixHR33(float z)
   { return TranslationDMatrixHR31((double)z); }

extern TDMatrixR2 RotationDMatrixR2(double theta);
extern TDMatrixR3 RotationDMatrixR31(double theta);
extern TDMatrixR3 RotationDMatrixR32(double theta);
extern TDMatrixR3 RotationDMatrixR33(double theta);
inline TDMatrixR2 RotationDMatrixR2(float theta)
   { return RotationDMatrixR2((double)theta); }
inline TDMatrixR3 RotationDMatrixR31(float theta)
   { return RotationDMatrixR31((double)theta); }
inline TDMatrixR3 RotationDMatrixR32(float theta)
   { return RotationDMatrixR32((double)theta); }
inline TDMatrixR3 RotationDMatrixR33(float theta)
   { return RotationDMatrixR33((double)theta); }

extern TDMatrixHR2 RotationDMatrixHR2(double theta);
extern TDMatrixHR3 RotationDMatrixHR31(double theta);
extern TDMatrixHR3 RotationDMatrixHR32(double theta);
extern TDMatrixHR3 RotationDMatrixHR33(double theta);
inline TDMatrixHR2 RotationDMatrixHR2(float theta)
   { return RotationDMatrixHR2((double)theta); }
inline TDMatrixHR3 RotationDMatrixHR31(float theta)
   { return RotationDMatrixHR31((double)theta); }
inline TDMatrixHR3 RotationDMatrixHR32(float theta)
   { return RotationDMatrixHR32((double)theta); }
inline TDMatrixHR3 RotationDMatrixHR33(float theta)
   { return RotationDMatrixHR33((double)theta); }






extern TFMatrixR2 ScalingFMatrixR2(float x,float y);
extern TFMatrixR2 ScalingFMatrixR21(float x);
extern TFMatrixR2 ScalingFMatrixR22(float y);
inline TFMatrixR2 ScalingFMatrixR2(double x,double y)
   { return ScalingFMatrixR2((float)x,(float)y); }
inline TFMatrixR2 ScalingFMatrixR21(double x)
   { return ScalingFMatrixR21((float)x); }
inline TFMatrixR2 ScalingFMatrixR22(double y)
   { return ScalingFMatrixR22((float)y); }

extern TFMatrixR3 ScalingFMatrixR3(float x,float y,float z);
extern TFMatrixR3 ScalingFMatrixR31(float x);
extern TFMatrixR3 ScalingFMatrixR32(float y);
extern TFMatrixR3 ScalingFMatrixR33(float z);
inline TFMatrixR3 ScalingFMatrixR3(double x,double y,double z)
   { return ScalingFMatrixR3((float)x,(float)y,(float)z); }
inline TFMatrixR3 ScalingFMatrixR31(double x)
   { return ScalingFMatrixR31((float)x); }
inline TFMatrixR3 ScalingFMatrixR32(double y)
   { return ScalingFMatrixR32((float)y); }
inline TFMatrixR3 ScalingFMatrixR33(double z)
   { return ScalingFMatrixR33((float)z); }

extern TFMatrixHR2 ScalingFMatrixHR2(float x,float y);
extern TFMatrixHR2 ScalingFMatrixHR21(float x);
extern TFMatrixHR2 ScalingFMatrixHR22(float y);
inline TFMatrixHR2 ScalingFMatrixHR2(double x,double y)
   { return ScalingFMatrixHR2((float)x,(float)y); }
inline TFMatrixHR2 ScalingFMatrixHR21(double x)
   { return ScalingFMatrixHR21((float)x); }
inline TFMatrixHR2 ScalingFMatrixHR22(double y)
   { return ScalingFMatrixHR22((float)y); }

extern TFMatrixHR3 ScalingFMatrixHR3(float x,float y,float z);
extern TFMatrixHR3 ScalingFMatrixHR31(float x);
extern TFMatrixHR3 ScalingFMatrixHR32(float y);
extern TFMatrixHR3 ScalingFMatrixHR33(float z);
inline TFMatrixHR3 ScalingFMatrixHR3(double x,double y,double z)
   { return ScalingFMatrixHR3((float)x,(float)y,(float)z); }
inline TFMatrixHR3 ScalingFMatrixHR31(double x)
   { return ScalingFMatrixHR31((float)x); }
inline TFMatrixHR3 ScalingFMatrixHR32(double y)
   { return ScalingFMatrixHR32((float)y); }
inline TFMatrixHR3 ScalingFMatrixHR33(double z)
   { return ScalingFMatrixHR33((float)z); }

extern TFMatrixHR2 TranslationFMatrixHR2(float x,float y);
extern TFMatrixHR2 TranslationFMatrixHR21(float x);
extern TFMatrixHR2 TranslationFMatrixHR22(float y);
inline TFMatrixHR2 TranslationFMatrixHR2(double x,double y)
   { return TranslationFMatrixHR2((float)x,(float)y); }
inline TFMatrixHR2 TranslationFMatrixHR21(double x)
   { return TranslationFMatrixHR21((float)x); }
inline TFMatrixHR2 TranslationFMatrixHR22(double y)
   { return TranslationFMatrixHR22((float)y); }

extern TFMatrixHR3 TranslationFMatrixHR3(float x,float y,float z);
extern TFMatrixHR3 TranslationFMatrixHR31(float x);
extern TFMatrixHR3 TranslationFMatrixHR32(float y);
extern TFMatrixHR3 TranslationFMatrixHR33(float z);
inline TFMatrixHR3 TranslationFMatrixHR3(double x,double y,double z)
   { return TranslationFMatrixHR3((float)x,(float)y,(float)z); }
inline TFMatrixHR3 TranslationFMatrixHR31(double x)
   { return TranslationFMatrixHR31((float)x); }
inline TFMatrixHR3 TranslationFMatrixHR32(double y)
   { return TranslationFMatrixHR31((float)y); }
inline TFMatrixHR3 TranslationFMatrixHR33(double z)
   { return TranslationFMatrixHR31((float)z); }

extern TFMatrixR2 RotationFMatrixR2(float theta);
extern TFMatrixR3 RotationFMatrixR31(float theta);
extern TFMatrixR3 RotationFMatrixR32(float theta);
extern TFMatrixR3 RotationFMatrixR33(float theta);
inline TFMatrixR2 RotationFMatrixR2(double theta)
   { return RotationFMatrixR2((float)theta); }
inline TFMatrixR3 RotationFMatrixR31(double theta)
   { return RotationFMatrixR31((float)theta); }
inline TFMatrixR3 RotationFMatrixR32(double theta)
   { return RotationFMatrixR32((float)theta); }
inline TFMatrixR3 RotationFMatrixR33(double theta)
   { return RotationFMatrixR33((float)theta); }

extern TFMatrixHR2 RotationFMatrixHR2(float theta);
extern TFMatrixHR3 RotationFMatrixHR31(float theta);
extern TFMatrixHR3 RotationFMatrixHR32(float theta);
extern TFMatrixHR3 RotationFMatrixHR33(float theta);
inline TFMatrixHR2 RotationFMatrixHR2(double theta)
   { return RotationFMatrixHR2((float)theta); }
inline TFMatrixHR3 RotationFMatrixHR31(double theta)
   { return RotationFMatrixHR31((float)theta); }
inline TFMatrixHR3 RotationFMatrixHR32(double theta)
   { return RotationFMatrixHR32((float)theta); }
inline TFMatrixHR3 RotationFMatrixHR33(double theta)
   { return RotationFMatrixHR33((float)theta); }






//======================================================================
//
// Default type establishment
//
#ifdef __Concat3
#undef __Concat3
#endif
#ifdef _Concat3
#undef _Concat3
#endif
#define __Concat3(a,b,c) a##b##c
#define _Concat3(a,b,c) __Concat3(a,b,c)

typedef _Concat3(T,DEF_PREC_MATRIX_NAME,R2)  TMatrixR2;
typedef _Concat3(T,DEF_PREC_MATRIX_NAME,R3)  TMatrixR3;
typedef _Concat3(T,DEF_PREC_MATRIX_NAME,HR2) TMatrixHR2;
typedef _Concat3(T,DEF_PREC_MATRIX_NAME,HR3) TMatrixHR3;

#define ScalingMatrixR2 _Concat3(Scaling,DEF_PREC_MATRIX_NAME,R2)
#define ScalingMatrixR21 _Concat3(Scaling,DEF_PREC_MATRIX_NAME,R21)
#define ScalingMatrixR22 _Concat3(Scaling,DEF_PREC_MATRIX_NAME,R22)
#define ScalingMatrixR3 _Concat3(Scaling,DEF_PREC_MATRIX_NAME,R3)
#define ScalingMatrixR31 _Concat3(Scaling,DEF_PREC_MATRIX_NAME,R31)
#define ScalingMatrixR32 _Concat3(Scaling,DEF_PREC_MATRIX_NAME,R32)
#define ScalingMatrixR33 _Concat3(Scaling,DEF_PREC_MATRIX_NAME,R33)
#define ScalingMatrixHR2 _Concat3(Scaling,DEF_PREC_MATRIX_NAME,HR2)
#define ScalingMatrixHR21 _Concat3(Scaling,DEF_PREC_MATRIX_NAME,HR21)
#define ScalingMatrixHR22 _Concat3(Scaling,DEF_PREC_MATRIX_NAME,HR22)
#define ScalingMatrixHR3 _Concat3(Scaling,DEF_PREC_MATRIX_NAME,HR3)
#define ScalingMatrixHR31 _Concat3(Scaling,DEF_PREC_MATRIX_NAME,HR31)
#define ScalingMatrixHR32 _Concat3(Scaling,DEF_PREC_MATRIX_NAME,HR32)
#define ScalingMatrixHR33 _Concat3(Scaling,DEF_PREC_MATRIX_NAME,HR33)

#define TranslationMatrixHR2 _Concat3(Translation,DEF_PREC_MATRIX_NAME,HR2)
#define TranslationMatrixHR21 _Concat3(Translation,DEF_PREC_MATRIX_NAME,HR21)
#define TranslationMatrixHR22 _Concat3(Translation,DEF_PREC_MATRIX_NAME,HR22)
#define TranslationMatrixHR3 _Concat3(Translation,DEF_PREC_MATRIX_NAME,HR3)
#define TranslationMatrixHR31 _Concat3(Translation,DEF_PREC_MATRIX_NAME,HR31)
#define TranslationMatrixHR32 _Concat3(Translation,DEF_PREC_MATRIX_NAME,HR32)
#define TranslationMatrixHR33 _Concat3(Translation,DEF_PREC_MATRIX_NAME,HR33)

#define RotationMatrixR2 _Concat3(Rotation,DEF_PREC_MATRIX_NAME,R2)
#define RotationMatrixHR2 _Concat3(Rotation,DEF_PREC_MATRIX_NAME,HR2)
#define RotationMatrixR31 _Concat3(Rotation,DEF_PREC_MATRIX_NAME,R31)
#define RotationMatrixR32 _Concat3(Rotation,DEF_PREC_MATRIX_NAME,R32)
#define RotationMatrixR33 _Concat3(Rotation,DEF_PREC_MATRIX_NAME,R33)
#define RotationMatrixHR31 _Concat3(Rotation,DEF_PREC_MATRIX_NAME,HR31)
#define RotationMatrixHR32 _Concat3(Rotation,DEF_PREC_MATRIX_NAME,HR32)
#define RotationMatrixHR33 _Concat3(Rotation,DEF_PREC_MATRIX_NAME,HR33)

#endif

