// JRinterpolate.C

#include "JRinterpolater.h"
#include <stdio.h>










//--------------------------------------------------------------------------------
//---- JRinterpolater()
//--------------------------------------------------------------------------------
//:a simple linear interpolater
// doesn't use the nice stl stuff or do error checking during interpolation since
// it is meant to be fast; was written to do interpolation of color ramps during
// drawing opengl surface plots.

JRinterpolater::JRinterpolater(float inmin,float inmax,int inorder)
{
  // initialize
  min=inmin;
  max=inmax;
  dif=max-min;
  order=inorder;
  controlpoints=0;
  datap=NULL;
  
  // reserve space for 10 points (can be expanded later)
  allocate(10);
}

JRinterpolater::~JRinterpolater()
{
  // free memory
  deallocate();
}


void JRinterpolater::addcontrolpoint(float *values)
{
  // add a new point, assumed of order 'order'
  int count;
  float *vloc=newpoint();

  for (count=0;count<order;++count)
    *vloc++=*values++;
}


void JRinterpolater::addcontrolpoint(float val1)
{
  // add a new point of order 1
  float *vloc=newpoint();

  if (order!=1)
    {
      cout << "JRinterpolater error: internal order is "<<order<<", but trying to set order 1.\n";
      return;
    }
  *vloc=val1;
}

void JRinterpolater::addcontrolpoint(float val1,float val2)
{
  // shortcut, add a new point of order 2
  float *vloc=newpoint();
  
  if (order!=2)
    {
      cout << "JRinterpolater error: internal order is "<<order<<", but trying to set order 2.\n";
      return;
    }
  *vloc++=val1;
  *vloc++=val2;
}

void JRinterpolater::addcontrolpoint(float val1,float val2,float val3)
{
  // shortcut, add a new point of order 3
  float *vloc=newpoint();
  
  if (order!=3)
    {
      cout << "JRinterpolater error: internal order is "<<order<<", but trying to set order 3.\n";
      return;
    }
  *vloc++=val1;
  *vloc++=val2;
  *vloc++=val3;
}

float *JRinterpolater::newpoint()
{
  // make sure there is room for another point
  if (controlpoints==maxpoints)
    {
      // we are at the max points already so we need to reallocate
      allocate(maxpoints*2+1);
    }

  // now return point to first element of the new point
  ++controlpoints;
  return &datap[(controlpoints-1)*order];
}


void JRinterpolater::allocate(int numpoints)
{
  // reserve/expand space
  float *newdatap;
  int count,count2;

  // reserve new space
  newdatap=new float[numpoints*order];

  // copy old values, if any
  for (count=0;count<controlpoints;++count)
    {
      for (count2=0;count2<order;++count2)
	newdatap[count*order+count2]=datap[count*order+count2];
    }

  // release old space
  deallocate();

  // now let permanent pointer point to new collection
  datap=newdatap;

  // lastly, adjust maxpoints
  maxpoints=numpoints;
}


void JRinterpolater::deallocate()
{
  // free the space
  if (datap)
    delete datap;
  datap=NULL;
}



void JRinterpolater::interpolate(float lookpos,float *ivalues)
{
  // simple LINEAR interpolation
  // look up surrounding elemtents of list and interpolate the values
  //  putting the output vector, of length order, into ivalues
  int count;
  float pos2,pos;
  float weightratio,iweightratio;
  int minpoint,maxpoint;
  float *minpointloc,*maxpointloc;

  if (controlpoints<2)
    {
      cout << "JRinterpolater error: must be at least 2 control points in interpolater before it can be queried.\n";
      return;
    }

  // figure out where it lies in interpolation table, and distance from min&max points
  if (lookpos<=min)
    {
      minpoint=maxpoint=0;
      weightratio=.5;
    }
  else if (lookpos>=max)
    {
      minpoint=maxpoint=controlpoints-1;
      weightratio=.5;
    }
  else
    {
      // figure out which 2 points it's inbetween
      pos2=(lookpos-min)/dif;
      pos=pos2*(controlpoints-1);
      minpoint=(int)pos;
      if (minpoint==pos)
	{
	  maxpoint=minpoint;
	  weightratio=.5;
	}
      else
	{
	  maxpoint=minpoint+1;
	  weightratio=(lookpos-position(minpoint))/(position(maxpoint)-position(minpoint));
	}
    }

  //  cout <<"JRinterpolate query.  In with "<<lookpos<<" ("<<min<<","<<max<<") = points "<<minpoint<<","<<maxpoint<<".\n";

  iweightratio=1.0-weightratio;
  minpointloc=location(minpoint);
  maxpointloc=location(minpoint);
  for (count=0;count<order;++count)
    *ivalues++=((*minpointloc++)*iweightratio)+((*maxpointloc++)*weightratio);
}




