// JRutils.C

#include "JRutils.h"
#include <stdio.h>
#include <cstdarg>

//Not implemented in gcc yet (2.8.1)
//using namespace std;
//namespace JR{






//--------------------------------------------------------------------------------
//---- JRSearchFileOpen() --------------------------------------------------------
//--------------------------------------------------------------------------------
//: Find and open the file stream for input, searching through directory list to find.
// Subtlety: if we do find it someplace unusual, change filename to the actual location
//  what we do is we search our list of directories, checking for first the exact filename passed,
//  and then the filename+extension.

fstream *JRSearchFileOpen(string filename, list<string> *dirs, string &outfilename, ios::openmode mode, const string ext)
{
  // Find and open the file stream for input, searching through directory list to find.
  // Subtlety: if we do find it someplace unusual, change filename to the actual location
  //  what we do is we search our list of directories, checking for first the exact filename passed,
  //  and then the filename+extension.
  string tempname;
  fstream *fsp;
  list<string>::iterator dirp=dirs->begin();
  
  fsp=new fstream();
  tempname=filename;
  while (dirp!=dirs->end())
    {
      // okay try to open a file here
      tempname=*dirp + filename;
      fsp->open(tempname.c_str(),mode);
      if (fsp->is_open())
	{
	  outfilename=tempname;
	  return fsp;
	}
      if (ext != "")
	{
	  // no good, try to open a file here with the extension
	  tempname=*dirp + filename +ext;
	  fsp->open(tempname.c_str(),mode);
	  if (fsp->is_open())
	    {
	      outfilename=tempname;
	      return fsp;
	    }
	}
      // no good, try the next dir
      dirp++;
    }

  // if we fall through to here then we couldn't find a file
  delete fsp;
  return NULL;
}









//--------------------------------------------------------------------------------
//---- JRShowStreamPos() ---------------------------------------------------------
//--------------------------------------------------------------------------------

//: Display a specified line from a file and highlight selected section.
// Finds the closest line and character in the file and display the line,
//  line #, and put carrots under the error location.
// This is mainly used as an error-reporting helper function:

void JRShowStreamPos(ostream *out, string sname, istream *sisp, long sline, long slinepos, string ename, istream *eisp, long eline, long elinepos)
{
  // Error Reporting helper function:
  // Finds the closest line and character in the file and display the line,
  //  line #, and put carrots under the error location.
  // Note: there are varying degrees of info we might have about the location.
  string prefstr="     ";
  string preflinestr="        ";

  // make sure we have a stream, and that it is open
  if (sisp==NULL)
    return;
  if (!*sisp)
    return;

  //  printf("%ld,%ld  to  %ld,%ld\n",sline,slinepos,eline,elinepos);

  // figure out what format we are going to use
  if (eisp==NULL || eisp==sisp)
    {
      // starts and ends in the same file
      if (sline==0)
	{
	  // no line number, so just tell them the stream name.
	  *out << prefstr << "In " << sname << ".\n";
	}
      else if (eline == sline || eline==0)
	{
	  // start and end on same line.
	  // so tell them the line number and show them the line and (possibly) position
	  *out << prefstr << sname << " line " << sline << ":\n";
	  *out << preflinestr;
	  JRShowStreamLine(out,sisp,sline,slinepos-1,elinepos-1);
	}
      else if (eline==-1)
	{
	  // ends at eof
	  *out << prefstr << sname << " lines " << sline << " and forward:\n";
	  *out << preflinestr;
	  JRShowStreamLine(out,sisp,sline,slinepos-1,NEXTLINE_RIGHT);
	}
      else if (eline==sline+1)
	{
	  // starts and ends on subsequent lines
	  *out << prefstr << sname << " lines " << sline << " and " << eline << ":\n";
	  *out << preflinestr;
	  JRShowStreamLine(out,sisp,sline,slinepos-1,NEXTLINE_RIGHT);
	  *out << preflinestr;
	  JRShowStreamLine(out,sisp,eline,PRIORLINE_LEFT,elinepos-1);
	}
      else
	{
	  // starts and ends on different lines
	  *out << prefstr << sname << " lines " << sline << "-" << eline << ":\n";
	  *out << preflinestr;
	  JRShowStreamLine(out,sisp,sline,slinepos-1,NEXTLINE_RIGHT);
	  *out << preflinestr;
	  JRShowStreamLine(out,sisp,eline,PRIORLINE_LEFT,elinepos-1);
	}
    }
  else
    {
      // starts and ends in different files
      // JESSE NOTES 12/28/98:  this is no longer allowed to happen by the configpp interpreter
      //  But someone might still want to allow it.
      if (sline==0 && eline==0)
	{
	  // two sep. streams with no line numbers
	  *out << prefstr << "Starts in " << sname << ", ends in " << ename << ".\n";
	}
      else if (eline==-1)
	{
	  // ends at eof
	  *out << prefstr << "Starts in ";
	  if (sline==0)
	    {
	      // no line number, so just tell them the stream name.
	      *out << sname << ".\n";
	    }
	  else
	    {
	      // so tell them the line number and show them the line and (possibly) position
	      *out << sname << " line " << sline << " and forward:\n";
	      *out << preflinestr;
	      JRShowStreamLine(out,sisp,sline,slinepos-1,NEXTLINE_RIGHT);
	    }
	}
      else
	{
	  // there are lines to show
	  // first show starting file loc
	  *out << prefstr << "Starts in ";
	  if (sline==0)
	    {
	      // no line number, so just tell them the stream name.
	      *out << sname << ".\n";
	    }
	  else
	    {
	      // so tell them the line number and show them the line and (possibly) position
	      *out << sname << " line " << sline << ":\n";
	      *out << preflinestr;
	      JRShowStreamLine(out,sisp,sline,slinepos-1,NEXTLINE_RIGHT);
	    }
	  // next show ending file loc
	  *out << prefstr << "Ends in ";
	  if (eline==0)
	    {
	      // no line number, so just tell them the stream name.
	      *out << ename << ".\n";
	    }
	  else
	    {
	      // so tell them the line number and show them the line and (possibly) position
	      *out << ename << " line " << eline << ":\n";
	      *out << preflinestr;
	      JRShowStreamLine(out,eisp,eline,PRIORLINE_LEFT,elinepos-1);
	    }
	}
    }
}





//--------------------------------------------------------------------------------
//---- JRShowStreamLine() --------------------------------------------------------
//--------------------------------------------------------------------------------

//: Helper for ShowStreamPos.
// Show the line from the stream, and carrot positions if specified

void JRShowStreamLine(ostream *out, istream *isp, long sline, long slinepos, long elinepos)
{
  // helper for ShowStreamPos.  Show the line from the stream, and carrot positions if specified
  string linestring;

  linestring=JRGetStreamLine(out,isp,sline);
  JRmysplicearound(linestring,slinepos,HIGHLIGHT_LEFT,elinepos,HIGHLIGHT_RIGHT);
  *out << linestring << "\n";
}



//--------------------------------------------------------------------------------
//---- JRGetStreamLine() ---------------------------------------------------------
//--------------------------------------------------------------------------------

//: Helper for ShowStreamLine.
//  Read the line specified, return fileposition.

string JRGetStreamLine(ostream *out, istream *isp, long sline)
{
  // helper for ShowStreamLine.  Read the line specified, return fileposition.
  long beforepos;
  string linestring="";
  long curline=1;
  int c;
  int llen=0;

  // first remember where we were in file, then set the pointer to begining of file
  beforepos=isp->tellg();
  isp->seekg(0,ios::beg);

  // now read until we get the line we want
  do
    {
      c=isp->get();
      if (c=='\n' || c==EOF)
	++curline;
      else if (curline==sline)
	{
	  // since this is our line we append the char to our line
	  linestring.append(1,(char)c);
	  if (++llen>=MAX_DISPLAYLINELENGTH)
	    {
	      // truncate it if it is sillily long
	      linestring.append("...");
	      break;
	    }
	}
    } while (c!=EOF && curline<=sline);

  // then reset the file position
  isp->seekg(beforepos,ios::beg);

  // change all tabs to spaces, since tabs print out funny in color
  JRstringcharreplace(linestring,'\t',' ');

  // and return the string
  return linestring;
}




//--------------------------------------------------------------------------------
//---- JRstringcharreplace() -----------------------------------------------------
//--------------------------------------------------------------------------------

//: Replace all occurences of fchar with rchar.

void JRstringcharreplace(string &instring, char fchar, char rchar)
{
  // replace all occurences of fchar with rchar
  int count,len;

  len=instring.length();
  for (count=0;count<len;++count)
    {
      if (instring[count]==fchar)
	instring[count]=rchar;
    }
}



//--------------------------------------------------------------------------------
//---- JRmysplicearound() --------------------------------------------------------
//--------------------------------------------------------------------------------
//: Insert two strings inside another, to the left and right of two positions

void JRmysplicearound(string &instring,long leftpos,string leftinsert,long rightpos,string rightinsert)
{
  // insert two strings inside another
  long leftadded=0;
  string prestr="... ";

  // left insert. do we want a leftmost mark?
  if (leftpos!=PRIORLINE_LEFT || INSERT_MULTILINE_MARKS)
    {
      if (leftpos==PRIORLINE_LEFT)
	{
	  instring.insert(0,prestr);
	  leftadded+=prestr.length();
	}
      if (leftpos<0)
	leftpos=0;
      if (leftpos>=instring.length())
	leftpos=instring.length();
      instring.insert(leftpos,leftinsert);
      leftadded+=leftinsert.length();
    }

  // right insert. do we want a rightmost mark?
  if (rightpos!=NEXTLINE_RIGHT || INSERT_MULTILINE_MARKS)
    {
      if (rightpos==NEXTLINE_RIGHT)
	instring.append(prestr);
      if (rightpos<0)
	rightpos=leftpos;
      rightpos+=leftadded;
      rightpos++;
      if (rightpos>=instring.length())
	{
	  instring.append(rightinsert);
	}
      else
	instring.insert(rightpos,rightinsert);
    }
}




//--------------------------------------------------------------------------------
//---- Configpp::JRlowerstring() --------------------------------------------------
//--------------------------------------------------------------------------------

string JRlowerstring(string instring)
{
  // return a lower case version of the string
  int len;

  len=instring.length();
  for (int i=0;i<len;++i)
    instring[i]=tolower(instring[i]);

  return instring;
}







//--------------------------------------------------------------------------------
//---- JRAliasmap::addalias() ----------------------------------------------------
//--------------------------------------------------------------------------------

void JRAliasmap::addalias(string alias,string keyword,bool casesensitive)
{
  // add a new alias to the list of aliases
  // if casesensitive==false, then comparisons of the key will be done in a
  //  case insensitive fashion.  It is okay to pass alias=keyword, such
  //  that the alias acts like a lower case conversion.
  map<string, string>::iterator p;
  
  if (casesensitive)
    {
      // we add the normal case-senstive alias
      if (alias != keyword)
	csaliasmap.insert(make_pair(alias,keyword));
    }
  else
    {
      // we add the case-insensitive alias
      string lowercaseversion=JRlowerstring(alias);
      cisaliasmap.insert(make_pair(lowercaseversion,keyword));
    }
}



//--------------------------------------------------------------------------------
//---- Configpp::aliases() --------------------------------------------------------
//--------------------------------------------------------------------------------

string JRAliasmap::alias(string wordstr)
{
  // see if wordstr is an alias for another canonical string, if so return the
  // canonical string, otherwise, return wordstr.
  // do replacements iteratively, to support alias chains up to depth 20, then bail out.
  map<string, string>::iterator p;
  string lowercaseversion;

  // allow up to depth 20 recursion of aliases, then bail out
  for (int depth=0;depth<20;++depth)
    {
      // first check the case sensitive alias map
      p=csaliasmap.find(wordstr);
      if (p==csaliasmap.end())
	{
	  // if there are no case-sensitive aliases, check the case-insensitive map
	  lowercaseversion=JRlowerstring(wordstr);
	  p=cisaliasmap.find(lowercaseversion);
	  if (p==cisaliasmap.end())
	    break;
	}
      // found an alias
      if (p->second!=wordstr)
	wordstr=p->second;
      else
	break;
    }

  return wordstr;
}




//--------------------------------------------------------------------------------
//---- JRstripword() -------------------------------------------------------------
//--------------------------------------------------------------------------------

string JRstripword(string &instring)
{
  // strip off and return any leading word (before space)
  // (ie from "robot chain link", return "robot" and modify instring="chain link")
  int ploc;
  string leadstr;
  int slen;

  // first clean the string of any leading or trailing spaces
  JRcleanstring(instring);

  slen=instring.length();
  ploc=instring.find(' ',0);

  if (ploc<0)
    {
      // no spaces or cmmas so return entire string as the word
      leadstr=instring;
      instring="";
    }
  else
    {
      leadstr=instring.substr(0,ploc);
      if (ploc>0)
	instring.erase(0,ploc+1);
    }

  return leadstr;
}


//--------------------------------------------------------------------------------
//---- JRstriptoken() ------------------------------------------------------------
//--------------------------------------------------------------------------------

string JRstriptoken(string &instring)
{
  // strip off and return any leading token (before space or \0 or comma, and with [ being its own token)
  // (ie from "[0 1 2], return "[" and modify instring="0 1 2]")
  int count,ploc;
  string leadstr,charstring;
  int slen;

  // first clean the string of any leading or trailing spaces
  JRcleanstring(instring);
  slen=instring.length();
  if (slen==0)
    return "";

  // first check for single character tokens
  for (count=0;count<slen;++count)
    {
      charstring=instring.substr(count,1);
      if (charstring==" " || charstring=="," || charstring=="[" || charstring=="]")
	break;
    }

  // if first character is a divider, then it is the token
  if (count==0)
    count=1;

  if (count>=slen)
    {
      // no spaces or cmmas so return entire string as the word
      leadstr=instring;
      instring="";
    }
  else
    {
      leadstr=instring.substr(0,count);
      instring.erase(0,count);
      // now clean remainder
      JRcleanstring(instring);
    }

  return leadstr;
}



//--------------------------------------------------------------------------------
//---- JRcleanstring()
//--------------------------------------------------------------------------------
//: remove any leading or trailing spaces

void JRcleanstring(string &instring)
{
  int len;

  // first remove leading spaces
  while ((len=instring.length())>0)
    {
      if (instring.substr(0,1)!=" ")
	break;
      else
	instring.erase(0,1);
    }

  // now remove trailing spaces
  while ((len=instring.length())>0)
    {
      if (instring.substr(len-1,1)!=" ")
	break;
      else
	instring.erase(len-1,1);
    }
}



//--------------------------------------------------------------------------------
//---- JRstripline() -------------------------------------------------------------
//--------------------------------------------------------------------------------

string JRstripline(string &instring)
{
  // strip off and return the any leading line (before \n or \0)
  // (ie from "hello there\nworld", return "hello there" and modify instring="world")
  int ploc;
  string leadstr;
  int slen;

  slen=instring.length();
  ploc=instring.find('\n',0);

  if (ploc<0)
    {
      // no spaces so return entire string as the word
      leadstr=instring;
      instring="";
    }
  else
    {
      leadstr=instring.substr(0,ploc);
      instring.erase(0,ploc+1);
    }

  return leadstr;
}







//--------------------------------------------------------------------------------
//---- JRstripleader() -----------------------------------------------------------
//--------------------------------------------------------------------------------

string JRstripleader(string &instring)
{
  // strip off and return any leading trail (before last period) or return ""
  // (ie from robot.chain.link.length, return robot.chain.link and modify instring=length)
  int ploc;
  string leadstr;
  int slen;

  slen=instring.length();
  ploc=instring.rfind('.',slen-1);

  if (ploc>=slen-1 || ploc<=0)
    return "";

  leadstr=instring.substr(0,ploc);
  instring.erase(0,ploc+1);
  return leadstr;
}






//--------------------------------------------------------------------------------
//---- JRstriptail() -------------------------------------------------------------
//--------------------------------------------------------------------------------

string JRstriptail(string &instring)
{
  // strip off and return any trail (before first period) or return instring
  // this is just like stripleader except that we return the tail not leader
  int ploc;
  string tailstr="";
  int slen;

  //  cout << "In striptail with '"+instring+"'.\n";
  slen=instring.length();
  if (slen==0)
    return tailstr;

  ploc=instring.rfind('.',slen-1);

  if (ploc==0)
    {
      // period is first char
      tailstr=instring.substr(ploc+1,(slen-ploc)-1);
      instring="";
    }
  else if (ploc<0)
    {
      // no periods, so we consider this to be that there is no tail (note 
      // be careful because multiple calls to us will never zero out the string!
      tailstr="";
    }
  else if (ploc==slen-1)
    {
      // here we have a period at the very end, how to we want to consider this
      //  return a trail of "" and kill the period
      instring="";
      instring.erase(ploc,1);
    }
  else
    {
      tailstr=instring.substr(ploc+1,(slen-ploc)-1);
      instring.erase(ploc,slen-ploc);
    }

  //  cout << "Leaving striptail with '"+instring+"' + '"+tailstr+"'.\n";
  return tailstr;
}




//--------------------------------------------------------------------------------
//---- JRstriptail2() -------------------------------------------------------------
//--------------------------------------------------------------------------------

string JRstriptail2(string &instring)
{
  // strip off and return any trail (before first period) or return instring
  // this is just like stripleader except that we return the tail not leader
  int ploc;
  string tailstr="";
  int slen;

  //  cout << "In striptail with '"+instring+"'.\n";
  slen=instring.length();
  if (slen==0)
    return tailstr;

  ploc=instring.rfind('.',slen-1);

  if (ploc==0)
    {
      // period is first char
      tailstr=instring.substr(ploc+1,(slen-ploc)-1);
      instring="";
    }
  else if (ploc<0)
    {
      // no periods, so we consider the string to be an entire tail
      tailstr=instring;
      instring="";
    }
  else if (ploc==slen-1)
    {
      // here we have a period at the very end, how to we want to consider this
      //  return a trail of "" and kill the period
      instring="";
      instring.erase(ploc,1);
    }
  else
    {
      tailstr=instring.substr(ploc+1,(slen-ploc)-1);
      instring.erase(ploc,slen-ploc);
    }

  //  cout << "Leaving striptail with '"+instring+"' + '"+tailstr+"'.\n";
  return tailstr;
}






//--------------------------------------------------------------------------------
//---- JRclcmp() -----------------------------------------------------------------
//--------------------------------------------------------------------------------

bool JRclcmp(char *ocomword,char *shortstr,char *longstr)
{
  // a commandline flag helper function
  // return true if the flag is in the word.  AND remove the matching flag
  char *p;
  char *comword;
  char c;

  comword=ocomword;
  p=longstr;
  if (*p!='\0')
    {
      while ( (p=strstr(comword,p)) )
	{
	  comword=p+strlen(longstr);
	  c=*comword;
	  if (c=='\0' || c=='-' || c==' ' || c==',' || c=='\n')
	    {
	      // okay we found a flag, so remove it and return true
	      strcpy(p,comword);
	      return true;
	    }
	};
    }

  comword=ocomword;
  p=shortstr;
  if (*p!='\0')
    {
      while ( (p=strstr(comword,p)) )
	{
	  comword=p+strlen(shortstr);
	  c=*comword;
	  if (c=='\0' || c=='-' || c==' ' || c==',' || c=='\n')
	    {
	      // okay we found a flag, so remove it and return true
	      strcpy(p,comword);
	      return true;
	    }
	};
    }

  // we didn't find it
  return false;
}




//--------------------------------------------------------------------------------
//---- JRinvalidcomargs() --------------------------------------------------------
//--------------------------------------------------------------------------------

bool JRinvalidcomargs(int argc, char **argv, bool complain, string *unknownargstr)
{
  // return true if there are any unaccounted for (non-blank) arguments on the commandline.
  // if string pointer is passed, it will be filled with a string list of the unknown args
  int count;
  bool invalids=false;
  string erstring="";

  for (count=1;count<argc;++count)
    {
      if (argv[count][0]!='\0')
	{
	  // found a non-zero comline arg
	  erstring+=" '";
	  erstring+=argv[count];
	  erstring+="'";
	  invalids=true;
	}
    }

  if (invalids && complain)
    cerr << "\nError.  Invalid commandline option(s):" << erstring << ".\n";

  if (unknownargstr)
    *unknownargstr=erstring;

  return invalids;
}




//--------------------------------------------------------------------------------
//---- JRcompresscomargs() -------------------------------------------------------
//--------------------------------------------------------------------------------
void JRcompresscomargs(int &argc, char **argv)
{
  // any blank command line args are removed.  this function is used when other funcs
  // remove (zero-out) commandline args as they consume them.
  // sets the new number of commandline arguments.
  int count;
  int newargc=1;

  // shift out zeroed out entries
  for (count=1;count<argc;++count)
    {
      if (argv[count][0]!='\0')
	{
	  if (newargc!=count)
	    argv[newargc]=argv[count];
	  ++newargc;
	}
    }

  // should we zero out trailing leftovers in case someone stupid ignores our changed argc
  for (count=newargc;count<argc;++count)
    {
      // adjust the pointer to the '\0';
      argv[count]+=strlen(argv[count]);
    }

  argc=newargc;
}




//--------------------------------------------------------------------------------
//---- JRstripcomarg() -----------------------------------------------------------
//--------------------------------------------------------------------------------

string JRstripcomarg(int &argc,char **argv)
{
  // remove the leading non-blank commandline argument and return it, setting new argc
  string argstr="";

  JRcompresscomargs(argc,argv);
  if (argc<2)
    return argstr; // an error - no arg to strip

  argstr+=argv[1];
  argv[1][0]='\0';
  JRcompresscomargs(argc,argv);
  return argstr;
}




//--------------------------------------------------------------------------------
//---- JRremquotes() -------------------------------------------------------------
//--------------------------------------------------------------------------------

void JRremquotes(string &instring)
{
  // remove enclosing quotes from string
  int slen;

  // check for the common
  if (instring=="\"\"")
    {
      instring="";
      return;
    }

  slen=instring.length();
  if (slen>2)
    {
      if ((instring[0]=='\"' && instring[slen-1]=='\"') || (instring[0]=='\'' && instring[slen-1]=='\''))
	{
	  instring.erase(slen-1,1);
	  instring.erase(0,1);
	}
    }
}




//--------------------------------------------------------------------------------
//---- JRsimiliarstr() -----------------------------------------------------------
//--------------------------------------------------------------------------------

bool JRsimiliarstr(string str1,string str2)
{
  // return true if the strings are "similiar" enough (used for warnings, hints, etc.)
  // do case insensitive matches and substrings,etc.
  int pos;
  char c;

  if (str1=="" || str2=="")
    return false;

  // first lowercase both strings
  str1=JRlowerstring(str1);
  str2=JRlowerstring(str2);

  // first, a case insensitice check
  if (str1==str2)
    return true;

  // now we want to detect non-trivial substring inclusion
  // first of str2 in str1
  pos=0;
  for(;;)
    {
      pos=str1.find(str2,pos);
      if (pos<0)
	break;
      // we found a substr, but might we want to make sure it's non-trivial (it it is not in the middle of a word)
      return true;
      ++pos;
    }
  // now of str1 in str2
  pos=0;
  for(;;)
    {
      pos=str2.find(str1,pos);
      if (pos<0)
	break;
      // we found a substr, but might we want to make sure it's non-trivial (it it is not in the middle of a word)
      return true;
      ++pos;
    }

  return false;
}




//--------------------------------------------------------------------------------
//---- JRescaperep() -------------------------------------------------------------
//--------------------------------------------------------------------------------

string JRescaperep(string instring)
{
  // return a string with any escaped symbols (\n,\t, etc) replace with their
  // ascii equivelants
  string retstring=instring;

  JRfindandreplace(retstring,"\\n",'\n');
  JRfindandreplace(retstring,"\\t",'\t');

  return retstring;
}


//--------------------------------------------------------------------------------
//---- JRfindandreplace() --------------------------------------------------------
//--------------------------------------------------------------------------------

void JRfindandreplace(string &instring,string findstr,char repc)
{
  // find and replace substrings with characters
  int reps=0;
  int pos;
  int wlen;

  wlen=findstr.length();
  while((pos=instring.find(findstr))>=0)
    {
      instring.erase(pos,wlen);
      if (pos<instring.length())
	instring.insert(pos,1,repc);
      else
	instring.append(1,repc);
      if (++reps>1000)
	{
	  cout << "Forcing exit from JRfindandreplace because we seem to be in endless loop.\n";
	  break;
	}
    }
}


void JRfindandreplace(string &instring,string findstr,string repstr)
{
  // find and replace substrings with substrings (beware infinite recursions)
  int reps=0;
  int pos;
  int wlen;

  wlen=findstr.length();
  while((pos=instring.find(findstr))>=0)
    {
      instring.erase(pos,wlen);
      if (pos<instring.length())
	instring.insert(pos,repstr);
      else
	instring.append(repstr);
      if (++reps>1000)
	{
	  cout << "Forcing exit from JRfindandreplace because we seem to be in endless loop.\n";
	  break;
	}
    }
}






//--------------------------------------------------------------------------------
//---- JRtostr() -----------------------------------------------------------------
//--------------------------------------------------------------------------------
string JRtostr(double val)
{
  char numbert[NUMBERSTR_MAXSIZE];

  sprintf(numbert,"%g",val);
  return string(numbert);
}


//--------------------------------------------------------------------------------
//---- JRtostr() -----------------------------------------------------------------
//--------------------------------------------------------------------------------
string JRtostr(int val)
{
  char numbert[NUMBERSTR_MAXSIZE];

  sprintf(numbert,"%d",val);
  return string(numbert);
}



//--------------------------------------------------------------------------------
//---- JRtostr() -----------------------------------------------------------------
//--------------------------------------------------------------------------------
string JRtostr(long val)
{
  char numbert[NUMBERSTR_MAXSIZE];

  sprintf(numbert,"%ld",val);
  return string(numbert);
}



//--------------------------------------------------------------------------------
//---- JRdtostr() ----------------------------------------------------------------
//--------------------------------------------------------------------------------

string JRdtostr(double val,string formatstr="")
{
  // return a string which is a "%lg" sprintf formated printing of a long double
  // if formatstr is non-blank, it will be used in place of "%lg"
  // warning, as with all sprintf type funcs, stack may get corrupt if format
  //  does not match arguments
  char numbert[NUMBERSTR_MAXSIZE];

  if (formatstr=="")
    sprintf(numbert,"%g",val);
  else
    sprintf(numbert,formatstr.c_str(),val);

  return string(numbert);
}



//--------------------------------------------------------------------------------
//---- JRstradr()
//--------------------------------------------------------------------------------

const char *JRstradr(string *instring)
{
  // return instring->c_str() or NULL if blank

  if (*instring=="")
    return NULL;
  else
    return instring->c_str();
}





// BAD - see JRrand.h for the better random number generators

double JRrand01()
{
  // simple uniform random from 0 to 1
  // NOTE: drand48 seems to be returning large values instead of [0.0,1.0] so i am using rand()/RAND_MAX
  return (rand()/(double)RAND_MAX);
}









//--------------------------------------------------------------------------------
//---- JRcleancharp()
//--------------------------------------------------------------------------------

void JRcleancharp(char *instring, string commentprefix)
{
  // remove anything after comments
  char *pos;
  int len;

  pos=strstr(instring,commentprefix.c_str());
  if (pos!=NULL)
    *pos='\0';

  len=strlen(instring);
  while (len>0)
    {
      if (instring[len-1]<33)
	instring[--len]='\0';
      else
	break;
    }

  if (instring[len-1]==';')
    instring[--len]='\0';
}





//--------------------------------------------------------------------------------
//---- JRstriptok()
//--------------------------------------------------------------------------------

void JRstriptok(char *instring, char *outword)
{
  // strip off and return any leading word (before space or \0)
  // (ie from "robot chain link", return "robot" and modify instring="chain link")
  int ploc=0;
  int slen;
  int count,count2;
  int c;

  // first kill any initial spaces,etc.
  slen=strlen(instring);
  for (count=0;count<slen;++count)
    {
      c=instring[count];
      if (c!=' ' && c!='\t' && c!=',')
	break;
    }
  if (count>0)
    {
      for (count2=0;count2<=slen-count;++count2)
	instring[count2]=instring[count2+count];
    }

  // now pick off the first word
  slen=strlen(instring);
  for (count=0;count<slen;++count)
    {
      c=instring[count];
      if (c==' ' || c=='\t' || c==',')
	break;
    }

  outword[0]='\0';
  for (count2=0;count2<count;++count2)
    outword[count2]=instring[count2];
  outword[count2]='\0';
  for (count2=1;count2<=slen-count;++count2)
    instring[count2-1]=instring[count2+count];
  instring[count2-1]='\0';
}




//--------------------------------------------------------------------------------
//---- JRfprintf()
//--------------------------------------------------------------------------------
void JRfprintf(ostream *ostreamp,const char *str, ...)
{
  // ue vprintf to duplicate sprintf then send to output stream
  int len;
  char *tempstr;
  va_list arg;

  // initialize the variable-argument pointer
  va_start(arg, str);

  // allocate space for the sprintf
  len=strlen(str);
  if (len<100)
    len=200;
  tempstr=new char[len];

  // now perform the sprintf
  vsprintf(tempstr,str,arg);
  // and write it to the output stream
  ostreamp->write(tempstr,strlen(tempstr));

  // cleanup
  delete [] tempstr;
  va_end(arg);
}





//--------------------------------------------------------------------------------
//---- JRstrprintf()
//--------------------------------------------------------------------------------
string JRstrprintf(const char *str, ...)
{
  // sprintf then convert to string
  int len;
  char *tempstr;
  va_list arg;
  string resultstring;

  // initialize the variable-argument pointer
  va_start(arg, str);

  // allocate space for the sprintf
  len=strlen(str);
  if (len<100)
    len=200;
  tempstr=new char[len];

  // now perform the sprintf
  vsprintf(tempstr,str,arg);

  // now convert to a string
  resultstring=(string)tempstr;

  // cleanup
  delete [] tempstr;
  va_end(arg);
  return resultstring;
}


//--------------------------------------------------------------------------------
//---- JRtostr(char *str)
//--------------------------------------------------------------------------------
string JRtostr(char *str)
{
  // convert char* to string
  string s;
  s=string(str);
  return s;
}




//--------------------------------------------------------------------------------
//---- JRsgn()
//--------------------------------------------------------------------------------
float JRsgn(float val)
{
  if (val==0)
    return 0.0;
  else if (val<0)
    return -1.0;
  else
    return 1.0;
}



// }
// Namespace close-brace

